/* $Id: camera.c,v 1.13 2004/04/01 02:30:30 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <sys/param.h>
#include <sys/time.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "timers.h"
#include "pngloader.h"
#include "jpegloader.h"
#include "mappings.h"
#include "defines.h"
#include "utils.h"
#include "camera.h"
#include "cobject.h"
#include "stattables.h"

extern int MWidth, MHeight;
extern cmap_t mapcache[9];
extern RMnode *mainchar, *rootnode;
camera_t maincamera;
extern uniqid_t player_id;
extern RMcolor4D curfogcolor;

void init_camera(void)
{
    maincamera.eye.x = EYE_OFF_X;
    maincamera.eye.y = EYE_OFF_Y;
    maincamera.eye.z = EYE_OFF_Z;
    maincamera.at.x = AT_OFF_X;
    maincamera.at.y = AT_OFF_Y;
    maincamera.at.z = AT_OFF_Z;
    maincamera.hither = 1.25;
    maincamera.yon = 120.0;
    maincamera.fov = 45.0;
    maincamera.camera = CAMERA_DEFAULT;
}

void addFog(RMnode *n, float start, float end)
{
    RMfog *f;
    
    f = rmFogNew();
    
    rmFogSetMode(f, GL_LINEAR);
    rmFogSetStartEnd(f, start, end);
    rmFogSetColor(f, &curfogcolor);
    rmNodeSetSceneFog(n, f);
    rmFogDelete(f);
}

/* the caller should set eye and at, setscene the cam, and delete it */

RMcamera3D *newCamera(void)
{
    RMcamera3D *camera;
    RMvertex3D up;
    float fov, hither, yon;

    camera = rmCamera3DNew();
    up.x = 0.;
    up.y = 1.;
    up.z = 0.;

    fov = 45.0F;
    hither = 1.25;
    yon = 120.0;
    
    rmCamera3DSetProjection(camera, RM_PROJECTION_PERSPECTIVE);
    rmCamera3DSetUpVector(camera, &up);
    rmCamera3DSetFOV(camera, fov);
    rmCamera3DSetHither(camera, hither);
    rmCamera3DSetYon(camera, yon);
    rmCamera3DSetFocalDistance(camera, 1.0F);
    rmCamera3DSetAspectRatio(camera, (float)MWidth/(float)MHeight);

    return(camera);
}

void changeCamera(int newcam)
{
    client_t *obj;

    obj = find_obj_by_id(&player_id);

    if (obj == NULL || (newcam == maincamera.camera && newcam))
	return;
    else
	maincamera.camera = newcam;

    switch (newcam) {
    case CAMERA_DEFAULT: {
	init_camera();
	maincamera.yon = vision_dist(obj->stats.Per);
	compute_fog(obj);
	break;
    }
    case CAMERA_OVERHEAD: {
	maincamera.eye.x = 5.0;
	maincamera.eye.y = 25.0;
	maincamera.eye.z = 5.0;
	maincamera.at.x = -7.0;
	maincamera.at.y = 0.0;
	maincamera.at.z = -7.0;
	maincamera.hither = 1.25;
	maincamera.yon = 120.0;
	maincamera.fov = 45.0;
	addFog(rootnode, 100.0, 100.0);
	break;
    }
    case CAMERA_FOVERH: {
	maincamera.eye.x = 5.0;
	maincamera.eye.y = 80.0;
	maincamera.eye.z = 5.0;
	maincamera.at.x = 0.0;
	maincamera.at.y = 0.0;
	maincamera.at.z = 0.0;
	maincamera.hither = 1.25;
	maincamera.fov = 5.0;
	maincamera.yon = 120.0;
	addFog(rootnode, 100.0, 100.0);
	break;
    }
    case CAMERA_MAP: {
	maincamera.eye.x = 1.0;
	maincamera.eye.y = 300.0;
	maincamera.eye.z = 1.0;
	maincamera.at.x = 0.0;
	maincamera.at.y = 0.0;
	maincamera.at.z = 0.0;
	maincamera.hither = 1.25;
	maincamera.fov = 0.7;
	maincamera.yon = 1000.0;
	addFog(rootnode, 1000.0, 1000.0);
	break;
    }
    case CAMERA_FAR: {
	maincamera.eye.x = 25.0;
	maincamera.eye.y = EYE_OFF_Y;
	maincamera.eye.z = 25.0;
	maincamera.at.x = -25.0;
	maincamera.at.y = AT_OFF_Y;
	maincamera.at.z = -25.0;
	maincamera.hither = 1.25;
	maincamera.yon = vision_dist(obj->stats.Per);
	maincamera.fov = 45.0;
	compute_fog(obj);
	break;
    }
    case CAMERA_FIRST: {
	maincamera.eye.x = -0.4;
	maincamera.eye.y = 9.0;
	maincamera.eye.z = -0.4;
	maincamera.at.x = -25.0;
	maincamera.at.y = 9.0;
	maincamera.at.z = -25.0;
	maincamera.hither = 1.25;
	maincamera.yon = vision_dist(obj->stats.Per);
	maincamera.fov = 60.0;
	compute_fog(obj);
	break;
    }
    }
}

void moveCamera(RMnode *node)
{
    RMvertex3D eye, at;
    RMmatrix m, old;
    RMcamera3D *camera;
    RMvertex3D *vert;
    float ghy, ghat;

    camera = newCamera();
    rmCamera3DSetFOV(camera, maincamera.fov);
    rmCamera3DSetHither(camera, maincamera.hither);
    rmCamera3DSetYon(camera, maincamera.yon);

    vert = malloc(sizeof(RMvertex3D));
    rmNodeGetTranslateVector(node, vert);

    rmMatrixIdentity(&m);
    m.m[0][3] = maincamera.eye.x;
    m.m[1][3] = vert->y;
    m.m[2][3] = maincamera.eye.z;

    if (rmNodeGetRotateMatrix(node ,&old) == RM_WHACKED)
        rmMatrixIdentity(&old);
    rmMatrixInverse(&old, &old);
    rmMatrixMultiply(&old, &m, &old);

    eye.x = old.m[0][3] + vert->x;
    eye.y = old.m[1][3] + maincamera.eye.y;
    eye.z = old.m[2][3] + vert->z;

    ghy = (float)groundheight_at_xz(mapcache[0].elevation,
        (double)eye.x, (double)eye.z, mapcache[0].height, mapcache[0].width);

    if (ghy > eye.y)
	eye.y = ghy + maincamera.eye.y;

    if (maincamera.eye.y + ghy > eye.y + 5.0)
	eye.y = ghy + maincamera.eye.y;

    rmCamera3DSetEye(camera, &eye);

    rmMatrixIdentity(&m);
    m.m[0][3] = maincamera.at.x;
    m.m[1][3] = vert->y;
    m.m[2][3] = maincamera.at.z;

    if (rmNodeGetRotateMatrix(node ,&old) == RM_WHACKED)
        rmMatrixIdentity(&old);
    rmMatrixInverse(&old, &old);
    rmMatrixMultiply(&old, &m, &old);

    at.x = old.m[0][3] + vert->x;
    at.y = old.m[1][3] + maincamera.at.y;
    at.z = old.m[2][3] + vert->z;

    ghat = (float)groundheight_at_xz(mapcache[0].elevation,
        (double)at.x, (double)at.z, mapcache[0].height, mapcache[0].width);

    if (ghat > at.y)
	at.y = ghat + maincamera.at.y;

    if (maincamera.at.y + ghat > at.y + 5.0)
	at.y = ghat + maincamera.at.y;

    rmCamera3DSetAt(camera, &at);

/*     printf("AT: x=%f y=%f z=%f\n", at.x, at.y, at.z); */
/*     printf("EYE: x=%f y=%f z=%f\n", eye.x, eye.y, eye.z); */

    rmNodeSetSceneCamera3D(rootnode, camera);
    rmCamera3DDelete(camera);
}

void addCamera(RMnode *node)
{
    RMcamera3D *camera;
    RMvertex3D eye, at, *vert;
    float ghy;

    vert = malloc(sizeof(RMvertex3D));
    rmNodeGetTranslateVector(node, vert);
    init_camera();
    changeCamera(0);
    camera = newCamera();
    rmCamera3DSetFOV(camera, maincamera.fov);
    rmCamera3DSetHither(camera, maincamera.hither);
    rmCamera3DSetYon(camera, maincamera.yon);

    eye.x = maincamera.eye.x + vert->x;
    eye.y = maincamera.eye.y + vert->y;
    eye.z = maincamera.eye.z + vert->z;

    ghy = (float)groundheight_at_xz(mapcache[0].elevation,
	(double)eye.x, (double)eye.z, mapcache[0].height, mapcache[0].width);
    if (ghy > eye.y)
	eye.y = ghy + maincamera.eye.y;    

    at.x = maincamera.at.x + vert->x;
    at.y = maincamera.at.y + vert->y;
    at.z = maincamera.at.z + vert->z;

    rmCamera3DSetAt(camera, &at);
    rmCamera3DSetEye(camera, &eye);
/*    printf("AT: x=%f y=%f z=%f\n", at.x, at.y, at.z);
      printf("EYE: x=%f y=%f z=%f\n", eye.x, eye.y, eye.z);*/
    rmNodeSetSceneCamera3D(rootnode, NULL);
    rmNodeSetSceneCamera3D(rootnode, camera);
    rmCamera3DDelete(camera);
}   
