/* $Id: cobj.c,v 1.38 2004/04/01 02:19:30 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <sys/param.h>
#include <sys/time.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
/*#define bool char*/
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <rm/rmaux.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <GL/gl.h>
#include <GL/glu.h>
#undef False
#undef True
#define bool enum Boolean
#include <cal3d/cal3d_wrapper.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "defines.h"
#include "timers.h"
#include "mappings.h"
#include "cmotion.h"
#include "shstr.h"
#include "jpegloader.h"
#include "pngloader.h"
#include "utils.h"
#include "xmlparse.h"
#include "cobject.h"
#include "modelloader.h"
#include "skelanim.h"
#include "camera.h"
#include "cnet.h"
#include "ui.h"
#include "interface.h"
#include "inventory.h"
#include "settings.h"
#include "stattables.h"
#include "netparser.h"

extern RMnode *mainchar;
extern RMnode *rootnode;
extern RMnode *transnode;
extern RMnode *objnode;
extern RMnode *tobjnode;
extern RMnode *groundnode;
extern RMpipe *mainpipe;
extern int needs_redraw;
extern csettings_t settings;

int requested_movement;
uniqid_t  player_id;
client_t * player_obj = NULL;
TAILQ_HEAD(, _client_t) objects = TAILQ_HEAD_INITIALIZER(objects);

client_t *new_object(void)
{
    client_t *obj;

    obj = smalloc(client_t);
    return(obj);
}

/* find an object by it's ID */

client_t *find_obj_by_id(uniqid_t* id)
{
    client_t *tmp;

    TAILQ_FOREACH(tmp, &objects, next) {
	if (uid_eq(&tmp->uid ,id))
	    return(tmp);
    }
    return(NULL);
}

#if 0
void parse_command_bbox(char **words)
{
    int i, id, scaled, rotated, extended;
    size_t span;
    client_t *obj;
    RMvertex3D bb[2];
    char *tmp;
    RMprimitive *prim;
    RMcolor4D bbColor={1.0,0.5,0.5,0.5};
    RMnode *node;

    obj = NULL;
    scaled = 0;
    rotated = 0;
    extended = 0;
    for (i=0;  i < 50 && words[i] != NULL; i++) {
	span = strcspn(words[i], ":");
	tmp = strrchr(words[i], ':');
	tmp++;
	if (!strncmp(words[i], "id", span)) {
	    id = atoi(tmp);
	    obj = find_obj_by_id(id);
	    /* something is seriously wrong here */
	    if (obj == NULL)
		return;
	} else if (!strncmp(words[i], "lx", span))
	    bb[0].x = atof(tmp)/obj->scale.x;
	else if (!strncmp(words[i], "ly", span))
	    bb[0].y = atof(tmp)/obj->scale.y;
	else if (!strncmp(words[i], "lz", span))
	    bb[0].z = atof(tmp)/obj->scale.z;
	else if (!strncmp(words[i], "ux", span))
	    bb[1].x = atof(tmp)/obj->scale.x;
	else if (!strncmp(words[i], "uy", span))
	    bb[1].y = atof(tmp)/obj->scale.y;
	else if (!strncmp(words[i], "uz", span))
	    bb[1].z = atof(tmp)/obj->scale.z;
    }
    prim = rmPrimitiveNew(RM_BOX3D_WIRE);
    rmPrimitiveSetVertex3D(prim, 2, bb, RM_COPY_DATA, NULL);
    rmPrimitiveSetColor4D(prim, 1, &bbColor, RM_COPY_DATA, NULL);
    if (strncmp(strrchr(obj->model, '.'), ".xgm", 4) == 0) {
	/* add it to the picking node */
	node = rmNodeGetIthChild(obj->node, 1);
	rmNodeAddPrimitive(node, prim);
    } else
	rmNodeAddPrimitive(obj->node, prim);
}
#endif


void model_object(client_t *obj, RMnode *attachnode)
{
    RMprimitive *prim;
    RMmatrix old, new;
    RMnode *tag;
    RMvertex3D cor[2];
    RMcolor4D color={0.1,0.8,0.8,0.0};
    RMvertex3D v;
    char file[MAX_PATH];

    /* XXX
       The tag node is a hideous hack.  By putting a node with a bounding
       box here, with a full alpha channel, we make something pickable
       that can reference the real object.  Grumble.
    */

    if (strncmp(strrchr(obj->model, '.'), ".xgm", 4) == 0) {
	obj->node = rmNodeNew(obj->name, RM_RENDERPASS_ALL,
			      RM_RENDERPASS_ALL);
	sprintf(file, "%s/%s", settings.archdir, obj->model);
	parse_genmodel(obj->node, file, obj);
	rmNodeUnionAllBoxes(obj->node);
	tag = rmNodeNew(obj->name, RM_RENDERPASS_3D,
			RM_RENDERPASS_TRANSPARENT);
	tag->clientData = obj;
	prim = rmPrimitiveNew(RM_BOX3D);
	rmNodeAddPrimitive(tag, prim);
	cor[0] = obj->node->bbox.max;
	cor[1] = obj->node->bbox.min;
	rmPrimitiveSetVertex3D(prim, 2, cor, RM_COPY_DATA, NULL);
	rmPrimitiveSetColor4D(prim, 1, &color, RM_COPY_DATA, NULL);
	rmNodeAddChild(obj->node, tag);
	if (attachnode == NULL)
	    rmNodeAddChild(tobjnode, obj->node);
	else
	    rmNodeAddChild(attachnode, obj->node);
    } else if (strncmp(strrchr(obj->model, '.'), ".xcm", 4) == 0) {
	sprintf(file, "%s/%s", settings.archdir, obj->model);
	obj->node = rmNodeNew(obj->name, RM_RENDERPASS_3D,
			      RM_RENDERPASS_ALL);
	prim = rmPrimitiveNew(RM_USERDEFINED_PRIM);
	rmPrimitiveSetRenderFunc(prim, rmCal3D);
	obj->anim = create_calmodel(file);
	obj->anim->node = obj->node;
	prim->clientData = (void *)obj->anim;
	rmPrimitiveSetBoundingBox(prim, &obj->anim->coremodel->bmin,
				  &obj->anim->coremodel->bmax);
	rmNodeSetBoundingBox(obj->node, &obj->anim->coremodel->bmin,
			     &obj->anim->coremodel->bmax);
	v.x = v.y = v.z = 0.0;
	rmPrimitiveSetVertex3D(prim, 1, &v, RM_COPY_DATA, NULL);
	rmNodeAddPrimitive(obj->node, prim);
	rmNodeUnionAllBoxes(obj->node);
	/* XXX this doesn't work and I have no idea why
rmNodeSetPreTraversalCallback(obj->node, RM_VIEW, rmNodeFrustumCullCallback);*/
	rmNodeSetCenter(obj->node, &v);
	if (attachnode == NULL)
	    rmNodeAddChild(objnode, obj->node);
	else
	    rmNodeAddChild(attachnode, obj->node);
	change_obj_state(obj, obj->state);
    }
    rmNodeSetClientData(obj->node, (void *)obj, NULL);

    rmMatrixIdentity(&new);
    new.m[0][0] = obj->scale.x;
    new.m[1][1] = obj->scale.y;
    new.m[2][2] = obj->scale.z;
    if (rmNodeGetScaleMatrix(obj->node, &old) == RM_WHACKED)
	rmMatrixIdentity(&old);
    rmMatrixMultiply(&old, &new, &old);
    rmNodeSetScaleMatrix(obj->node, &old);
    /* set the rotation and shit too */
    rotate_obj(mainpipe, obj->node, obj->facing, ROT_NORMAL);
    v.x = obj->coord.x;
    v.y = obj->coord.y;
    v.z = obj->coord.z;
    rmNodeSetTranslateVector(obj->node, &v);
}

void compute_fog(client_t *obj)
{
    float vismax, vismin;

    vismax = vision_dist(obj->stats.Per);
    vismin = vismax - 5.0;
    vismax += 30.0;
    /*printf("FOG: %f %f\n", vismin, vismax);*/
    addFog(rootnode, vismin, vismax);
}

void add_main_char(uniqid_t* id)
{
    client_t *obj;

    obj = find_obj_by_id(id);

    if (obj == NULL)
	bailout();
    player_obj=obj;

    mainchar = obj->node;
    forced_redraw();
    /*addFloor(groundnode, 0);*/
    addCamera(obj->node);
    rmNodeRemoveChild(rootnode, objnode);
    rmNodeAddChild(rootnode, objnode);
    /* fix the initial camera offset */
    rotate_self(mainpipe, obj->node, 135.0);
    obj->facing.y = add_euler(obj->facing.y, -135.0);
    obj->direction = obj->facing;
    rotate_self(mainpipe, obj->node, 0);
    update_main_display_from_object(obj);
    /* fixed */
    set_sky_condition(14, 0);
    compute_fog(obj);
    addLighting(rootnode);
    moveCamera(obj->node);
    needs_redraw = 1;
}

void req_self_movement(float dir, int state)
{
    char buf[256];
    client_t *obj;
    static float olddir=0.0;

    if (olddir == 0.0)
	olddir = dir;

    obj = (client_t *)rmNodeGetClientData(mainchar);
    if (olddir != dir) {
	/* tell the other end we want to move in the opposite direction. */
	obj->direction.y = add_euler(obj->direction.y, 180);
	sprintf(buf, "rot dy:%f\r", obj->direction.y/*, 180*/);
	do_write(buf);
    }
    olddir = dir;
    change_player_state(state);
}

void change_obj_state(client_t *obj, int newstate)
{
    int i;

    if (obj->state != newstate) {
	obj->state = newstate;
	if (obj->anim != NULL) {
	    if (newstate == STATE_RUN) {
		i = find_anim_for_state(obj->anim->coremodel, ANIM_RUN);
		if (i != -1)
		    obj->anim->curanimid = i;
	    } else if (newstate == STATE_WALK) {
		i = find_anim_for_state(obj->anim->coremodel, ANIM_WALK);
		if (i != -1)
		    obj->anim->curanimid = i;
	    } else if (newstate == STATE_IDLE) {
		i = find_anim_for_state(obj->anim->coremodel, ANIM_IDLE);
		if (i != -1)
		    obj->anim->curanimid = i;
	    }
	}
    }
}

void change_player_state(int newstate)
{
    client_t *obj;
    char buf[256],b2[40];

    obj = (client_t *)rmNodeGetClientData(mainchar);
    if (obj->state != newstate) {
	change_obj_state(obj, newstate);
	sprintf(buf, "modpl id:%s st:%d\r", uid_sprint(b2,&obj->uid), newstate);
	do_write(buf);
    }
}

/* remvoe an object from the screen */

void remove_obj(client_t *obj)
{
    int i, j;
    RMnode *child;

    i = rmNodeGetNumChildren(obj->node);
    for (j=0; j<i; j++) {
	child = rmNodeGetIthChild(obj->node, 0);
	rmNodeRemoveChild(obj->node, child);
	if (obj->gmod != NULL)
	    if (obj->gmod->node != child)
		rmNodeDelete(child);
	if (obj->anim != NULL) {
	    rmNodeDelete(child);
	    delete_calmodel(obj->anim);
	}
    }
    if (i == 0 && obj->anim != NULL) {
	delete_calmodel(obj->anim);
    }
    rmNodeRemoveChild(tobjnode, obj->node);
    rmNodeRemoveChild(objnode, obj->node);
    rmNodeDelete(obj->node);
    obj->node = NULL;
    if (obj->gmod != NULL)
	obj->gmod->refcount--;
}
