/* $Id: connect.c,v 1.7 2004/03/16 19:18:45 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <inttypes.h>
#include <sys/time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>

#include "defines.h"
#include "timers.h"
#include "utils.h"
#include "shstr.h"
#include "cnet.h"
#include "settings.h"
#include "metaserv.h"

extern GtkWidget *window, *glarea;
extern csettings_t settings;
extern int connected;
extern TAILQ_HEAD(, _mservdata) mservdata;
extern csettings_t settings;
/* XXX */
extern void init_metaservers(void);

GtkWidget *username, *charname, *password, *server;
int got_mservdata;
GtkListStore *list_store;
GtkTreePath *cur_connect_id;

enum {
    MS_COLUMN_NAME,
    MS_COLUMN_HOSTNAME,
    MS_COLUMN_USERS,
    MS_COLUMN_ID,
    MS_COLUMN_PORTNUM,
    MS_N_COLUMNS
};

static gint do_connect(GtkWidget *widget)
{
    char *uname, *pword, *genserver;
    char *buf, *p;
    int ret;
    int port;

    uname = (char *)gtk_entry_get_text(GTK_ENTRY(username));
    pword = (char *)gtk_entry_get_text(GTK_ENTRY(password));
    genserver = (char *)gtk_entry_get_text(GTK_ENTRY(server));
    printf("do connect: %s %s %s\n", uname, pword, genserver);

    if (strlen(uname) == 0 || strlen(pword) == 0)
        return(TRUE);
    if (strlen(genserver) == 0)
        genserver = "localhost";

    p = strchr(genserver, ':');
    if (p == NULL)
	port = 0;
    else {
	*p = '\0';
	p++;
	if (*p == '\0')
	    port = 0;
	else
	    port = atoi(p);
    }

    ret = connect_server(genserver, 0);
    if (ret == 0)
        return(TRUE);

    connected++;
    buf = safer_malloc(sizeof(char)*strlen(uname)+ strlen(pword)+20);
    sprintf(buf, "adpl nm:%s pass:%s\r", uname, pword);
    do_write(buf);
    gtk_widget_destroy(widget);
    return(TRUE);
}

static gint cancel_connect(GtkWidget *widget)
{
   gtk_widget_destroy(widget);
   return(FALSE);
}

void draw_list_from_servers(void)
{
    GtkTreeIter iter;
    mservdata_t *dat;

    /* clean the tree out */
    while (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(list_store), &iter))
	gtk_list_store_remove(list_store, &iter);

    TAILQ_FOREACH(dat, &mservdata, next) {
	gtk_list_store_append(list_store, &iter);
	if (dat->preferip)
	    gtk_list_store_set(list_store, &iter, MS_COLUMN_NAME, dat->name,
			   MS_COLUMN_HOSTNAME, dat->ip,
			   MS_COLUMN_USERS, dat->numusers,
			   MS_COLUMN_ID, dat->id,
			   MS_COLUMN_PORTNUM, dat->port, -1);
	else
	    gtk_list_store_set(list_store, &iter, MS_COLUMN_NAME, dat->name,
			   MS_COLUMN_HOSTNAME, dat->hostname,
			   MS_COLUMN_USERS, dat->numusers,
			   MS_COLUMN_ID, dat->id,
			   MS_COLUMN_PORTNUM, dat->port, -1);
    }
}

static gint check_metaserver_datastream(GtkWidget *widget)
{
    if (got_mservdata == 0 || connected)
	return(TRUE);

    draw_list_from_servers();
    return(FALSE);
}

int update_from_metaserver(void)
{
    int ret;
    char buf[40];

    if (connected)
	return(0);

    ret = connect_server(settings.metaservername, settings.metaserverport);
    if (ret == 0)
	return(ret);

    got_mservdata = 0;

    sprintf(buf, "get\r");
    do_write(buf);
    gtk_idle_add((GtkFunction)check_metaserver_datastream, NULL);
    return(1);
}

static gint do_update_from_metaserver(GtkWidget *widget)
{
    (void)update_from_metaserver();
    return(TRUE);
}

static gint do_refresh_from_metaserver(GtkWidget *widget)
{
    while (TAILQ_FIRST(&mservdata) != NULL)
	TAILQ_REMOVE(&mservdata, TAILQ_FIRST(&mservdata), next);
    init_metaservers();
    (void)update_from_metaserver();
    return(TRUE);
}

static gboolean connect_selection_func(GtkTreeSelection *selection,
				       GtkTreeModel *model,
				       GtkTreePath *path, gboolean pcurr,
				       gpointer userdata)
{
    GtkTreeIter iter;

    if (gtk_tree_model_get_iter(model, &iter, path))
    {
	gchar *hostname;
	int portnum;
	char *buf;

	gtk_tree_model_get(model, &iter, MS_COLUMN_HOSTNAME, &hostname,
			   MS_COLUMN_PORTNUM, &portnum, -1);
	if (strlen(hostname) < 1)
	    return(FALSE);
	buf = safer_malloc(sizeof(char) * (strlen(hostname) + 10));
	if (portnum != settings.defserverport)
	    sprintf(buf, "%s:%d", hostname, portnum);
	else
	    sprintf(buf, "%s", hostname);
	gtk_entry_set_text(GTK_ENTRY(server), buf);
	cur_connect_id = path;
	return TRUE; /* allow selection state to change */
    }
    return(FALSE);
}



/* create and display the connect to server window */

gint splash_up(GtkWidget * widget)
{
    GtkWidget *splash;
    GtkWidget *username_menu, *charname_menu;
    GtkWidget *getlist_button, *refresh_button;
    GtkWidget *cancel_button, *connect_button;
    GtkWidget *label, *vbox1, *alignment1, *hbox1, *image, *vseparator1;
    GtkWidget *scrolledwindow1, *table1, *hseparator;
    GtkWidget *treeview1;
    GtkWidget *game_logo;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    GtkTreeSelection *iselection;
    char logoname[MAX_PATH];

    got_mservdata = 0;
    splash = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(splash), "Connect to Server");
    gtk_window_set_position(GTK_WINDOW(splash), GTK_WIN_POS_MOUSE);
    gtk_widget_set_size_request(splash, 430, 450);

    table1 = gtk_table_new(10, 5, FALSE);
    gtk_widget_show(table1);
    gtk_container_add(GTK_CONTAINER(splash), table1);

    charname_menu = gtk_option_menu_new();
    gtk_widget_show(charname_menu);
    gtk_table_attach(GTK_TABLE(table1), charname_menu, 3, 4, 3, 4,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);

    username = gtk_entry_new();
    gtk_widget_show(username);
    gtk_table_attach(GTK_TABLE(table1), username, 3, 4, 1, 2,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);

    charname = gtk_entry_new();
    gtk_widget_show(charname);
    gtk_table_attach(GTK_TABLE(table1), charname, 3, 4, 4, 5,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);

    password = gtk_entry_new();
    gtk_widget_show(password);
    gtk_table_attach(GTK_TABLE(table1), password, 3, 4, 6, 7,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);
    gtk_entry_set_visibility(GTK_ENTRY(password), FALSE);

    server = gtk_entry_new();
    gtk_widget_show(server);
    gtk_table_attach(GTK_TABLE(table1), server, 3, 4, 8, 9,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);

    hseparator = gtk_hseparator_new();
    gtk_widget_show(hseparator);
    gtk_table_attach(GTK_TABLE(table1), hseparator, 2, 4, 2, 3,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);

    hseparator = gtk_hseparator_new();
    gtk_widget_show(hseparator);
    gtk_table_attach(GTK_TABLE(table1), hseparator, 2, 4, 5, 6,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);

    hseparator = gtk_hseparator_new();
    gtk_widget_show(hseparator);
    gtk_table_attach(GTK_TABLE(table1), hseparator, 2, 4, 7, 8,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);

    username_menu = gtk_option_menu_new();
    gtk_widget_show(username_menu);
    gtk_table_attach(GTK_TABLE(table1), username_menu, 3, 4, 0, 1,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);

    label = gtk_label_new("Username");
    gtk_widget_show(label);
    gtk_table_attach(GTK_TABLE(table1), label, 2, 3, 0, 2,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);

    label = gtk_label_new("Character");
    gtk_widget_show(label);
    gtk_table_attach(GTK_TABLE(table1), label, 2, 3, 3, 5,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);

    vbox1 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox1);
    gtk_table_attach(GTK_TABLE(table1), vbox1, 4, 5, 0, 9,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);

    refresh_button = gtk_button_new_from_stock("gtk-refresh");
    gtk_widget_show(refresh_button);
    gtk_box_pack_start(GTK_BOX(vbox1), refresh_button, TRUE, TRUE, 0);
    g_signal_connect_swapped(G_OBJECT(refresh_button), "clicked",
                             G_CALLBACK(do_refresh_from_metaserver),
			     G_OBJECT(splash));

    cancel_button = gtk_button_new_from_stock("gtk-cancel");
    gtk_widget_show(cancel_button);
    gtk_box_pack_start(GTK_BOX(vbox1), cancel_button, TRUE, TRUE, 0);
    g_signal_connect_swapped(G_OBJECT(cancel_button), "clicked",
                             G_CALLBACK(cancel_connect), G_OBJECT(splash));

    connect_button = gtk_button_new();
    gtk_widget_show(connect_button);
    gtk_box_pack_start(GTK_BOX(vbox1), connect_button, TRUE, TRUE, 0);
    g_signal_connect_swapped(G_OBJECT(connect_button), "clicked",
			     G_CALLBACK(do_connect), G_OBJECT(splash));

    alignment1 = gtk_alignment_new(0.5, 0.5, 0, 0);
    gtk_widget_show(alignment1);
    gtk_container_add(GTK_CONTAINER(connect_button), alignment1);

    hbox1 = gtk_hbox_new(FALSE, 2);
    gtk_widget_show(hbox1);
    gtk_container_add(GTK_CONTAINER(alignment1), hbox1);

    image = gtk_image_new_from_stock("gtk-apply", GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(image);
    gtk_box_pack_start(GTK_BOX(hbox1), image, FALSE, FALSE, 0);

    label = gtk_label_new_with_mnemonic("Connect");
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(hbox1), label, FALSE, FALSE, 0);

    getlist_button = gtk_button_new();
    gtk_widget_show(getlist_button);
    gtk_box_pack_start(GTK_BOX(vbox1), getlist_button, TRUE, TRUE, 0);
    g_signal_connect_swapped(G_OBJECT(getlist_button), "clicked",
                             G_CALLBACK(do_update_from_metaserver),
			     G_OBJECT(splash));

    alignment1 = gtk_alignment_new(0.5, 0.5, 0, 0);
    gtk_widget_show(alignment1);
    gtk_container_add(GTK_CONTAINER(getlist_button), alignment1);

    hbox1 = gtk_hbox_new(FALSE, 2);
    gtk_widget_show(hbox1);
    gtk_container_add(GTK_CONTAINER(alignment1), hbox1);

    image = gtk_image_new_from_stock("gtk-execute", GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(image);
    gtk_box_pack_start(GTK_BOX(hbox1), image, FALSE, FALSE, 0);

    label = gtk_label_new_with_mnemonic("Get Servers");
    gtk_widget_show(label);
    gtk_box_pack_start(GTK_BOX(hbox1), label, FALSE, FALSE, 0);

    vseparator1 = gtk_vseparator_new();
    gtk_widget_show(vseparator1);
    gtk_table_attach(GTK_TABLE(table1), vseparator1, 1, 2, 0, 9,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);

    label = gtk_label_new("Server");
    gtk_widget_show(label);
    gtk_table_attach(GTK_TABLE(table1), label, 2, 3, 8, 9,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);

    scrolledwindow1 = gtk_scrolled_window_new(NULL, NULL);
    gtk_widget_show(scrolledwindow1);
    gtk_table_attach(GTK_TABLE(table1), scrolledwindow1, 0, 5, 9, 10,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);

    /* treeview */

    treeview1 = gtk_tree_view_new();
    gtk_widget_show(treeview1);
    gtk_container_add(GTK_CONTAINER(scrolledwindow1), treeview1);

    list_store = gtk_list_store_new (MS_N_COLUMNS, G_TYPE_STRING,
				     G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING,
				     G_TYPE_INT);
    gtk_tree_view_set_model(GTK_TREE_VIEW(treeview1),
			    GTK_TREE_MODEL(list_store));
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview1), TRUE);
    draw_list_from_servers();
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes("Name", renderer, "text",
						      MS_COLUMN_NAME, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview1), column);

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes("Hostname", renderer,
	"text", MS_COLUMN_HOSTNAME, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview1), column);

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes("Users Online",
        renderer, "text", MS_COLUMN_USERS, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview1), column);
    iselection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview1));
    gtk_tree_selection_set_select_function(iselection, connect_selection_func,
					   NULL, NULL);

    /* Genecys Logo */

    sprintf(logoname, "%s/genecys.jpg", settings.texdir);
    game_logo = gtk_image_new_from_file(logoname);
    gtk_widget_show(game_logo);
    gtk_table_attach(GTK_TABLE(table1), game_logo, 0, 1, 0, 9,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);
    gtk_widget_set_size_request(game_logo, 128, 128);

    label = gtk_label_new("Password");
    gtk_widget_show(label);
    gtk_table_attach(GTK_TABLE(table1), label, 2, 3, 6, 7,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);

    gtk_widget_show_all(splash);
    return(1);
}
