/* $Id: interface.c,v 1.29 2004/03/31 05:27:49 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#if defined(__linux__)
#include "queue.h"
#include <sys/time.h>
#else
#include <sys/queue.h>
#endif

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gtk/gtkgl.h>
#include <rm/rm.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <locale.h>
#include "cnet.h"
#include "shstr.h"
#include "defines.h"
#include "cobject.h"
#include "camera.h"
#include "ui.h"
#include "utils.h"
#include "inventory.h"
#include "settings.h"
#include "xmlparse.h"
#include "netparser.h"
#include "metaserv.h"
#include "connect.h"
#include "mappings.h"

GtkWidget *spval;
GtkWidget *stmval;
GtkWidget *playername;
GtkWidget *hpval;
GtkWidget *foodval;
GtkWidget *waterval;
GtkWidget *waternum;
GtkWidget *hpnum;
GtkWidget *spnum;
GtkWidget *stmnum;
GtkWidget *foodnum;
GtkWidget *strval;
GtkWidget *dexval;
GtkWidget *wisval;
GtkWidget *intval;
GtkWidget *conval;
GtkWidget *perval;
GtkWidget *Speedval;
GtkWidget *EXPval;
GtkWidget *Apprval;
GtkWidget *ACval;
GtkWidget *Fameval;
GtkWidget *messages;
GtkWidget *commands;
GtkWidget *glarea, *target_glarea;
GtkWidget *window;

struct timeval last_time;
csettings_t settings;
servdefs_t serverdefs;
TAILQ_HEAD(, _mservdata) mservdata = TAILQ_HEAD_INITIALIZER(mservdata);

extern int MWidth, MHeight;
extern int TWidth, THeight;
extern GtkWidget *Inventory;
extern GtkTreeIter b_spell, b_job, b_inv, b_skill, b_key, b_equip;
extern GtkTreeStore *invstore;
extern GtkTreeSelection *iselection;
extern RMnode *mainchar;

void init_settings(void)
{
    settings.archdir = PACKAGE_DATA_DIR ARCH_SUFX;
    settings.etcdir = PACKAGE_DATA_DIR ETC_SUFX;
    settings.mapdir = PACKAGE_DATA_DIR MAP_SUFX;
    settings.texdir = PACKAGE_DATA_DIR TEX_SUFX;
    settings.libdir = PACKAGE_DATA_DIR LIB_SUFX;
    settings.defservername = add_string("localhost");
    settings.defserverport = 5501;
    settings.detail = 0;
    settings.metaservername = add_string("metaserver.genecys.org");
    settings.metaserverport = 5502;
    settings.usemetaserver = 1;
}

void init_metaservers(void)
{
    mservdata_t *dat;
    TAILQ_INIT(&mservdata);
    dat = smalloc(mservdata_t);
    dat->id = add_string("defaultserver:5501");
    dat->name = add_string(settings.defservername);
    dat->port = settings.defserverport;
    dat->desc = add_string("The default server");
    dat->hostname = add_string(settings.defservername);
    dat->lastupdate = dat->firstupdate = time(NULL);
    TAILQ_INSERT_TAIL(&mservdata, dat, next);
}

void parse_settings(void)
{
    char file[MAX_PATH], *sf, *home;
    xmlDocPtr doc;
    xmlNodePtr cur;

    sf = "genclientconfig.xml";
    home = getenv("HOME");

    sprintf(file, "./%s", sf);
    doc = xmlParseFile(file);
    if (doc == NULL && home != NULL) {
	sprintf(file, "%s/%s", home, sf);
	doc = xmlParseFile(file);
    }
    if (doc == NULL) {
	sprintf(file, "%s/%s", settings.etcdir, sf);
	doc = xmlParseFile(file);
    }
    if (doc == NULL) {
	sprintf(file, "/etc/%s", sf);
	doc = xmlParseFile(file);
    }
    if (doc == NULL) {
	fprintf(stderr, "Warning, cannot find settings file %s", sf);
	return;
    }
    /* if we got this file, we found one */

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "Settings file %s is empty", file);
        xmlFreeDoc(doc);
	return;
    }
    if (strcmp(cur->name, "ClientSettings")) {
        fprintf(stderr, "%s: root node != ClientSettings", file);
        xmlFreeDoc(doc);
	return;
    }
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "archdir"))
	    parsestring(doc, cur, &settings.archdir);
	if (!strcmp(cur->name, "etcdir"))
	    parsestring(doc, cur, &settings.etcdir);
	if (!strcmp(cur->name, "mapdir"))
	    parsestring(doc, cur, &settings.mapdir);
	if (!strcmp(cur->name, "texdir"))
	    parsestring(doc, cur, &settings.texdir);
	if (!strcmp(cur->name, "libdir"))
	    parsestring(doc, cur, &settings.libdir);
	if (!strcmp(cur->name, "defservername"))
	    parsestring(doc, cur, &settings.defservername);
	if (!strcmp(cur->name, "defserverport"))
	    parseint(doc, cur, &settings.defserverport);
	if (!strcmp(cur->name, "detail"))
	    parseint(doc, cur, &settings.detail);
	if (!strcmp(cur->name, "metaservername"))
	    parsestring(doc, cur, &settings.metaservername);
	if (!strcmp(cur->name, "metaserverport"))
	    parseint(doc, cur, &settings.metaserverport);
	if (!strcmp(cur->name, "usemetaserver"))
	    parseint(doc, cur, &settings.usemetaserver);
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

void new_pagebutton(GtkWidget *container, const char *label, int target)
{
    GtkWidget *button;

    button = gtk_button_new_with_label(label);
    gtk_box_pack_start(GTK_BOX(container), button, TRUE, TRUE, 0);
    gtk_widget_show(button);
    switch (target) {
    case T_NONE:
	g_signal_connect((gpointer)button, "clicked",
			 G_CALLBACK(command_no_target), NULL);
	break;
    case T_SINGLE:
	g_signal_connect((gpointer)button, "clicked",
			 G_CALLBACK(command_single_target), NULL);
	break;
    case T_SINGOPT:
	g_signal_connect((gpointer)button, "clicked",
			 G_CALLBACK(command_optional_target), NULL);
	break;
    }
}

void new_pageline(GtkWidget *page, const char *l1, int t1,  const char *l2,
		  int t2, const char *l3, int t3)
{
    GtkWidget *hbox;

    hbox = gtk_hbox_new(TRUE, 0);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, FALSE, 0);

    new_pagebutton(hbox, l1, t1);
    new_pagebutton(hbox, l2, t2);
    new_pagebutton(hbox, l3, t3);
}

static void changecam(char *string)
{
    changeCamera(atoi(string));
    if (mainchar != NULL)
	moveCamera(mainchar);
}

static void command_string(GtkWidget *entry)
{
    char *buf, *p, *buf2;
    const char *entry_text;

    entry_text = gtk_entry_get_text(GTK_ENTRY(entry));
    /*buf = malloc(sizeof(char)*strlen(entry_text)+1);
    strncpy(buf, entry_text, strlen(entry_text));
    buf[strlen(entry_text)] = '\0';*/
    buf = strdup(entry_text);
    gtk_entry_set_text(GTK_ENTRY(entry), "");
    gtk_widget_grab_focus(GTK_WIDGET(glarea));
    g_signal_handlers_unblock_by_func(GTK_OBJECT(window),
	GTK_SIGNAL_FUNC(key_press_event), NULL);
    g_signal_handlers_unblock_by_func(GTK_OBJECT(window),
	GTK_SIGNAL_FUNC(key_release_event), NULL);
    p = strchr(buf, ' ');
    /* we have a command with no args, pass it */
    if (p == NULL) {
	buf2 = malloc(sizeof(char)*strlen(buf)+5);
	sprintf(buf2, "%s\r", buf);
	do_write(buf2);
	free(buf);
	free(buf2);
    } else {
	buf2 = malloc(sizeof(char)*strlen(buf)+50);
	*p = '\0';
	p++;

	/* now, we have to dis-assemble what the player typed, and try to
	   interpret what he wanted to do.  We change these arguments, into
	   strings that the server can understand. */

	if (!strcmp(buf, "say"))
	    sprintf(buf2, "say data:\"%s\"\r", p);
	else if (!strcmp(buf, "whisper"))
	    sprintf(buf2, "say volume:1 data:\"%s\"\r", p);
	else if (!strcmp(buf, "shout"))
	    sprintf(buf2, "say volume:3 data:\"%s\"\r", p);
	else
	    sprintf(buf2, "%s %s\r", buf, p);
	do_write(buf2);
	free(buf);
	free(buf2);
    }
}

int main(int argc, char **argv)
{
    GtkWidget *vbox1;
    GtkWidget *Mainmenu;
    GtkWidget *menuitem4;
    GtkWidget *menuitem4_menu;
    GtkWidget *connect1;
    GtkWidget *image3;
    GtkWidget *disconnect1;
    GtkWidget *image4;
    GtkWidget *quit1;
    GtkWidget *cam1;
    GtkWidget *menuitem7;
    GtkWidget *menuitem7_menu;
    GtkWidget *about1;
    GtkWidget *hbox2;
    GtkWidget *vbox2;
    GtkWidget *scrolledwindow1;
    GtkWidget *vbox3;
    GtkWidget *table1;
    GtkWidget *table2;
    GtkWidget *stats;
    GtkWidget *Strlabel;
    GtkWidget *Dexlabel;
    GtkWidget *Wislabel;
    GtkWidget *Intlabel;
    GtkWidget *Conlabel;
    GtkWidget *Perlabel;
    GtkWidget *Speedlabel;
    GtkWidget *AClabel;
    GtkWidget *EXPlabel;
    GtkWidget *Famelabel;
    GtkWidget *Apprlabel;
    GtkWidget *notebook1;
    GtkWidget *emotions_page;
    GtkWidget *common_page;
    GtkWidget *combat_page;
    GtkWidget *page1;
    GtkWidget *page2;
    GtkWidget *page3;
    GtkWidget *scrolledwindow2;
    GtkAccelGroup *accel_group;
    GtkTooltips *tooltips;
    GtkCellRenderer *irenderer;
    GtkTreeViewColumn *icol;

    /* initialize things */
    init_hash_table();
    init_settings();
    parse_settings();
    init_argcomm();
    init_metaservers();
    (void)memset(&serverdefs, 0, sizeof(servdefs_t));

    (void)gettimeofday(&last_time, NULL);
    gtk_init(&argc, &argv);
	 
	 /* gtk set the locale frem the ebv , we need it in C */
    setlocale(LC_NUMERIC,"C");
    gtk_gl_init(&argc, &argv);

    glarea = create_glarea();
    target_glarea = create_target_glarea();

    load_sky_textures();

    tooltips = gtk_tooltips_new();

    accel_group = gtk_accel_group_new();

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_size_request(window, 1024, 768);
    gtk_window_set_title(GTK_WINDOW(window), "Genecys Client");
    gtk_window_set_default_size(GTK_WINDOW(window), 1024, 768);

    vbox1 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox1);
    gtk_container_add(GTK_CONTAINER(window), vbox1);

    Mainmenu = gtk_menu_bar_new();
    gtk_widget_show(Mainmenu);
    gtk_box_pack_start(GTK_BOX(vbox1), Mainmenu, FALSE, FALSE, 0);

    /* connect/disconnect menu */

    menuitem4 = gtk_menu_item_new_with_mnemonic("_Game");
    gtk_widget_show(menuitem4);
    gtk_container_add(GTK_CONTAINER(Mainmenu), menuitem4);

    menuitem4_menu = gtk_menu_new();
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem4), menuitem4_menu);

    connect1 = gtk_image_menu_item_new_with_mnemonic("Connect");
    gtk_widget_show(connect1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), connect1);
    gtk_tooltips_set_tip(tooltips, connect1, "Connect To Server", NULL);

    image3 = gtk_image_new_from_stock("gtk-convert", GTK_ICON_SIZE_MENU);
    gtk_widget_show(image3);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(connect1), image3);

    disconnect1 = gtk_image_menu_item_new_with_mnemonic("Disconnect");
    gtk_widget_show(disconnect1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), disconnect1);
    gtk_tooltips_set_tip(tooltips, disconnect1, "Disconnect from server", NULL);

    image4 = gtk_image_new_from_stock("gtk-cancel", GTK_ICON_SIZE_MENU);
    gtk_widget_show(image4);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(disconnect1), image4);

    quit1 = gtk_image_menu_item_new_from_stock("gtk-quit", accel_group);
    gtk_widget_show(quit1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), quit1);

    /* camera menu */

    menuitem4 = gtk_menu_item_new_with_mnemonic("Camera");
    gtk_widget_show(menuitem4);
    gtk_container_add(GTK_CONTAINER(Mainmenu), menuitem4);

    menuitem4_menu = gtk_menu_new();
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem4), menuitem4_menu);

    cam1 = gtk_image_menu_item_new_with_mnemonic("Default");
    gtk_widget_show(cam1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), cam1);
    gtk_tooltips_set_tip(tooltips, cam1, "Set camera to default", NULL);
    g_signal_connect_swapped((gpointer)cam1, "activate", G_CALLBACK(changecam),
		     (gpointer)g_strdup("0"));

    cam1 = gtk_image_menu_item_new_with_mnemonic("Overhead");
    gtk_widget_show(cam1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), cam1);
    gtk_tooltips_set_tip(tooltips, cam1, "Overhead camera", NULL);
    g_signal_connect_swapped((gpointer)cam1, "activate", G_CALLBACK(changecam),
		     (gpointer)g_strdup("1"));

    cam1 = gtk_image_menu_item_new_with_mnemonic("Far Overhead");
    gtk_widget_show(cam1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), cam1);
    gtk_tooltips_set_tip(tooltips, cam1, "Far overhead camera", NULL);
    g_signal_connect_swapped((gpointer)cam1, "activate", G_CALLBACK(changecam),
		     (gpointer)g_strdup("2"));

    cam1 = gtk_image_menu_item_new_with_mnemonic("Map");
    gtk_widget_show(cam1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), cam1);
    gtk_tooltips_set_tip(tooltips, cam1, "Map View", NULL);
    g_signal_connect_swapped((gpointer)cam1, "activate", G_CALLBACK(changecam),
		     (gpointer)g_strdup("3"));

    cam1 = gtk_image_menu_item_new_with_mnemonic("Default Far");
    gtk_widget_show(cam1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), cam1);
    gtk_tooltips_set_tip(tooltips, cam1, "Default camera, but farther", NULL);
    g_signal_connect_swapped((gpointer)cam1, "activate", G_CALLBACK(changecam),
		     (gpointer)g_strdup("4"));

    cam1 = gtk_image_menu_item_new_with_mnemonic("First Person");
    gtk_widget_show(cam1);
    gtk_container_add(GTK_CONTAINER(menuitem4_menu), cam1);
    gtk_tooltips_set_tip(tooltips, cam1, "A first person camera", NULL);
    g_signal_connect_swapped((gpointer)cam1, "activate", G_CALLBACK(changecam),
		     (gpointer)g_strdup("5"));

    /* Help menu */

    menuitem7 = gtk_menu_item_new_with_mnemonic("_Help");
    gtk_widget_show(menuitem7);
    gtk_container_add(GTK_CONTAINER(Mainmenu), menuitem7);

    menuitem7_menu = gtk_menu_new();
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem7), menuitem7_menu);

    about1 = gtk_menu_item_new_with_mnemonic("_About");
    gtk_widget_show(about1);
    gtk_container_add(GTK_CONTAINER(menuitem7_menu), about1);



    hbox2 = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox2);
    gtk_box_pack_start(GTK_BOX(vbox1), hbox2, TRUE, TRUE, 0);

    vbox2 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox2);
    gtk_box_pack_start(GTK_BOX(hbox2), vbox2, TRUE, TRUE, 0);

    gtk_widget_show(glarea);
    gtk_widget_show(target_glarea);
    gtk_box_pack_start(GTK_BOX(hbox2), glarea, FALSE, FALSE, 0);
/*    gtk_widget_set_size_request(glarea, MWidth, MHeight);*/

    gtk_box_pack_start(GTK_BOX(vbox2), target_glarea, FALSE, FALSE, 0);
/*    gtk_widget_set_size_request(target_glarea, TWidth, THeight);*/

    scrolledwindow1 = gtk_scrolled_window_new(NULL, NULL);
    gtk_widget_show(scrolledwindow1);
    gtk_box_pack_start(GTK_BOX(vbox2), scrolledwindow1, TRUE, TRUE, 0);

    /* Inventory tree */
    Inventory = gtk_tree_view_new();
    gtk_widget_show(Inventory);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolledwindow1),
					  Inventory);
    gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(Inventory)), GTK_SELECTION_SINGLE);
    icol = gtk_tree_view_column_new();
    gtk_tree_view_column_set_title(icol, "Object");
    gtk_tree_view_append_column(GTK_TREE_VIEW(Inventory), icol);
    irenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_column_pack_start(icol, irenderer, TRUE);
    gtk_tree_view_column_add_attribute(icol, irenderer, "text", COL_OBJ_NAME);
    invstore = gtk_tree_store_new(NUM_COLS, G_TYPE_STRING, G_TYPE_POINTER);

    create_buckets(&b_spell, "Spells");
    create_buckets(&b_job, "Jobs");
    create_buckets(&b_inv, "Inventory");
    create_buckets(&b_skill, "Skills");
    create_buckets(&b_key, "Key Items");
    create_buckets(&b_equip, "Equipment");
    gtk_tree_view_set_model(GTK_TREE_VIEW(Inventory), GTK_TREE_MODEL(invstore));
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(Inventory), FALSE);
    iselection = gtk_tree_view_get_selection(GTK_TREE_VIEW(Inventory));
    gtk_tree_selection_set_select_function(iselection, view_selection_func,
					   NULL, NULL);

    vbox3 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox3);
    gtk_box_pack_start(GTK_BOX(hbox2), vbox3, TRUE, TRUE, 0);

    table1 = gtk_table_new(6, 3, FALSE);
    gtk_widget_show(table1);
    gtk_box_pack_start(GTK_BOX(vbox3), table1, FALSE, TRUE, 0);

    playername = gtk_label_new("Bob");
    gtk_widget_show(playername);
    gtk_table_attach(GTK_TABLE(table1), playername, 0, 3, 0, 1,
		     (GtkAttachOptions)(0),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_misc_set_alignment(GTK_MISC(playername), 0, 0.5);

    /* Progress Bars */

    hpval = gtk_progress_bar_new();
    gtk_widget_show(hpval);
    gtk_table_attach(GTK_TABLE(table1), hpval, 0, 3, 1, 2,
		     (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);

    spval = gtk_progress_bar_new();
    gtk_widget_show(spval);
    gtk_table_attach(GTK_TABLE(table1), spval, 0, 3, 2, 3,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);

    stmval = gtk_progress_bar_new();
    gtk_widget_show(stmval);
    gtk_table_attach(GTK_TABLE(table1), stmval, 0, 3, 3, 4,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);

    foodval = gtk_progress_bar_new();
    gtk_widget_show(foodval);
    gtk_table_attach(GTK_TABLE(table1), foodval, 0, 3, 4, 5,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);

    waterval = gtk_progress_bar_new();
    gtk_widget_show(waterval);
    gtk_table_attach(GTK_TABLE(table1), waterval, 0, 3, 5, 6,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);


    table2 = gtk_table_new(7, 4, FALSE);
    gtk_widget_show(table2);
    gtk_box_pack_start(GTK_BOX(vbox3), table2, FALSE, FALSE, 0);

    stats = gtk_label_new("Stats");
    gtk_widget_show(stats);
    gtk_table_attach(GTK_TABLE(table2), stats, 0, 4, 0, 1,
		     (GtkAttachOptions)(GTK_EXPAND),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_misc_set_alignment(GTK_MISC(stats), 0, 0.5);

    Strlabel = gtk_label_new("Str");
    gtk_widget_show(Strlabel);
    gtk_table_attach(GTK_TABLE(table2), Strlabel, 0, 1, 1, 2,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Strlabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Strlabel), 0, 0.5);

    Dexlabel = gtk_label_new("Dex");
    gtk_widget_show(Dexlabel);
    gtk_table_attach(GTK_TABLE(table2), Dexlabel, 0, 1, 2, 3,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Dexlabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Dexlabel), 0, 0.5);

    Wislabel = gtk_label_new("Wis");
    gtk_widget_show(Wislabel);
    gtk_table_attach(GTK_TABLE(table2), Wislabel, 2, 3, 1, 2,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Wislabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Wislabel), 0, 0.5);

    Intlabel = gtk_label_new("Int");
    gtk_widget_show(Intlabel);
    gtk_table_attach(GTK_TABLE(table2), Intlabel, 2, 3, 2, 3,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Intlabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Intlabel), 0, 0.5);

    strval = gtk_label_new("0");
    gtk_widget_show(strval);
    gtk_table_attach(GTK_TABLE(table2), strval, 1, 2, 1, 2,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(strval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(strval), 0, 0.5);

    dexval = gtk_label_new("0");
    gtk_widget_show(dexval);
    gtk_table_attach(GTK_TABLE(table2), dexval, 1, 2, 2, 3,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(dexval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(dexval), 0, 0.5);

    wisval = gtk_label_new("0");
    gtk_widget_show(wisval);
    gtk_table_attach(GTK_TABLE(table2), wisval, 3, 4, 1, 2,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(wisval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(wisval), 0, 0.5);

    intval = gtk_label_new("0");
    gtk_widget_show(intval);
    gtk_table_attach(GTK_TABLE(table2), intval, 3, 4, 2, 3,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(intval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(intval), 0, 0.5);

    Conlabel = gtk_label_new("Con");
    gtk_widget_show(Conlabel);
    gtk_table_attach(GTK_TABLE(table2), Conlabel, 0, 1, 3, 4,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Conlabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Conlabel), 0, 0.5);

    Perlabel = gtk_label_new("Per");
    gtk_widget_show(Perlabel);
    gtk_table_attach(GTK_TABLE(table2), Perlabel, 2, 3, 3, 4,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Perlabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Perlabel), 0, 0.5);

    conval = gtk_label_new("0");
    gtk_widget_show(conval);
    gtk_table_attach(GTK_TABLE(table2), conval, 1, 2, 3, 4,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(conval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(conval), 0, 0.5);

    perval = gtk_label_new("0");
    gtk_widget_show(perval);
    gtk_table_attach(GTK_TABLE(table2), perval, 3, 4, 3, 4,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(perval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(perval), 0, 0.5);

    Speedlabel = gtk_label_new("Speed");
    gtk_widget_show(Speedlabel);
    gtk_table_attach(GTK_TABLE(table2), Speedlabel, 0, 1, 4, 5,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Speedlabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Speedlabel), 0, 0.5);

    AClabel = gtk_label_new("AC");
    gtk_widget_show(AClabel);
    gtk_table_attach(GTK_TABLE(table2), AClabel, 2, 3, 4, 5,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(AClabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(AClabel), 0, 0.5);

    EXPlabel = gtk_label_new("EXP");
    gtk_widget_show(EXPlabel);
    gtk_table_attach(GTK_TABLE(table2), EXPlabel, 0, 1, 5, 6,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(EXPlabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(EXPlabel), 0, 0.5);

    Famelabel = gtk_label_new("Fame");
    gtk_widget_show(Famelabel);
    gtk_table_attach(GTK_TABLE(table2), Famelabel, 2, 3, 5, 6,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Famelabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Famelabel), 0, 0.5);

    Apprlabel = gtk_label_new("Appr");
    gtk_widget_show(Apprlabel);
    gtk_table_attach(GTK_TABLE(table2), Apprlabel, 0, 1, 6, 7,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Apprlabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Apprlabel), 0, 0.5);

    Speedval = gtk_label_new("0");
    gtk_widget_show(Speedval);
    gtk_table_attach(GTK_TABLE(table2), Speedval, 1, 2, 4, 5,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Speedval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Speedval), 0, 0.5);

    EXPval = gtk_label_new("0");
    gtk_widget_show(EXPval);
    gtk_table_attach(GTK_TABLE(table2), EXPval, 1, 2, 5, 6,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(EXPval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(EXPval), 0, 0.5);

    Apprval = gtk_label_new("0");
    gtk_widget_show(Apprval);
    gtk_table_attach(GTK_TABLE(table2), Apprval, 1, 2, 6, 7,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Apprval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Apprval), 0, 0.5);

    ACval = gtk_label_new("0");
    gtk_widget_show(ACval);
    gtk_table_attach(GTK_TABLE(table2), ACval, 3, 4, 4, 5,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(ACval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(ACval), 0, 0.5);

    Fameval = gtk_label_new("0");
    gtk_widget_show(Fameval);
    gtk_table_attach(GTK_TABLE(table2), Fameval, 3, 4, 5, 6,
		     (GtkAttachOptions)(GTK_FILL),
		     (GtkAttachOptions)(0), 0, 0);
    gtk_label_set_justify(GTK_LABEL(Fameval), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(Fameval), 0, 0.5);

    /* NoteBook */

    notebook1 = gtk_notebook_new();
    gtk_widget_show(notebook1);
    gtk_box_pack_start(GTK_BOX(vbox3), notebook1, TRUE, TRUE, 0);

    /* emotions page */

    emotions_page = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(emotions_page);
    gtk_container_add(GTK_CONTAINER(notebook1), emotions_page);

    new_pageline(emotions_page, "Wave", T_NONE, "Wave", T_NONE,
		 "Wave", T_NONE);

    page1 = gtk_label_new("Emotions");
    gtk_widget_show(page1);
    gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook1), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook1), 0), page1);
    gtk_label_set_justify(GTK_LABEL(page1), GTK_JUSTIFY_LEFT);


    /* common page */
    common_page = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(common_page);
    gtk_container_add(GTK_CONTAINER(notebook1), common_page);

    new_pageline(common_page, "Get", T_SINGOPT, "Wield", T_SINGLE,
		 "Drop", T_SINGLE);
    new_pageline(common_page, "UnWear", T_SINGLE, "Throw", T_SINGLE,
		 "Drop", T_SINGLE);

    page2 = gtk_label_new("Common");
    gtk_widget_show(page2);
    gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook1), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook1), 1), page2);
    gtk_label_set_justify(GTK_LABEL(page2), GTK_JUSTIFY_LEFT);

    /* combat page */

    combat_page = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(combat_page);
    gtk_container_add(GTK_CONTAINER(notebook1), combat_page);

    page3 = gtk_label_new("Combat");
    gtk_widget_show(page3);
    gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook1), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook1), 2), page3);
    gtk_label_set_justify(GTK_LABEL(page3), GTK_JUSTIFY_LEFT);

    scrolledwindow2 = gtk_scrolled_window_new(NULL, NULL);
    gtk_widget_show(scrolledwindow2);
    gtk_box_pack_start(GTK_BOX(vbox1), scrolledwindow2, TRUE, TRUE, 0);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow2), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

    messages = gtk_text_view_new();
    gtk_widget_show(messages);
    gtk_container_add(GTK_CONTAINER(scrolledwindow2), messages);
    gtk_text_view_set_editable(GTK_TEXT_VIEW(messages), FALSE);
    gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(messages), GTK_WRAP_WORD);
    gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(messages), FALSE);

    commands = gtk_entry_new();
    (void)gtk_editable_set_editable(GTK_EDITABLE(commands), TRUE);
    gtk_widget_show(commands);
    gtk_box_pack_start(GTK_BOX(vbox1), commands, FALSE, FALSE, 0);

    g_signal_connect((gpointer)commands, "activate",
		     G_CALLBACK(command_string), NULL);

    g_signal_connect((gpointer)connect1, "activate",
		     G_CALLBACK(splash_up), NULL);
/*    g_signal_connect((gpointer) disconnect1, "activate",
		      G_CALLBACK(on_disconnect1_activate),
		      NULL);
    g_signal_connect((gpointer) about1, "activate",
		      G_CALLBACK(on_about1_activate),
		      NULL);
*/
    g_signal_connect((gpointer)quit1, "activate",
		      GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    gtk_quit_add_destroy(1, GTK_OBJECT(window));

    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
		       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    gtk_signal_connect(GTK_OBJECT(window), "destroy",
		       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
    gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
		       GTK_SIGNAL_FUNC(key_press_event), NULL);
    gtk_signal_connect(GTK_OBJECT(window), "key_release_event",
		       GTK_SIGNAL_FUNC(key_release_event), NULL);
    gtk_container_set_border_width(GTK_CONTAINER(window), 10);

    gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);

    gtk_widget_show(window);

    /* GTK_WIDGET_SET_FLAGS(glarea, GTK_CAN_FOCUS);*/
    GTK_WIDGET_SET_FLAGS(commands, GTK_CAN_FOCUS);
    /*gtk_widget_grab_focus(GTK_WIDGET(glarea));*/

    /* record the actual width and height so the aspect ratios arent hosed */

    MWidth = glarea->allocation.width;
    MHeight = glarea->allocation.height;
    TWidth = glarea->allocation.width;
    THeight = glarea->allocation.height;

    gtk_idle_add((GtkFunction)update_target_screen, target_glarea);
    gtk_idle_add((GtkFunction)update_screen, glarea);
    gtk_idle_add((GtkFunction)networker, glarea);

    gtk_main();

    return(0);
}

void update_main_display_from_object(client_t *obj)
{
    char buf[256];

    gtk_label_set_text(GTK_LABEL(playername), obj->name);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(hpval),
				  (double)obj->stats.hp/
				  (double)obj->stats.maxhp);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(spval),
				  (double)obj->stats.sp/
				  (double)obj->stats.maxsp);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(stmval),
				  (double)obj->stats.stm/
				  (double)obj->stats.maxstm);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(foodval),
				  (double)obj->stats.food/1000.0);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(waterval),
				  (double)obj->stats.water/1000.0);

    sprintf(buf, "Water: %d/1000", obj->stats.water);
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(waterval), buf);
    sprintf(buf, "HP: %d/%d", obj->stats.hp, obj->stats.maxhp);
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(hpval), buf);
    sprintf(buf, "SP: %d/%d", obj->stats.sp, obj->stats.maxsp);
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(spval), buf);
    sprintf(buf, "STM: %d/%d", obj->stats.stm, obj->stats.maxstm);
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(stmval), buf);
    sprintf(buf, "Food: %d/1000", obj->stats.food);
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(foodval), buf);

    gtk_label_set_text(GTK_LABEL(strval), print_stat(obj->stats.Str));
    gtk_label_set_text(GTK_LABEL(dexval), print_stat(obj->stats.Dex));
    gtk_label_set_text(GTK_LABEL(wisval), print_stat(obj->stats.Wis));
    gtk_label_set_text(GTK_LABEL(intval), print_stat(obj->stats.Int));
    gtk_label_set_text(GTK_LABEL(perval), print_stat(obj->stats.Per));
    gtk_label_set_text(GTK_LABEL(conval), print_stat(obj->stats.Con));
    sprintf(buf, "%d", obj->stats.speed);
    gtk_label_set_text(GTK_LABEL(Speedval), buf);
    sprintf(buf, "%d", obj->stats.exp);
    gtk_label_set_text(GTK_LABEL(EXPval), buf);
    sprintf(buf, "%d", obj->stats.appr);
    gtk_label_set_text(GTK_LABEL(Apprval), buf);
    sprintf(buf, "%d", obj->stats.ac);
    gtk_label_set_text(GTK_LABEL(ACval), buf);
    sprintf(buf, "%d", obj->stats.fame);
    gtk_label_set_text(GTK_LABEL(Fameval), buf);
}

/**
   \brief scroll the messages window to the bottom
   \param data unused
   \return true/false (allways false)
*/

gboolean status_page_scroll_to_end(gpointer data)
{
    GtkTextIter iter;
    GtkTextBuffer *buffer;

    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(messages));
    gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(buffer), &iter);
    gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(messages), &iter, 0.0,
				 TRUE, 0.0, 1.0);
    return(FALSE);
}

/**
   \brief Display a message to the client, in the messages window
   \param buf buffer to display
*/

void display_message(char *buf)
{
    GtkTextIter iter;
    GtkTextBuffer *buffer;

    if (buf == NULL)
	return;

    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(messages));
    gtk_text_buffer_get_end_iter(buffer, &iter);
    gtk_text_buffer_insert(buffer, &iter, buf, -1);
    g_idle_add((GSourceFunc)status_page_scroll_to_end, NULL);
}
