/* $Id: inventory.c,v 1.12 2004/04/01 02:53:49 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <rm/rm.h>

#include "cnet.h"
#include "shstr.h"
#include "defines.h"
#include "cobject.h"
#include "ui.h"
#include "utils.h"
#include "inventory.h"

GtkWidget *Inventory;
GtkTreeIter b_spell, b_job, b_inv, b_skill, b_key, b_equip;
GtkTreeStore *invstore;
GtkTreeSelection *iselection;

extern client_t *pickedobj;
/*extern int pickedobjnum;*/

gboolean view_selection_func(GtkTreeSelection *selection, GtkTreeModel *model,
			      GtkTreePath *path, gboolean pcurr,
			      gpointer userdata)
{
    GtkTreeIter iter;
    client_t *obj;

    if (gtk_tree_model_get_iter(model, &iter, path))
    {
	gchar *name;

	gtk_tree_model_get(model, &iter, COL_OBJ_NAME, &name,
			   COL_OBJ_OBJ, &obj, -1); 
	if (obj != pickedobj && obj != NULL) {
	    if (!pcurr) {
		g_print ("%s is going to be selected.\n", name);
		g_free(name);
		set_target(obj);
	    }
	}
    }

    return TRUE; /* allow selection state to change */
}


void create_buckets(GtkTreeIter *subtree, char *label)
{
    gtk_tree_store_append(invstore, subtree, NULL);
    gtk_tree_store_set(invstore, subtree, COL_OBJ_NAME, label,
		       COL_OBJ_OBJ, NULL, -1);
}

void add_obj_to_bucket(client_t *obj, int bucket)
{
    GtkTreeIter *iter;

    switch(bucket) {
    case BUCKET_SPELL:
	iter = &b_spell;
	break;
    case BUCKET_JOB:
	iter = &b_job;
	break;
    case BUCKET_INV:
	iter = &b_inv;
	break;
    case BUCKET_SKILL:
	iter = &b_skill;
	break;
    case BUCKET_KEY:
	iter = &b_key;
	break;
    case BUCKET_EQUIP:
	iter = &b_equip;
	break;
    default:
	return;
    }
    gtk_tree_store_append(invstore, &obj->iter, iter);
    gtk_tree_store_set(invstore, &obj->iter, COL_OBJ_NAME, obj->name,
		       COL_OBJ_OBJ, obj, -1); 
}

/* this does not properly handle "bags" */

void remove_object_from_inv(client_t *obj)
{
    gtk_tree_store_remove(invstore, &obj->iter);
}

void command_no_target(GtkButton *button, gpointer user_data)
{
    const char *label;
    char buf[256];

    label = gtk_button_get_label(button);
    if (label == NULL)
	return;

    /* the server needs only the command, in this case */
 
    sprintf(buf, "%s\r", label);
    do_write(buf);
}

void command_optional_target(GtkButton *button, gpointer user_data)
{
    const char *label;
    char buf[256],b2[40];

    label = gtk_button_get_label(button);
    if (label == NULL)
	return;

    /* the user has nothing selected */
    if (pickedobj == NULL)/* || pickedobjnum == -1) */
	sprintf(buf, "%s\r", label);
    else
	sprintf(buf, "%s tar:%s\r", label, uid_sprint(b2, &pickedobj->uid));
    do_write(buf);
}

void command_single_target(GtkButton *button, gpointer user_data)
{
    const char *label;
    char buf[256],b2[40];

    label = gtk_button_get_label(button);
    if (label == NULL)
	return;

    /* the user has nothing selected */
    if (pickedobj == NULL)/* || pickedobjnum == -1)*/
	return;

    /* the server wants a command, followed by a target ID.  Lets give it
       one */
    sprintf(buf, "%s tar:%s\r", label, uid_sprint(b2, &pickedobj->uid));
    do_write(buf);
}
