/* $Id: mappings.c,v 1.16 2004/04/01 07:45:41 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <inttypes.h>
#include <sys/time.h>

#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "timers.h"
#include "pngloader.h"
#include "jpegloader.h"
#include "mappings.h"
#include "defines.h"
#include "utils.h"
#include "shstr.h"
#include "textures.h"
#include "settings.h"
#include "weatherdefs.h"
#include "cobject.h"

/*
    781
    602
    543
*/

cmap_t mapcache[9];
tmap_t *txtimage;
skymap_t *skies;
int textures_loaded;
RMcolor4D curfogcolor = {0.07, 0.88, 0.81, 1.0};

extern RMpipe *mainpipe;
extern int numgtextures;
extern csettings_t settings;
extern servdefs_t serverdefs;
extern RMnode *groundnode;
extern RMnode *bgnode;

/**
   \brief initialize everything we need for the ground
*/

void init_ground(void)
{
    /* zero out the mapcaches */
    memset(&mapcache, 0, sizeof(cmap_t) * 9);
    textures_loaded = 0;
}

int rmPrimitiveFrustumCullCallback(const RMprimitive *p, const RMstate *s)
{
    int i;
    float v[4],d[4];
    RMvertex3D bmin, bmax;
    RMvertex3D t[8];
    int isVisible;
    const RMmatrix *c;

    /* get the bounding box. */
    if (rmPrimitiveGetBoundingBox(p, &bmin, &bmax) == RM_WHACKED)
        return 1;               /* not sure what to do - just draw it. */

    /* from the two corners of the source box, construct the other
     6 of the fully-specified box. we assume the original box is an
     axis aligned thing. */
    t[0] = bmin;
    t[1] = t[0];
    t[1].x = bmax.x;
    t[2] = t[1];
    t[2].y = bmax.y;
    t[3] = t[2];
    t[3].x = bmin.x;

    t[4] = bmin;
    t[4].z = bmax.z;
    t[5] = t[4];
    t[5].x = bmax.x;
    t[6] = t[5];
    t[6].y = bmax.y;
    t[7] = t[6];
    t[7].x = bmin.x;

    isVisible = 0;
    c = &(s->composite) ;       

    for (i=0;i<8;i++)
    {
        int mask = (1 << i);
        
        v[0] = t[i].x;
        v[1] = t[i].y;
        v[2] = t[i].z;
        v[3] = 1.F;

        /* transform a box vertex through the current
         model+view+projection matrix. we include the project matrix
         to simplify comparisons in Z. t*/
        rmPoint4MatrixTransform(v, c, d);

        /* this test is valid for perspective projections. need to
         test with orthographic projections. */
        if ((d[2] > 0.0) && ( d[2] <= d[3]) && (fabs(d[0]) <= d[3]) && (fabs(d[1
]) <= d[3]))
            isVisible |= mask;

        if (isVisible != 0)     /* one of bbox corners is visible */
            break;
    }

    if (isVisible == 0)
        return(0);
    else
        return(1);
}

int GenNodeFrustumCullCallback(const RMnode *n, const RMstate *s)
{
    int i, nrofp;
    int isVisible;
    RMprimitive *p;

    isVisible = rmNodeFrustumCullCallback(n, s);
    if (isVisible)
	return(1);
    nrofp = rmNodeGetNumPrims(n);
    for (i=0; i < nrofp; i++) {
	p = rmNodeGetPrimitive(n, i);
	isVisible = rmPrimitiveFrustumCullCallback(p, s);
	if (isVisible)
	    return(1);
    }
    return(0);
}

static void load_sky_cond(skycond_t *cond)
{
    uint8_t *pixels=NULL;
    char file[PATH_MAX];

    if (cond->path == NULL)
	return;
    if (cond->image != NULL)
	return;

    sprintf(file, "%s/%s", settings.texdir, cond->path);
    if (cond->tformat == TFORM_JPEG) {
	pixels = readJPEG(file, &cond->width, &cond->height);
	if (pixels == NULL)
	    bailout();
	cond->image = rmImageNew(2, cond->width, cond->height, 1,
				 RM_IMAGE_RGB, RM_UNSIGNED_BYTE, RM_COPY_DATA);
	rmImageSetPixelData(cond->image, pixels, RM_COPY_DATA, NULL);
	free(pixels);
    }
    if (cond->tformat == TFORM_PNG) {
	pixels = parse_pngRGBA(file, &cond->width, &cond->height);
	if (pixels == NULL)
	    bailout();
	cond->image = rmImageNew(2, cond->width, cond->height, 1,
			        RM_IMAGE_RGBA, RM_UNSIGNED_BYTE, RM_COPY_DATA);
	rmImageSetPixelData(cond->image, pixels, RM_COPY_DATA, NULL);
	free(pixels);
    }
}

/**
   \brief set the sky
   \param curhour the current hour
   \param cloudlevel the current sky condition
   \todo eventually the server needs to tell us the time of dawn and dusk.
   for now we hardcode it.
*/

void set_sky_condition(int curhour, int cloudlevel)
{
    client_t *obj;

    obj = find_obj_by_id(&player_id);

    if (cloudlevel >= SKY_NROF)
	return;

    while (curhour >= serverdefs.hoursday)
	curhour -= serverdefs.hoursday;

    rmNodeSetSceneBackgroundImage(bgnode, NULL);
    if (curhour < 7 || curhour > 21) { /* night */
	rmNodeSetSceneBackgroundColor(bgnode, &skies[cloudlevel].night->sky);
	if (skies[cloudlevel].night->image != NULL)
	    rmNodeSetSceneBackgroundImage(bgnode,
					  skies[cloudlevel].night->image);
	curfogcolor = skies[cloudlevel].night->fog;
    }
    if (curhour >= 7 && curhour <= 21) { /* day */
	rmNodeSetSceneBackgroundColor(bgnode, &skies[cloudlevel].day->sky);
	if (skies[cloudlevel].day->image != NULL)
	    rmNodeSetSceneBackgroundImage(bgnode,skies[cloudlevel].day->image);
	curfogcolor = skies[cloudlevel].day->fog;
    }

    compute_fog(obj);
}

/**
   \brief load all the sky image files
*/

void load_sky_textures(void)
{
    int j;

    skies = parse_skymaps("skymaps.xml", settings.libdir);

/*    for (j=0; j < SKY_NROF; j++) {
	if (skies[j].day != NULL)
	    load_sky_cond(skies[j].day);
	if (skies[j].night != NULL)
	    load_sky_cond(skies[j].night);
	if (skies[j].dusk != NULL)
	    load_sky_cond(skies[j].dusk);
	if (skies[j].dawn != NULL)
	    load_sky_cond(skies[j].dawn);
	    }*/
}

void loadFloorTextures(void)
{
    RMimage *foo;
    int j;
    uint8_t *pixels;
    char file[PATH_MAX];

    if (textures_loaded)
	return;

    txtimage = parsegtextures("gtextures.xml", settings.libdir);

    for (j=0; j <= numgtextures; j++) {
	sprintf(file, "%s/%s", settings.texdir, txtimage[j].path);
	pixels = readJPEG(file, &txtimage[j].width, &txtimage[j].height);
	if (pixels == NULL)
	    bailout();
	foo = rmImageNew(2, txtimage[j].width, txtimage[j].height,
			 1, RM_IMAGE_RGB, RM_UNSIGNED_BYTE, RM_COPY_DATA);
	rmImageSetPixelData(foo, pixels, RM_COPY_DATA, NULL);
	txtimage[j].texture = rmTextureNew(2);
	rmTextureSetImages(txtimage[j].texture, &foo, 1, RM_FALSE);
	rmImageDelete(foo);
	free(pixels);
	/*rmTextureSetFilterMode(txtimage[0].texture, GL_LINEAR_MIPMAP_LINEAR, GL_NEAREST);*/
    }
    textures_loaded = 1;
}

/* this will cleanse out a previously used floor section for re-use */

void cleanFloor(RMnode *addto, int m)
{
    int i, j, num;
    RMnode *n;

    if (mapcache[m].mapnode != NULL && mapcache[m].tnodes != NULL) {
	for (i=0; i <= numgtextures; i++) {
	    num = rmNodeGetNumChildren(mapcache[m].tnodes[i]);
	    for (j=0; j < num; j++) {
		n = rmNodeGetIthChild(mapcache[m].tnodes[i], 0);
		rmNodeRemoveChild(mapcache[m].tnodes[i], n);
		rmSubTreeDelete(n);
	    }
	    rmNodeRemoveChild(mapcache[m].mapnode, mapcache[m].tnodes[i]);
	    rmSubTreeDelete(mapcache[m].tnodes[i]);
	}
	free(mapcache[m].tnodes);
	rmNodeRemoveChild(addto, mapcache[m].mapnode);
	rmSubTreeDelete(mapcache[m].mapnode);
    }
    if (mapcache[m].elevation != NULL && mapcache[m].terrain != NULL) {
	for (i=0; i < mapcache[m].height; i++) {
	    free(mapcache[m].elevation[i]);
	    free(mapcache[m].terrain[i]);
	}
	free(mapcache[m].elevation);
	free(mapcache[m].terrain);
    }
    memset(&mapcache[m], 0, sizeof(cmap_t));
}

void addFloor(RMnode *addto, int m)
{
    int x, z, y, i, j, n;
    RMvertex3D v[4];
    RMvertex2D t[4];
    RMvertex3D nor[4];
    float bx, by, bz;

    loadFloorTextures();

    if (mapcache[m].elevation == NULL || mapcache[m].terrain == NULL)
	return;

    bx = mapcache[m].basex;
    by = mapcache[m].basey;
    bz = mapcache[m].basez;

    mapcache[m].mapnode = rmNodeNew("ground", RM_RENDERPASS_3D,
				    RM_RENDERPASS_OPAQUE);

    mapcache[m].tnodes = malloc(sizeof(RMnode *)*numgtextures+1);
    for (i=0; i <= numgtextures; i++) {
	mapcache[m].tnodes[i] = rmNodeNew(txtimage[i].name, RM_RENDERPASS_3D,
					  RM_RENDERPASS_OPAQUE);
	rmNodeSetPolygonCullMode(mapcache[m].tnodes[i], RM_CULL_FRONT);
    }

    /* each quad should be a 16x16 square of primitives of same texture */

    for (n=0; n <= numgtextures; n++) {
	for (i=0; i < mapcache[m].height-1; i+=16) {
	    for (j=0; j < mapcache[m].width-1; j+=16) {
		RMnode *quad;
		char buf[30];
		
		sprintf(buf, "ground-%d-%d-%d", n, i, j);
		quad = rmNodeNew(buf, RM_RENDERPASS_3D, RM_RENDERPASS_OPAQUE);
		for (y=0,x=i; x < i+16 && x < mapcache[m].height-1; x++) {
		    for (z=j; z < j+16 && z < mapcache[m].width-1; z++) {
			RMprimitive *p;

			if (mapcache[m].terrain[x][z] == n) {
			    p = rmPrimitiveNew(RM_TRIANGLE_STRIP);
			    if (p == NULL)
				bailout();

			    nor[0].z = v[0].z = (z+1)*SCALE_XZ+bz;
			    nor[0].x = v[0].x = x*SCALE_XZ+bx;
			    nor[0].y = v[0].y = by +
				(float)mapcache[m].elevation[x][z+1]*SCALE_Y;
			    nor[1].z = v[1].z = z*SCALE_XZ + bz;
			    nor[1].x = v[1].x = x*SCALE_XZ + bx;
			    nor[1].y = v[1].y = by +
				(float)mapcache[m].elevation[x][z]*SCALE_Y;
			    nor[2].z = v[2].z = (z+1) * SCALE_XZ + bz;
			    nor[2].x = v[2].x = (x+1) * SCALE_XZ + bx;
			    nor[2].y = v[2].y = by +
				(float)mapcache[m].elevation[x+1][z+1]*SCALE_Y;
			    nor[3].z = v[3].z = z*SCALE_XZ + bz;
			    nor[3].x = v[3].x = v[2].x;
			    nor[3].y = v[3].y = by +
				(float)mapcache[m].elevation[x+1][z]*SCALE_Y;

			    t[0].x = 0.0;
			    t[0].y = 1.0;
			    t[1].x = 0.0;
			    t[1].y = 0.0;
			    t[2].x = 1.0;
			    t[2].y = 1.0;
			    t[3].x = 1.0;
			    t[3].y = 0.0;

			    rmVertex3DNormalize(&nor[0]);
			    rmVertex3DNormalize(&nor[1]);
			    rmVertex3DNormalize(&nor[2]);
			    rmVertex3DNormalize(&nor[3]);

			    rmPrimitiveSetVertex3D(p, 4, v, RM_COPY_DATA,
						   NULL);
			    rmPrimitiveSetTexcoord2D(p, 4, t, RM_COPY_DATA,
						     NULL);
			    rmPrimitiveSetNormal3D(p, 4, nor, RM_COPY_DATA, NULL);
			    rmPrimitiveComputeBoundingBox(p);
			    rmNodeAddPrimitive(quad, p);
			}
		    }
		}
		if (rmNodeGetNumPrims(quad) == 0)
		    rmNodeDelete(quad);
		else {
		    rmNodeComputeBoundingBox(quad);
		    rmNodeSetPreTraversalCallback(quad, RM_VIEW,
		      GenNodeFrustumCullCallback);
		    rmNodeAddChild(mapcache[m].tnodes[n], quad);
		}
	    }
	}
    }
    for (i = 0; i <= numgtextures; i++) {
	/* rmNodeSetSceneTexture(mapcache[m].tnodes[i], txtimage[i].texture, RM_DONT_COPY_DATA); */
	rmNodeSetSceneTexture(mapcache[m].tnodes[i], txtimage[i].texture);
	rmNodeAddChild(mapcache[m].mapnode, mapcache[m].tnodes[i]);
    }
    rmNodeAddChild(addto, mapcache[m].mapnode);
}

/**
   \brief move a map from one slot to another
   \param from slot to move from
   \param to slot to move to
*/

void move_map(int from, int to)
{
    memcpy(&mapcache[to], &mapcache[from], sizeof(cmap_t));
    memset(&mapcache[from], 0, sizeof(cmap_t));
}

/**
   \brief shift the known maps by a given direction.  Clear all other maps
   \param dir direction to shift maps in
   \warning map "dir" must be valid
*/

void shift_maps(int dir)
{
    int vmap1, vmap2, vmap3, i;

    vmap1 = vmap2 = vmap3 = 0;

    if (!dir)
        return;

    /* we were called with no map in the direction to switch to */
    if (mapcache[dir].mapnode == NULL)
	return;

    if (dir+4 > 8) {
	move_map(0, dir-4);
	vmap1 = dir-4;
    } else {
	move_map(0, dir+4);
	vmap1 = dir+4;
    }

    if (mapcache[dir].mapnode != NULL)
	move_map(dir, 0);

    if (dir == 1) {
        if (mapcache[8].mapnode != NULL) {
	    move_map(8, 6);
	    vmap2 = 6;
	}
        if (mapcache[2].mapnode != NULL) {
	    move_map(2, 4);
	    vmap3 = 4;
	}
    } else if (dir == 2) {
        if (mapcache[1].mapnode != NULL) {
	    move_map(1, 8);
	    vmap2 = 8;
	}
        if (mapcache[3].mapnode != NULL) {
	    move_map(3, 4);
	    vmap3 = 4;
	}
    } else if (dir == 3) {
        if (mapcache[2].mapnode != NULL) {
	    move_map(2, 8);
	    vmap2 = 8;
	}
        if (mapcache[4].mapnode != NULL) {
	    move_map(4, 6);
	    vmap3 = 6;
	}
    } else if (dir == 4) {
        if (mapcache[3].mapnode != NULL) {
	    move_map(3, 2);
	    vmap2 = 2;
	}
        if (mapcache[5].mapnode != NULL) {
	    move_map(5, 6);
	    vmap3 = 6;
	}
    } else if (dir == 5) {
        if (mapcache[4].mapnode != NULL) {
	    move_map(4, 2);
	    vmap2 = 2;
	}
        if (mapcache[6].mapnode != NULL) {
	    move_map(6, 8);
	    vmap3 = 8;
	}
    } else if (dir == 6) {
        if (mapcache[7].mapnode != NULL) {
	    move_map(7, 8);
	    vmap2 = 8;
	}
        if (mapcache[5].mapnode != NULL) {
	    move_map(5, 4);
	    vmap3 = 4;
	}
    } else if (dir == 7) {
        if (mapcache[8].mapnode != NULL) {
	    move_map(8, 2);
	    vmap2 = 2;
	}
        if (mapcache[6].mapnode != NULL) {
	    move_map(6, 4);
	    vmap3 = 4;
	}
    } else if (dir == 8) {
        if (mapcache[1].mapnode != NULL) {
	    move_map(1, 2);
	    vmap2 = 2;
	}
        if (mapcache[7].mapnode != NULL) {
	    move_map(7, 6);
	    vmap3 = 6;
	}
    }

    /* clean out all other maps */
    for (i=1; i<9; i++) {
	if (i == vmap1 || i == vmap2 || i == vmap3)
	    continue;
	cleanFloor(groundnode, i);
    }
}
