/* $Id: modelloader.c,v 1.19 2004/04/01 23:23:50 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <sys/param.h>
#include <sys/time.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <rm/rmaux.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <gtk/gtk.h>

#include "defines.h"
#include "timers.h"
#include "mappings.h"
#include "cmotion.h"
#include "shstr.h"
#include "jpegloader.h"
#include "pngloader.h"
#include "utils.h"
#include "xmlparse.h"
#include "cobject.h"
#include "modelloader.h"
#include "settings.h"

TAILQ_HEAD(, _genmaterial_t) genmats = TAILQ_HEAD_INITIALIZER(genmats);
TAILQ_HEAD(, _genmodel_t) genmodels = TAILQ_HEAD_INITIALIZER(genmodels);

extern RMpipe *mainpipe;
extern csettings_t settings;

typedef struct _tmpmaterial_t {
    char *name;
    char *image;
    RMcolor4D ambient, diffuse, specular;
    RMtexture *tex;
    int alpha;
} tmpmaterial_t;

void parse_quadfloat(xmlDocPtr doc, xmlNodePtr cur, RMcolor4D *quad)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    sscanf(tmp, "%f %f %f %f", &quad->r, &quad->g, &quad->b, &quad->a);
    free(tmp);
}

tmpmaterial_t *parse_materials(xmlDocPtr doc, xmlNodePtr cur, int *num)
{
    tmpmaterial_t *mats;
    int nrofmats = 0;
    xmlNodePtr ncur;
    genmaterial_t *gmat;
    RMimage *tmpimg, *image;
    RMenum format, type;
    int nh, nw;
    uint8_t *pixels;
    char file[MAX_PATH];

    cur = cur->xmlChildrenNode;
    mats = smalloc(tmpmaterial_t);
    pixels = NULL;
    format = 0;

    while (cur != NULL) {
        if (!strcmp(cur->name, "material")) {
	    nrofmats++;
	    mats = realloc(mats, sizeof(tmpmaterial_t)*nrofmats);
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "ambient"))
		    parse_quadfloat(doc, ncur, &mats[nrofmats-1].ambient);
		if (!strcmp(ncur->name, "diffuse"))
		    parse_quadfloat(doc, ncur, &mats[nrofmats-1].diffuse);
		if (!strcmp(ncur->name, "specular")) {
		    parse_quadfloat(doc, ncur, &mats[nrofmats-1].specular);
		}
		if (!strcmp(ncur->name, "name")) {
		    mats[nrofmats-1].name =
			xmlNodeListGetString(doc, ncur->xmlChildrenNode, 1);
		}
		if (!strcmp(ncur->name, "image")) {
		    int found=0;

		    mats[nrofmats-1].image =
			xmlNodeListGetString(doc, ncur->xmlChildrenNode, 1);
		    /* look for existing texture */
		    TAILQ_FOREACH(gmat, &genmats, next) {
			if (strcmp(mats[nrofmats-1].name, gmat->name) == 0) {
			    mats[nrofmats-1].tex = gmat->texture;
			    mats[nrofmats-1].alpha = gmat->alpha;
			    found++;
			    break;
			}
		    }
		    if (found) {
			ncur = ncur->next;
			continue;
		    }
		    /* allocate a new texture */
		    gmat = smalloc(genmaterial_t);
		    gmat->name = add_string(mats[nrofmats-1].name);
		    gmat->imagefile = add_string(mats[nrofmats-1].image);
		    gmat->alpha = 0;
		    sprintf(file, "%s/%s", settings.texdir, gmat->imagefile); 

		    if (strncmp(strrchr(gmat->imagefile, '.'),
				".jpg", 4) == 0) {
			pixels = readJPEG(file, &gmat->width, &gmat->height);
			format = RM_IMAGE_RGB;
		    } else if (strncmp(strrchr(gmat->imagefile, '.'),
				       ".png", 4) == 0) {
			pixels = parse_pngRGBA(file, &gmat->width, &gmat->height);
			format = RM_IMAGE_RGBA;
			gmat->alpha = 1;
		    }
		    mats[nrofmats-1].alpha = gmat->alpha;
		    type = RM_UNSIGNED_BYTE;
		    tmpimg = rmImageNew(2, gmat->width, gmat->height,
					1, format, type, RM_COPY_DATA);
		    rmImageSetPixelData(tmpimg, pixels,	RM_COPY_DATA, NULL);
		    nw = rmNearestPowerOfTwo(gmat->width);
		    nh = rmNearestPowerOfTwo(gmat->height);
		    if (nw != gmat->width || nh != gmat->height) {
                        gmat->width = nw;
                        gmat->height = nh;
                        image = rmImageNew(2, gmat->width, gmat->height,
                            1, format, type, RM_COPY_DATA);
                        rmImageResize(tmpimg, image, RM_SOFTWARE, mainpipe);
			rmImageDelete(tmpimg);
                    } else {
                        image = tmpimg;
                    }
		    gmat->texture = rmTextureNew(2);
		    rmTextureSetImages(gmat->texture, &image, 1,
				       RM_FALSE);
		    free(pixels);
		    rmImageDelete(image);
		    rmTextureSetEnv(gmat->texture, GL_REPLACE, NULL);
		    rmTextureSetWrapMode(gmat->texture, GL_REPEAT);
		    TAILQ_INSERT_TAIL(&genmats, gmat, next);
		    mats[nrofmats-1].tex = gmat->texture;
		}
		ncur = ncur->next;
	    }
	}
	cur = cur->next;
    }
    num = &nrofmats;
    return(mats);
}

/* the primitive data is a shared string */
/*ARGSUSED*/
void freeprimdata(RMprimitive *p, void *ptr)
{
    free_string((char *)ptr);
}

static int same_vert(RMvertex3D *v1, RMvertex3D *v2)
{
    if (v1->x == v2->x && v1->y == v2->y && v1->z == v2->z)
	return(1);
    return(0);
}

static int same_triangle(RMvertex3D *t1, RMvertex3D *t2, RMvertex3D *t3,
			 RMvertex3D *v1, RMvertex3D *v2, RMvertex3D *v3)
{
    int skip=0;

    if (same_vert(t1, v1)) {
	if (same_vert(t2, v2)) {
	    if (same_vert(t3, v3)) {
		skip++;
	    }
	} else if (same_vert(t2, v3)) {
	    if (same_vert(t3, v2)) {
		skip++;
	    }
	}
    } else if (same_vert(t1, v2)) {
	if (same_vert(t2, v1)) {
	    if (same_vert(t3, v3)) {
		skip++;
	    }
	} else if (same_vert(t2, v3)) {
	    if (same_vert(t3, v1)) {
		skip++;
	    }
	}
    } else if (same_vert(t1, v3)) {
	if (same_vert(t2, v1)) {
	    if (same_vert(t3, v2)) {
		skip++;
	    }
	} else if (same_vert(t2, v2)) {
	    if (same_vert(t3, v1)) {
		skip++;
	    }
	}
    }
    return(skip);
}

void parse_genmodel(RMnode *node, char *filename, client_t *obj)
{
    xmlDocPtr doc;
    xmlNodePtr cur, ncur, xcur, ycur;
    RMprimitive *p=NULL;
    RMnode *n=NULL;
    RMvertex3D *v=NULL, *fv=NULL, *fn=NULL, *nor=NULL;
    RMvertex2D *t=NULL, *ft=NULL;
    int *collid=NULL;
    tmpmaterial_t *mats=NULL;
    int nrofmats, nrofpoints, *gnor=NULL;
    char *tmp;
    genmodel_t *gmod;

    doc = xmlParseFile(filename);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "genmodel")) {
        fprintf(stderr, 
                "document of the wrong type, root node != genmodel");
        xmlFreeDoc(doc);
        bailout();
    }

    TAILQ_FOREACH(gmod, &genmodels, next) {
	if (strcmp(gmod->model, filename) == 0) {
	    gmod->refcount++;
	    rmNodeAddChild(node, gmod->node);
	    rmNodeSetName(node, gmod->name);
	    obj->gmod = gmod;
	    return;
	}
    }
    /* otherwise, we create a new gmodel and DTRT */
    gmod = smalloc(genmodel_t);
    if (gmod == NULL)
	bailout();
    gmod->model = add_string(filename);
    gmod->refcount = 1;

    cur = cur->xmlChildrenNode;

    while (cur != NULL) {
	if (!strcmp(cur->name, "name")) {
	    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    rmNodeSetName(node, tmp);
	    gmod->node = rmNodeNew(tmp, RM_RENDERPASS_3D,
				   RM_RENDERPASS_ALL);
	    gmod->name = add_string(tmp);
	    free(tmp);
	}
        if (!strcmp(cur->name, "materials")) {
	    mats = parse_materials(doc, cur, &nrofmats);
	}
	if (!strcmp(cur->name, "surface")) {
	    int optmesh=0;
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "vertexs")) {
		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL) {
			int nrofverts;
			static int cvert;

			if (!strcmp(xcur->name, "verts")) {
			    parseint(doc, xcur, &nrofverts);
			    v = safer_malloc(sizeof(RMvertex3D) * nrofverts);
			    collid = safer_malloc(sizeof(int) * nrofverts);
			    /*collnum = malloc(sizeof(int) * nrofverts);*/
			    t = safer_malloc(sizeof(RMvertex2D) * nrofverts);
			    nor = safer_malloc(sizeof(RMvertex3D) * nrofverts);
			    gnor = safer_malloc(sizeof(int) * nrofverts);
			    cvert = -1;
			}
			if (!strcmp(xcur->name, "vertex")) {
			    cvert++;
			    collid[cvert] = cvert;
			    /*collnum[cvert] = 0;*/
			    ycur = xcur->xmlChildrenNode;
			    while (ycur != NULL) {
				if (!strcmp(ycur->name, "vertex_x"))
				    parsefloat(doc, ycur, &v[cvert].x);
				if (!strcmp(ycur->name, "vertex_y"))
				    parsefloat(doc, ycur, &v[cvert].y);
				if (!strcmp(ycur->name, "vertex_z"))
				    parsefloat(doc, ycur, &v[cvert].z);
				if (!strcmp(ycur->name, "nor_x")) {
				    gnor[cvert]++;
				    parsefloat(doc, ycur, &nor[cvert].x);
				}
				if (!strcmp(ycur->name, "nor_y"))
				    parsefloat(doc, ycur, &nor[cvert].y);
				if (!strcmp(ycur->name, "nor_z"))
				    parsefloat(doc, ycur, &nor[cvert].z);
				if (!strcmp(ycur->name, "tex_u"))
				    parsefloat(doc, ycur, &t[cvert].x);
				if (!strcmp(ycur->name, "tex_v"))
				    parsefloat(doc, ycur, &t[cvert].y);
				if (!strcmp(ycur->name, "collid"))
				    parseint(doc, ycur, &collid[cvert]);
				/*if (!strcmp(ycur->name, "collc"))
				  parseint(doc, ycur, &collnum[cvert]);*/
				ycur = ycur->next;
			    }
			}
			xcur = xcur->next;
		    }
		}
		if (!strcmp(ncur->name, "triangle")) {
		    int type=0, cface=0, vnum, matnum, newnode=0;
		    int lodcount=0, tskip=0;
		    char *mname=NULL;

		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL && !optmesh) {
			if (!strcmp(xcur->name, "name")) {
			    mname = xmlNodeListGetString(doc,
			        xcur->xmlChildrenNode, 1);
			}
			if (!strcmp(xcur->name, "optmesh") && settings.detail > 5)
			    optmesh++;
			if (!strcmp(xcur->name, "lodcount"))
			    parseint(doc, xcur, &lodcount);
			if (!strcmp(xcur->name, "matnum")) {
			    parseint(doc, xcur, &matnum);
			    n = NULL;
			    if (rmNodeGetNumChildren(gmod->node) > matnum)
				n = rmNodeGetIthChild(gmod->node, matnum);
			    if (n == NULL) {
				if (mats[matnum].alpha)
				    n = rmNodeNew(mname, RM_RENDERPASS_3D,
					      RM_RENDERPASS_TRANSPARENT);
				else
				    n = rmNodeNew(mname, RM_RENDERPASS_3D,
					      RM_RENDERPASS_OPAQUE);
				newnode++;
			    }
			    free(mname);
			    if (newnode) {
				/*rmNodeSetSceneTexture(n, mats[matnum].tex,
				  RM_DONT_COPY_DATA);*/
				rmNodeSetSceneTexture(n, mats[matnum].tex);
				rmNodeSetAmbientColor(n,&mats[matnum].ambient);
				rmNodeSetDiffuseColor(n,&mats[matnum].diffuse);
				rmNodeSetSpecularColor(n, &mats[matnum].specular);
				/* removing this for now, it interferes with
				   the target window operation. XXX */
				/*rmNodeSetPreTraversalCallback(n,
				  RM_VIEW, rmNodeFrustumCullCallback);*/
			    }
			}
			if (!strcmp(xcur->name, "points")) {
			    parseint(doc, xcur, &nrofpoints);
			}
			if (!strcmp(xcur->name, "type")) {
			    tmp = xmlNodeListGetString(doc,
                                xcur->xmlChildrenNode, 1);
			    if (!strcmp(tmp, "strip")) {
				fv = safer_malloc(sizeof(RMvertex3D) * nrofpoints);
				fn = safer_malloc(sizeof(RMvertex3D) * nrofpoints);
				ft = safer_malloc(sizeof(RMvertex2D) * nrofpoints);
				p = rmPrimitiveNew(RM_TRIANGLE_STRIP);
				type = 1;
			    }
			    else if (!strcmp(tmp, "block")) {
				fv = safer_malloc(sizeof(RMvertex3D) * nrofpoints);
				fn = safer_malloc(sizeof(RMvertex3D) * nrofpoints);
				ft = safer_malloc(sizeof(RMvertex2D) * nrofpoints);
				p = rmPrimitiveNew(RM_TRIANGLES);
				type = 2;
			    }
			    rmPrimitiveSetClientData(p,
				add_string(gmod->model), freeprimdata);
			    free(tmp);
			}
			if (!strcmp(xcur->name, "face")) {
			    if (type == 1) {
				parseint(doc, xcur, &vnum);
				fv[cface] = v[vnum];
				fn[cface] = nor[vnum];
				if (gnor[vnum] == 0) {
				    fn[cface] = v[vnum];
				    (void)rmVertex3DNormalize(&fn[cface]);
				}
				ft[cface] = t[vnum];
				cface++;
			    } else if (type == 2) {
				int v1, v2, v3, skip=0, of, cdet;

				parsetriint(doc, xcur, &v1, &v2, &v3);
				/* cull degenerate triangles */
				if (v1 == v2 || v1 == v3 || v2 == v3)
				    skip++;
				if (settings.detail > 0 && lodcount > 0) {
				    if (rndm(0, 8) < settings.detail && !skip) {
					cdet = 0;
					while (v1 != collid[v1] &&
					       cdet < settings.detail &&
					       collid[collid[v1]] != v1) {
					    v1 = collid[v1];
					    cdet++;
					}
					cdet = 0;
					while (v2 != collid[v2] &&
					       cdet < settings.detail &&
					       collid[collid[v2]] != v2) {
					    v2 = collid[v2];
					    cdet++;
					}
					cdet = 0;
					while (v3 != collid[v3] &&
					       cdet < settings.detail &&
					       collid[collid[v3]] != v3) {
					    v3 = collid[v3];
					    cdet++;
					}
				    }
				    /* cull degenerate triangles */
				    if (v1 == v2 || v1 == v3 || v2 == v3)
					skip++;

				    if (settings.detail == 8)
					cdet = lodcount;
				    else if (settings.detail == 9)
					cdet = 9999999;
				    else {
					cdet = lodcount/10;
					cdet *= settings.detail;
				    }
				    /* look for triangles we allready have */
				    if (skip == 0 && tskip <= cdet &&
					rndm(0, 8) < settings.detail) {
					for (of=0; of < cface;) {
					    skip += same_triangle(&fv[of],
					        &fv[of+1], &fv[of+2], &v[v1],
					        &v[v2], &v[v3]);
					    of+=3;
					}
				    }
				}
				if (skip) {
				    tskip++;
				} else {
				    fv[cface] = v[v1];
				    fn[cface] = nor[v1];
				    if (gnor[v1] == 0) {
					fn[cface] = v[v1];
					(void)rmVertex3DNormalize(&fn[cface]);
				    }
				    ft[cface] = t[v1];
				    cface++;

				    fv[cface] = v[v2];
				    fn[cface] = nor[v2];
				    if (gnor[v2] == 0) {
					fn[cface] = v[v2];
					(void)rmVertex3DNormalize(&fn[cface]);
				    }
				    ft[cface] = t[v2];
				    cface++;

				    fv[cface] = v[v3];
				    fn[cface] = nor[v3];
				    if (gnor[v3] == 0) {
					fn[cface] = v[v3];
					(void)rmVertex3DNormalize(&fn[cface]);
				    }
				    ft[cface] = t[v3];
				    cface++;
				}
			    }
			}
			xcur = xcur->next;
		    } /* while triangle */
		    if (!optmesh) {
			/*printf("lodcount %d tskip %d\n", lodcount, tskip);*/
			rmPrimitiveSetVertex3D(p, cface, fv, RM_COPY_DATA,
					       NULL);
			rmPrimitiveSetNormal3D(p, cface, fn, RM_COPY_DATA,
					       NULL);
			rmPrimitiveSetTexcoord2D(p, cface, ft, RM_COPY_DATA,
						 NULL);
			rmPrimitiveComputeBoundingBox(p);
			rmNodeAddPrimitive(n, p);
			rmNodeComputeBoundingBox(n);
			if (newnode)
			    rmNodeAddChild(gmod->node, n);
		    } else {
			printf("skipped optional mesh\n");
		    }
		} /* triangle */
		ncur = ncur->next;
	    } /* while */
	    if (!optmesh) {
		free(fn);
		free(fv);
		free(ft);
	    }
	    free(v);
	    free(nor);
	    free(gnor);
	    free(t);
	    free(collid);
	} /* surface */
	cur = cur->next;
    }
    free(mats);
    rmNodeAddChild(node, gmod->node);
    obj->gmod = gmod;
    TAILQ_INSERT_TAIL(&genmodels, gmod, next);
    xmlFreeDoc(doc);
}
