/* $Id: net.c,v 1.28 2004/04/01 23:23:50 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 1998-2002 David Watson and Able Data Systems
    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <ctype.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <fcntl.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <rm/rm.h>

#include "defines.h"
#include "shstr.h"
#include "cnet.h"
#include "cobject.h"
#include "netparser.h"
#include "interface.h"
#include "parser.h"

typedef int (*parsefunc_t)(pargs_t *args, char *pos);

/** \brief parser command to function lookup table */
parsefunc_t parsefunc[] = {
    parse_command_null,
    parse_command_ur,
    parse_command_nuob,
    parse_command_bbox,
    parse_command_chob,
    parse_command_nomv,
    parse_command_inv,
    parse_command_rminv,
    parse_command_eq,
    parse_command_ueq,
    parse_command_rmob,
    parse_command_eact,
    parse_command_null,
    parse_command_null,
    parse_command_text,
    parse_command_msrvdone,
    parse_command_server,
    parse_command_mape,
    parse_command_mapt,
    parse_command_maps,
    parse_command_mapd,
    parse_command_srvfull,
    parse_command_srvd
};

extern uniqid_t player_id;

char	*in_buf;		/* Input buffer */
char	*out_buf;		/* Output buffer */
int mainsocket;	/* main socket */
int inbuflen;

int connect_to(char *host, int port)
{
    struct sockaddr_in sa;
    struct hostent *hp;
    int s, rbuf;
    char buf[256];

    /* Resolve the host */
    bzero(&sa, sizeof(sa));

    if ((hp = gethostbyname(host)))
	bcopy(hp->h_addr, &sa.sin_addr, hp->h_length);
    else if (isdigit(*host))
	sa.sin_addr.s_addr = htonl(inet_network(host));

    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);

    /* Create the socket */
    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	sprintf(buf, "Error while trying to connect to %s port %d:\n"
		"socket: %s\n", host, port, strerror(errno));
	display_message(buf);
	return(-1);
    }

    /* Make the connection */
    if (connect(s, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
	sprintf(buf, "Error while trying to connect to %s port %d:\n"
		"connect: %s\n", host, port, strerror(errno));
	display_message(buf);
	return(-1);
    }

    /* Set receive buffer to 8K */
    rbuf = 8192;

    if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *)&rbuf, sizeof(rbuf)) < 0){
	sprintf(buf, "Error while trying to connect to %s port %d:\n"
		"setsockopt(): %s\n", host, port, strerror(errno));
	display_message(buf);
	return(-1);
    }

#ifdef NONBLOCK
    /* Turn on No Delay (or non-blocking) for descriptor. If it is not
     * set, then the process will sleep while waiting for data on a
     * read() call. This causes the whole process to stop all operations
     * until it receives data. Depending on what your client does, you
     * may or may not want the descriptor to be non-delayed. If the
     * client is an interactive client where data may be sent and
     * received at the same time, you may want it non-delayed. If the
     * client is more of a step-by-step model, then turning off No Delay
     * may be more desireable.
     */
    if (fcntl(s, F_SETFL, O_NDELAY) < 0) {
	sprintf(buf, "Error while trying to connect to %s port %d:\n"
		"fcntl() non-block: %s\n", host, port, strerror(errno));
	display_message(buf);
	return(-1);
    }
#endif /* #ifdef NONBLOCK */

    return(s);
}

/* Read data from descriptor s and store it in in_buf */
int process_input(int s)
{
    char buf[8193], *p;
    int i;

    if ((i=read(s, buf, 8192)) < 1)
	return(-1);

    /*buf[i] = '\0';*/

    if (!in_buf) {
	in_buf = (char *)malloc(sizeof(char) * (i+1));
	memcpy(in_buf, buf, i);
	/*strcpy(in_buf, buf);*/
	inbuflen = i;
    } else {
	in_buf = (char *)realloc(in_buf, sizeof(char) * (inbuflen+i+2));
	p = in_buf;
	p += inbuflen;
	memcpy(p, buf, i);
	inbuflen += i;
	/*strcat(in_buf, buf);*/
    }
    return(i);
}

/*
 * You may want to use this function to store data in out_buf to be sent to
 * the server.
 */
void do_write(char *buf)
{
    if (!out_buf) {
	out_buf = (char *)malloc(strlen(buf)+1);
	strcpy(out_buf, buf);
    } else {
	out_buf = (char *)realloc(out_buf, (strlen(out_buf)+strlen(buf))+2);
	strcat(out_buf, buf);
    }
}

/* Send data that is stored in out_buf to server */
int process_output(int s)
{
    int i=0, buflen, cpos;
    char *p;

    if (out_buf) {
        p = out_buf;
        buflen = strlen(p);
        cpos = 0;
        while (cpos < buflen) {
            i = write(s, p, buflen-cpos);
            if (i < 0) {
		if (errno == EAGAIN)
		    return(0);
		else
		    return(-1);
            } else {
                cpos += i;
                p += i;
            }
        }
        free(out_buf);
        out_buf=0;
        return(cpos);
    } else
        return(0);
}

/* This is where the commands will be parsed */
void process_command()
{
    char **words, *buf, *p, *pp;
    int numwords, command, i, skip=0;
    pargs_t *args=NULL;

    /*printf("Client Got: \n::%s::\n :END\n", in_buf);*/

    p = in_buf;
    pp = NULL;
    for (buf=in_buf; buf != NULL && inbuflen > 0; ) {
	pp = p;
	p = memchr(buf, '\r',inbuflen);
	if (p == NULL)
	    goto next;
	*p = '\0';
	p++;
	inbuflen -= (strlen(buf)+1);
	/*printf("INBUFLEN = %d\n", inbuflen);*/
	words = parse_netcommand(buf, &numwords);
	if (words == NULL)
	    goto next;
	args = parse_command(words, &command, numwords);
	if (command < 1 || command > COMMAND_MAX || args == NULL)
	    goto next;
	skip = parsefunc[command](args, p);
	if (skip == -1) {
	    /* undo the \0 */
	    inbuflen += (strlen(buf)+1);
	    p--;
	    *p = '\r';
	    p++;
	    p = NULL;
	} else {
	    p += skip;
	    inbuflen -= skip;
	}
    next:
	if (args) {
	    for (i=0; args[i].cword != -1; i++)
		if (args[i].type == PTCHAR)
		    free_string(args[i].arg.c);
	    free(args);
	    args = NULL;
	}
	buf = p;
	skip = 0;
    }
    if (p == NULL && pp != NULL) {
	if (pp != in_buf)
	    memmove(in_buf, pp, inbuflen);
    } else {
	free(in_buf);
	in_buf=0;
	inbuflen=0;
    }
    if (skip == -1)
	(void)networker(NULL);
}

int connect_server(char *servername, int port)
{
    if (port == 0)
	port = DEFAULT_PORT;

    mainsocket = connect_to(servername, port);
    if (mainsocket < 0) {
	mainsocket = 0;
	return(0);
    }
    return(1);
}

/**
   \brief Main network handling function
   \param glarea the gtkgl widget
   \return TRUE
   \todo make the client recover from disconnect, and allow a new connection
   to be made
*/

gint networker(GtkWidget *glarea)
{
    fd_set input, output, errfds;
    struct timeval tval;
    char buf[256];

    FD_ZERO(&input);
    FD_ZERO(&output);
    FD_ZERO(&errfds);

#ifdef NONBLOCK
    /*
     * We will only be using input and output file descriptor sets
     * in a non-delayed type client.
     */
    FD_SET(mainsocket, &input);
    FD_SET(mainsocket, &output);
#endif /* #ifdef NONBLOCK */
    FD_SET(mainsocket, &errfds);

    tval.tv_sec = 0;
    tval.tv_usec = 1000;

    /* XXX these should be recoverable, no need to quit client. */
    if (select(mainsocket + 1, &input, &output, &errfds, &tval) < 0) {
	sprintf(buf, "Error while communicating with server:\n"
		"select(): %s\n", strerror(errno));
	fprintf(stderr, buf);
	display_message(buf);
	exit(1);
    }

    if (FD_ISSET(mainsocket, &errfds)) {
	sprintf(buf, "Error with connection, disconnecting.\n");
	display_message(buf);
	fprintf(stderr, buf);
	close(mainsocket);
	exit(-1);
    }

#ifdef NONBLOCK
    /*
     * We only check if FD_ISSET on a non-delayed type client. You
     * may need to switch the input and output around if the server
     * expects something from the client before it sends anything,
     * rather than sending something before it waits for input.
     */
    if (FD_ISSET(mainsocket, &input))
#endif /* #ifdef NONBLOCK */
	if (process_input(mainsocket) < 0) {
	    sprintf(buf, "Lost connection with server.\n");
	    display_message(buf);
	    fprintf(stderr, buf);
	    close(mainsocket);
	    exit(1);
	}

#ifdef NONBLOCK
    if (FD_ISSET(mainsocket, &output))
#endif /* #ifdef NONBLOCK */
	if (process_output(mainsocket) < 0) {
	    sprintf(buf, "Lost connection with server.\n");
	    display_message(buf);
	    fprintf(stderr, buf);
	    close(mainsocket);
	    exit(1);
	}

    if (in_buf)
	process_command();
    return(TRUE);
}
