/* $Id: parser.c,v 1.24 2004/04/01 23:23:50 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <sys/param.h>
#include <sys/time.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
/*#define bool char*/
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <rm/rmaux.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <GL/gl.h>
#include <GL/glu.h>
#undef False
#undef True
#define bool enum Boolean
#include <cal3d/cal3d_wrapper.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "defines.h"
#include "timers.h"
#include "mappings.h"
#include "cmotion.h"
#include "shstr.h"
#include "jpegloader.h"
#include "pngloader.h"
#include "utils.h"
#include "xmlparse.h"
#include "cobject.h"
#include "modelloader.h"
#include "skelanim.h"
#include "camera.h"
#include "cnet.h"
#include "interface.h"
#include "inventory.h"
#include "settings.h"
#include "stattables.h"
#include "netparser.h"
#include "parser.h"
#include "metaserv.h"

/**
   \file parser.c
   \brief Client network parsing functions
   \author Tim Rightnour
*/

extern RMnode *mainchar;
extern RMnode *rootnode;
extern RMnode *transnode;
extern RMnode *objnode;
extern RMnode *tobjnode;
extern RMnode *groundnode;
extern RMpipe *mainpipe;
extern int needs_redraw;
extern csettings_t settings;
extern servdefs_t serverdefs;
extern uniqid_t player_id;
extern TAILQ_HEAD(, _client_t) objects;
extern GtkWidget *messages;
extern int mainsocket;
extern int connected;
extern TAILQ_HEAD(, _mservdata) mservdata;
extern int got_mservdata;
extern int inbuflen;
extern cmap_t mapcache[9];

/** \brief the command table for the client parser */

argtable_t commtable[] = {
    {"ur", COMMAND_UR, 0},		/* server */
    {"nuob", COMMAND_NUOB, 0},		/* server */
    {"bbox", COMMAND_BBOX, 0},		/* server */
    {"chob", COMMAND_CHOB, 0},		/* server */
    {"nomv", COMMAND_NOMV, 0},		/* server */
    {"inv", COMMAND_INV, 0},		/* server */
    {"rminv", COMMAND_RMINV, 0},	/* server */
    {"eq", COMMAND_EQ, 0},		/* server */
    {"ueq", COMMAND_UEQ, 0},		/* server */
    {"rmob", COMMAND_RMOB, 0},		/* server */
    {"eact", COMMAND_EACT, 0},		/* server */
    {"ready", COMMAND_READY, 0},	/* server */
    {"noadpl", COMMAND_NOADPL, 0},	/* server */
    {"text", COMMAND_TEXT, 0},		/* server */
    {"msrvdone", COMMAND_MSRVDONE, 0},	/* metaserver */
    {"server", COMMAND_SERVER, 0},	/* metaserver */
    {"mape", COMMAND_MAPELEV, 0},	/* server */
    {"mapt", COMMAND_MAPTERR, 0},	/* server */
    {"maps", COMMAND_MAPSHIFT, 0},	/* server */
    {"mapd", COMMAND_MAPDELETE, 0},	/* server */
    {"srvfull", COMMAND_SRVFULL, 0},	/* server */
    {"srvd", COMMAND_SRVD, 0}		/* server */
};

const size_t comm_size = sizeof(commtable) / sizeof(argtable_t);

/**
   \brief Generic null handler for commands
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_null(pargs_t *args, char *pos)
{
    return(0);
}

/**
   \brief Parse a metaserver done command from the metaserver
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_msrvdone(pargs_t *args, char *pos)
{
    close(mainsocket);
    mainsocket = 0;
    connected = 0;
    got_mservdata = 1;
    return(0);
}

/**
   \brief Parse a server info command from the metaserver
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_server(pargs_t *args, char *pos)
{
    mservdata_t *dat;
    int i;

    dat = smalloc(mservdata_t);
    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case SRVID: dat->id = add_string(args[i].arg.c); break;
	case NAME: dat->name = add_string(args[i].arg.c); break;
	case IPADDR: dat->ip = add_string(args[i].arg.c); break;
	case PORTNUM: dat->port = args[i].arg.i; break;
	case DESC: dat->desc = add_string(args[i].arg.c); break;
	case MAPINFO: dat->mapinfo = add_string(args[i].arg.c); break;
	case HOSTNAME: dat->hostname = add_string(args[i].arg.c); break;
	case SRVVERSION: dat->version = add_string(args[i].arg.c); break;
	case WEBPAGE: dat->webpage = add_string(args[i].arg.c); break;
	case OWNERNAME: dat->ownername = add_string(args[i].arg.c); break;
	case OWNEREMAIL: dat->owneremail = add_string(args[i].arg.c); break;
	case LASTUPDATE: dat->lastupdate = args[i].arg.l; break;
	case FIRSTUPDATE: dat->firstupdate = args[i].arg.l; break;
	case NUMUSERS: dat->numusers = args[i].arg.i; break;
	case USERHWM: dat->userhwm = args[i].arg.i; break;
	case MAXUSERS: dat->maxusers = args[i].arg.i; break;
	case PREFERIP: dat->preferip = 1; break;
	}
    }
    TAILQ_INSERT_HEAD(&mservdata, dat, next);
    return(0);
}

/**
   \brief Parse a bounding box command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_bbox(pargs_t *args, char *pos)
{
    return(0);
}

/**
   \brief Parse a you are command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_ur(pargs_t *args, char *pos)
{
    int i;
    char b2[40];
    i = find_arg_by_id(args, ID);
    if (i == -1) {
	fprintf(stderr, "Did not recieve valid UR\n");
	return(0);
    }
    uid_cpy(&player_id ,&args[i].arg.uid);
    printf("I AM id %s\n", uid_sprint(b2,&player_id));
    add_main_char(&player_id);
    return(0);
}

/**
   \brief Parse a new object command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_nuob(pargs_t *args, char *pos)
{
    int i, placed, draw=1, state=STATE_IDLE;
    client_t *obj, *tobj;

    obj = new_object();
    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case ID: uid_cpy(&obj->uid,&args[i].arg.uid); break;
	case NROF: obj->nrof = args[i].arg.i; break;
	case DRAW: draw = args[i].arg.i; break;
	case STATE:
	    change_obj_state(obj, args[i].arg.i);
	    state = args[i].arg.i;
	    break;
	case NAME: obj->name = add_string(args[i].arg.c); break;
	case NAMEPL: obj->name_pl = add_string(args[i].arg.c); break;
	case MODEL: obj->model = add_string(args[i].arg.c); break;
	case FACEX: obj->facing.x = args[i].arg.f; break;
	case FACEY: obj->facing.y = args[i].arg.f; break;
	case FACEZ: obj->facing.z = args[i].arg.f; break;
	case DIRX: obj->direction.x = args[i].arg.f; break;
	case DIRY: obj->direction.y = args[i].arg.f; break;
	case DIRZ: obj->direction.z = args[i].arg.f; break;
	case COORDX: obj->coord.x = args[i].arg.f; break;
	case COORDY: obj->coord.y = args[i].arg.f; break;
	case COORDZ: obj->coord.z = args[i].arg.f; break;
	case SCALEX: obj->scale.x = args[i].arg.f; break;
	case SCALEY: obj->scale.y = args[i].arg.f; break;
	case SCALEZ: obj->scale.z = args[i].arg.f; break;
	case VELOCITY: obj->velocity = args[i].arg.f; break;
	/* extended info */
	case SEX: obj->stats.sex = args[i].arg.i; break;
	case TYPE: obj->stats.type = args[i].arg.i; break;
	case SUBTYPE: obj->stats.subtype = args[i].arg.i; break;
	case WEIGHT: obj->stats.weight = args[i].arg.i; break;
	case LEVEL: obj->stats.level = args[i].arg.i; break;
	case EXP: obj->stats.exp = args[i].arg.i; break;
	case MATERIAL: obj->stats.material = args[i].arg.i; break;
	case VALUE: obj->stats.value = args[i].arg.i; break;
	case STR: obj->stats.Str = args[i].arg.s; break;
	case DEX: obj->stats.Dex = args[i].arg.s; break;
	case CON: obj->stats.Con = args[i].arg.s; break;
	case WIS: obj->stats.Wis = args[i].arg.s; break;
	case INT: obj->stats.Int = args[i].arg.s; break;
	case PER: obj->stats.Per = args[i].arg.s; break;
	case FAME: obj->stats.fame = args[i].arg.i; break;
	case APPEAR: obj->stats.appr = args[i].arg.i; break;
	case HP: obj->stats.hp = args[i].arg.i; break;
	case MAXHP: obj->stats.maxhp = args[i].arg.i; break;
	case SP: obj->stats.sp = args[i].arg.i; break;
	case MAXSP: obj->stats.maxsp = args[i].arg.i; break;
	case STAMINA: obj->stats.stm = args[i].arg.i; break;
	case MAXSTAMINA: obj->stats.maxstm = args[i].arg.i; break;
	case FOOD: obj->stats.food = args[i].arg.i; break;
	case WATER: obj->stats.water = args[i].arg.i; break;
	case AC: obj->stats.ac = args[i].arg.i; break;
	case SPEED: obj->stats.speed = args[i].arg.i; break;
	}
    }

    placed = 0;
    TAILQ_FOREACH(tobj, &objects, next) {
	if (uid_eq(&tobj->uid ,&obj->uid)) {
	    /* the server has told us nuob for an object we allready know of */
	    parse_command_chob(args, pos);
	    return(0);
	}
    }
    TAILQ_FOREACH(tobj, &objects, next) {
	if (uid_cmp(&tobj->uid ,& obj->uid)<0) {
	    TAILQ_INSERT_BEFORE(tobj, obj, next);
	    placed++;
	    break;
	}
    }
    if (!placed)
	TAILQ_INSERT_TAIL(&objects, obj, next);

    /* now parse the model */
    if (obj->model != NULL && draw)
	model_object(obj, NULL);
    if (draw && obj->anim != NULL && state != STATE_IDLE) {
	change_obj_state(obj, STATE_IDLE);
	change_obj_state(obj, state);
    }
    return(0);
}

/**
   \brief Parse a change object command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_chob(pargs_t *args, char *pos)
{
    int i,  scaled, rotated, extended, draw=1;
    client_t *obj;
    RMvertex3D v;
    uniqid_t id;

    obj = NULL;
    scaled = 0;
    rotated = 0;
    extended = 0;
    uid_clear(&id);
    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case ID:
	    uid_cpy(&id,&args[i].arg.uid);
	    obj = find_obj_by_id(&id);
	    if (obj == NULL)
		return(0);
	    break;
	case NROF: obj->nrof = args[i].arg.i; break;
	case DRAW: draw = args[i].arg.i; break;
	case STATE: change_obj_state(obj, args[i].arg.i); break;
	case NAME: obj->name = add_string(args[i].arg.c); break;
	case NAMEPL: obj->name_pl = add_string(args[i].arg.c); break;
	case MODEL: obj->model = add_string(args[i].arg.c); break;
	case FACEX: obj->facing.x = args[i].arg.f; rotated++; break;
	case FACEY: obj->facing.y = args[i].arg.f; rotated++; break;
	case FACEZ: obj->facing.z = args[i].arg.f; rotated++; break;
	case DIRX: obj->direction.x = args[i].arg.f; break;
	case DIRY: obj->direction.y = args[i].arg.f; break;
	case DIRZ: obj->direction.z = args[i].arg.f; break;
	case COORDX: obj->coord.x = args[i].arg.f; break;
	case COORDY: obj->coord.y = args[i].arg.f; break;
	case COORDZ: obj->coord.z = args[i].arg.f; break;
	case SCALEX: obj->scale.x = args[i].arg.f; scaled++; break;
	case SCALEY: obj->scale.y = args[i].arg.f; scaled++; break;
	case SCALEZ: obj->scale.z = args[i].arg.f; scaled++; break;
	case VELOCITY: obj->velocity = args[i].arg.f; break;
	/* extended info */
	case SEX: obj->stats.sex = args[i].arg.i; extended++; break;
	case TYPE: obj->stats.type = args[i].arg.i; extended++; break;
	case SUBTYPE: obj->stats.subtype = args[i].arg.i; extended++; break;
	case WEIGHT: obj->stats.weight = args[i].arg.i; extended++; break;
	case LEVEL: obj->stats.level = args[i].arg.i; extended++; break;
	case EXP: obj->stats.exp = args[i].arg.i; extended++; break;
	case MATERIAL: obj->stats.material = args[i].arg.i; extended++; break;
	case VALUE: obj->stats.value = args[i].arg.i; extended++; break;
	case STR: obj->stats.Str = args[i].arg.s; extended++; break;
	case DEX: obj->stats.Dex = args[i].arg.s; extended++; break;
	case CON: obj->stats.Con = args[i].arg.s; extended++; break;
	case WIS: obj->stats.Wis = args[i].arg.s; extended++; break;
	case INT: obj->stats.Int = args[i].arg.s; extended++; break;
	case PER: obj->stats.Per = args[i].arg.s; extended++; break;
	case FAME: obj->stats.fame = args[i].arg.i; extended++; break;
	case APPEAR: obj->stats.appr = args[i].arg.i; extended++; break;
	case HP: obj->stats.hp = args[i].arg.i; extended++; break;
	case MAXHP: obj->stats.maxhp = args[i].arg.i; extended++; break;
	case SP: obj->stats.sp = args[i].arg.i; extended++; break;
	case MAXSP: obj->stats.maxsp = args[i].arg.i; extended++; break;
	case STAMINA: obj->stats.stm = args[i].arg.i; extended++; break;
	case MAXSTAMINA: obj->stats.maxstm = args[i].arg.i; extended++; break;
	case FOOD: obj->stats.food = args[i].arg.i; extended++; break;
	case WATER: obj->stats.water = args[i].arg.i; extended++; break;
	case AC: obj->stats.ac = args[i].arg.i; extended++; break;
	case SPEED: obj->stats.speed = args[i].arg.i; extended++; break;
	}
    }

    /* now move the object */
    if (obj->model != NULL && draw) {
	RMnode *node;
	RMmatrix old, new, model;

	node = obj->node;
	if (node == NULL) {
	    model_object(obj, NULL);
	    node = obj->node;
	}

	v.x = obj->coord.x;
	v.y = obj->coord.y;
	v.z = obj->coord.z;
	rmNodeGetCompositeModelMatrix(node, &model);
	/*rmPrintMatrix(&model);*/
	rmNodeSetTranslateVector(obj->node, &v);
	if (scaled) {
	    rmMatrixIdentity(&new);
	    new.m[0][0] = obj->scale.x;
	    new.m[1][1] = obj->scale.y;
	    new.m[2][2] = obj->scale.z;
	    rmMatrixIdentity(&old);
	    rmMatrixMultiply(&old, &new, &old);
	    rmNodeSetScaleMatrix(obj->node, &old);
	}
	if (rotated) {
	    rotate_obj(mainpipe, obj->node, obj->facing, ROT_FROMZERO);
	}
	if (uid_eq(&id, &player_id)) {
	    moveCamera(node);
	    if (extended)
		update_main_display_from_object(obj);
	}
    }
    return(0);
}

/**
   \brief Parse a no move command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_nomv(pargs_t *args, char *pos)
{
    client_t *obj;

    obj = (client_t *)rmNodeGetClientData(mainchar);
    change_obj_state(obj, STATE_IDLE);
    return(0);
}

/**
   \brief Parse a action command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_eact(pargs_t *args, char *pos)
{
    int i, ac=-1;
    client_t *obj;

    obj = NULL;
    i = find_arg_by_id(args, ID);
    if (i == -1)
	return(0);
    obj = find_obj_by_id(&args[i].arg.uid);
    if (obj == NULL)
	return(0);

    i = find_arg_by_id(args, ACT);
    if (i == -1)
	return(0);
    ac = args[i].arg.i;

    if (ac == -1 || obj->anim == NULL)
	return(0);
    perform_action(obj, ac);
    return(0);
}

/**
   \brief Parse an add inventory command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_inv(pargs_t *args, char *pos)
{
    int i,  bucket=BUCKET_INV;
    client_t *obj, *container;
    uniqid_t id, cont;

    obj = NULL;
    uid_clear(&id);
    uid_clear(&cont);
    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case ID:
	    uid_cpy(&id,& args[i].arg.uid);
	    obj = find_obj_by_id(&id);
	    if (obj == NULL)
		return(0);
	    break;
	case CONTAINER:
	    uid_cpy(&cont ,&args[i].arg.uid);
	    container = find_obj_by_id(&cont);
	    if (container == NULL)
		return(0);
	    break;
	case BUCKET:
	    bucket = args[i].arg.i;
	    break;
	}
    }
    if (i < 2)
	return(0);
    if (uid_isnull(&obj->container) )
	remove_object_from_inv(obj);
    uid_cpy(&obj->container ,&id);
    if (uid_eq(&cont,&player_id)) {
	add_obj_to_bucket(obj, bucket);
    }
    return(0);
}

/**
   \brief Parse a remove inventory command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_rminv(pargs_t *args, char *pos)
{
    int i;
    client_t *obj, *container;
    uniqid_t  cont;
    uid_clear(&cont);
    obj = NULL;
    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case ID:
	    obj = find_obj_by_id(&args[i].arg.uid);
	    if (obj == NULL)
		return(0);
	    break;
	case CONTAINER:
	    uid_cpy(&cont ,&args[i].arg.uid);
	    container = find_obj_by_id(&cont);
	    if (container == NULL)
		return(0);
	    break;
	}
    }
    if (i < 1)
	return(0);
    uid_clear(&obj->container);
    if (uid_eq(&cont,&player_id)) {
	remove_object_from_inv(obj);
    }
    return(0);
}

/**
   \brief Parse an object removal command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_rmob(pargs_t *args, char *pos)
{
    int i;
    client_t *obj;

    obj = NULL;
    i = find_arg_by_id(args, ID);
    if (i == -1)
	return(0);
    obj = find_obj_by_id(&args[i].arg.uid);
    /* something is seriously wrong here */
    if (obj == NULL)
	return(0);

    if (obj->model != NULL && obj->node != NULL)
	remove_obj(obj);
    return(0);
}

/**
   \brief Parse an equip command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_eq(pargs_t *args, char *pos)
{
    int i, slot=0, draw=1, fbone=0;
    client_t *obj, *eq;
    attach_t *attach;

    eq = obj = NULL;
    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case ID:
	    obj = find_obj_by_id(&args[i].arg.uid);
	    if (obj == NULL)
		return(0);
	    break;
	case EQUIP:
	    eq = find_obj_by_id(&args[i].arg.uid);
	    if (eq == NULL)
		return(0);
	    break;
	case SLOT:
	    slot = args[i].arg.i;
	    break;
	case DRAW:
	    draw = args[i].arg.i;
	    break;
	}
    }
    if (i < 2)
	return(0);

    if (obj->anim == NULL)
	draw = 0;
    else {
	TAILQ_FOREACH(attach, &obj->anim->coremodel->attachments, next) {
	    if (attach->attachpoint == slot) {
		fbone++;
		break;
	    }
	}
	if (fbone) {
	    /* the object is presumably from another object's inventory.  That
	       implies, that it isn't modelled. */
	    if (eq->model != NULL && draw) {
		model_object(eq, NULL);
		attach->node = eq->node;
	    }
	}
    }
    return(0);
}

/**
   \brief Parse an unequip command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_ueq(pargs_t *args, char *pos)
{
    int i, draw=1, found;
    client_t *obj, *eq;
    attach_t *attach;

    eq = obj = NULL;
    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case ID:
	    obj = find_obj_by_id(&args[i].arg.uid);
	    if (obj == NULL)
		return(0);
	    break;
	case EQUIP:
	    eq = find_obj_by_id(&args[i].arg.uid);
	    if (eq == NULL)
		return(0);
	    break;
	case DRAW:
	    draw = args[i].arg.i;
	    break;
	}
    }
    if (i < 1)
	return(0);

    if (obj->anim == NULL)
	draw = 0;
    else {
	found = 0;
	TAILQ_FOREACH(attach, &obj->anim->coremodel->attachments, next)
	    if (attach->node == eq->node) {
		found++;
		break;
	    }
	if (found) {
	    printf("FOUND and detached\n");
	    TAILQ_REMOVE(&obj->anim->coremodel->attachments, attach, next);
	    free(attach);
	}
	remove_obj(eq);
    }
    return(0);
}

/**
   \brief Parse a text command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_text(pargs_t *args, char *pos)
{
    int i, type=0,volume=2;
    char *data=NULL, *buf=NULL,*from=NULL;

    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case NAME:
	    from = args[i].arg.c;
	    break;
	case TYPE:
	    type = args[i].arg.i;
	    break;
	case DATA:
	    data = args[i].arg.c;
	    break;
	case VOLUME:
	    volume = args[i].arg.i;
	    break;
	}
    }
	 
    /*printf("SAY name:%s type:%i vol:%i data:%s\n",from,type,volume,data); */
    if (data == NULL)
	return(0);

    buf = safer_malloc(sizeof(char)*(strlen(data)+256));
    switch (type) {
    case DATA_MOTD:
    case DATA_NEWS:
    case DATA_INFO:
	sprintf(buf, "%s\n", data);
	break;
    case DATA_SAY:
	switch (volume) {
	case 1 : /*whisper */
	    if (!strcmp(from, player_obj->name))
		sprintf(buf, "You whisper: %s\n", data);
	    else if (from != NULL)
		sprintf(buf, "%s whispers: %s\n", from, data);
	    else
		sprintf(buf, "Someone whispers: %s\n", data);
	    break;
	case 3 : /*shout */
	    if (!strcmp(from, player_obj->name))
		sprintf(buf, "You shout: %s\n", data);
	    else if (from != NULL)
		sprintf(buf, "%s shouts: %s\n", from, data);
	    else
		sprintf(buf, "Someone shouts: %s\n", data);
	    break;
	default : /*say */
	    if (!strcmp(from, player_obj->name))
		sprintf(buf, "You say: %s\n", data);
	    else if (from != NULL)
		sprintf(buf, "%s says: %s\n", from, data);
	    else
		sprintf(buf, "Someone says: %s\n", data);
	    break;
	}
	break; 
    }

    display_message(buf);
    free(buf);
    return(0);
}

/**
   \brief Parse a mapelevation command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_mape(pargs_t *args, char *pos)
{
    int i, datasize=0, id;
    FILE *fp;
    char filename[MAX_PATH];

    i = find_arg_by_id(args, MAPID);
    if (i == -1)
	return(0);
    id = args[i].arg.i;

    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case WIDTH:
	    mapcache[id].width = args[i].arg.i;
	    break;
	case HEIGHT:
	    mapcache[id].height = args[i].arg.i;
	    break;
	case BINARY:
	    datasize = args[i].arg.i;
	    break;
	case COORDX:
	    mapcache[id].basex = args[i].arg.f;
	    break;
	case COORDY:
	    mapcache[id].basey = args[i].arg.f;
	    break;
	case COORDZ:
	    mapcache[id].basez = args[i].arg.f;
	    break;
	}
    }

    /* now we have parsed the text portion, the binary stream should follow
       directly after this.  If not, we're fucked and must bail. */

    if (datasize > inbuflen || datasize == 0)
	return(-1); /* bad juju */

    sprintf(filename, "/tmp/gencli.mape.%d.%d", id, getpid());
    fp = fopen(filename, "wb");
    if (fp == NULL)
	return(-1);
    i = fwrite(pos, datasize, 1, fp);
    fclose(fp);
    mapcache[id].elevation = parse_pngmap8bit(filename, mapcache[id].width,
					      mapcache[id].height);
    unlink(filename);

    addFloor(groundnode, id);

    return(datasize);
}

/**
   \brief Parse a mapterrain command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_mapt(pargs_t *args, char *pos)
{
    int i, datasize=0, id;
    FILE *fp;
    char filename[MAX_PATH];

    i = find_arg_by_id(args, MAPID);
    if (i == -1)
	return(0);
    id = args[i].arg.i;

    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case WIDTH:
	    mapcache[id].width = args[i].arg.i;
	    break;
	case HEIGHT:
	    mapcache[id].height = args[i].arg.i;
	    break;
	case BINARY:
	    datasize = args[i].arg.i;
	    break;
	case COORDX:
	    mapcache[id].basex = args[i].arg.f;
	    break;
	case COORDY:
	    mapcache[id].basey = args[i].arg.f;
	    break;
	case COORDZ:
	    mapcache[id].basez = args[i].arg.f;
	    break;
	}
    }

    /* now we have parsed the text portion, the binary stream should follow
       directly after this.  If not, we're fucked and must bail. */

    if (datasize > inbuflen)
	return(-1); /* bad juju */

    sprintf(filename, "/tmp/gencli.mapt.%d.%d", id, getpid());
    fp = fopen(filename, "wb");
    if (fp == NULL)
	return(-1);
    i = fwrite(pos, datasize, 1, fp);
    fclose(fp);
    mapcache[id].terrain = parse_pngmap8bit(filename, mapcache[id].width,
					    mapcache[id].height);
    unlink(filename);

    addFloor(groundnode, id);

    return(datasize);
}

/**
   \brief Parse a mapshift command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_maps(pargs_t *args, char *pos)
{
    int i, dir;

    i = find_arg_by_id(args, ACT);
    if (i == -1)
	return(0);
    dir = args[i].arg.i;

    shift_maps(dir);
    return(0);
}

/**
   \brief Parse a mapdelete command from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_mapd(pargs_t *args, char *pos)
{
    int i, dir;

    i = find_arg_by_id(args, ACT);
    if (i == -1)
	return(0);
    dir = args[i].arg.i;

    cleanFloor(groundnode, dir);
    return(0);
}

/**
   \brief Parse the fact that the server will allow no more connections
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_srvfull(pargs_t *args, char *pos)
{
    display_message("The server is full, try again later.\n");
    return(0);
}

/**
   \brief Parse server definitions from the server
   \param args arguments
   \param pos current position in the inbuf stream
   \return amount of data taken off the stream
*/

int parse_command_srvd(pargs_t *args, char *pos)
{
    int i;

    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case HP:
	    serverdefs.hoursday = args[i].arg.i;
	    break;
	}
    }
    return(0);
}
