/* $Id: skelanim.c,v 1.23 2004/04/01 23:23:50 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <sys/param.h>
#include <sys/time.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
/*#define bool char*/
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <rm/rmaux.h>
#include <gtk/gtk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#undef False
#undef True
#define bool enum Boolean
#include <cal3d/cal3d_wrapper.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "defines.h"
#include "timers.h"
#include "mappings.h"
#include "cmotion.h"
#include "shstr.h"
#include "jpegloader.h"
#include "pngloader.h"
#include "utils.h"
#include "xmlparse.h"
#include "cutils.h"
#include "cobject.h"
#include "ui.h"
#include "skelanim.h"
#include "settings.h"

extern cmap_t mapcache[9];
RMnode *mainchar = NULL;
char *progname;
extern csettings_t settings;

TAILQ_HEAD(, _anim_t) animtable = TAILQ_HEAD_INITIALIZER(animtable);
TAILQ_HEAD(, _calcore_t) calcoretable = TAILQ_HEAD_INITIALIZER(calcoretable);

/* return the animation number that matches the requested state */

int find_anim_for_state(calcore_t *corem, int state)
{
    int i;

    for (i=0; corem->animations[i] != -1; i++) {
	if (corem->animations[i] == state)
	    return(i);
    }
    return(-1); /* no animation matches */
}

void perform_action(client_t *obj, int action)
{
    anim_t *an;
    int i;

    an = obj->anim;
    i = find_anim_for_state(an->coremodel, action);
    if (i == -1)
	return;
    CalMixer_ExecuteAction(CalModel_GetMixer(an->calmodel), i,
			   an->animtimeleft, an->blendtime);
}

/* do basic animation updates on all calmodels before entering the true
   render stage. This solves the problem of equipment not being in sync
   with the models when rendering occurs.
*/

void prerender_calmodels(void)
{
    anim_t *an;
    RMvertex3D *cvec, v, v2;
    RMmatrix m1, m2;
    attach_t *attach;
    quat_t *quat;

    TAILQ_FOREACH(an, &animtable, next) {
	if(!an->paused) {
	    /* check if the time has come to blend to the next animation */
#if 1
	    if(CalCoreModel_GetCoreAnimationCount(an->coremodel->coremodel) > 1) {
		an->animtimeleft -= 0.1;
/*		CalMixer_ClearCycle(CalModel_GetMixer(an->calmodel), 1, 0.0);*/
		if(an->animtimeleft <= an->blendtime) {
		    /* get the next animation */
		    /* an->curanimid = (an->curanimid + 1) %
		       CalCoreModel_GetCoreAnimationCount(an->coremodel->coremodel);*/
		    /* fade in the new animation */
		    /*CalMixer_ExecuteAction(CalModel_GetMixer(an->calmodel),
					   an->curanimid, an->animtimeleft,
					   an->blendtime);*/
		    CalMixer_BlendCycle(CalModel_GetMixer(an->calmodel), an->curanimid, 1.0, 0.0);
		    /* adjust the animation time left until next animation flip */
		    an->animtimeleft = CalCoreAnimation_GetDuration(CalCoreModel_GetCoreAnimation(an->coremodel->coremodel, an->curanimid)) - an->blendtime;
		}
		if (an->prevanimid != an->curanimid) {
		    CalMixer_ClearCycle(CalModel_GetMixer(an->calmodel),
					an->prevanimid, 0.3);
		    CalMixer_BlendCycle(CalModel_GetMixer(an->calmodel),
					an->curanimid, 1.0, 0.0);
		    an->prevanimid = an->curanimid;
		}
		    
	    }
#endif
	    CalModel_Update(an->calmodel, 0.1); /* XXX */
	}

	TAILQ_FOREACH(attach, &an->coremodel->attachments, next) {
	    if (attach->node != NULL) {
		cvec = (RMvertex3D *)CalBone_GetTranslationAbsolute(CalSkeleton_GetBone(CalModel_GetSkeleton(an->calmodel), attach->boneid));
		rmMatrixIdentity(&m1);
		prep_xmat(&m2, an->coremodel->tiltangle);
		m1.m[3][0] = cvec->x;
		m1.m[3][1] = cvec->y;
		m1.m[3][2] = cvec->z;
		rmMatrixMultiply(&m1, &m2, &m1);
		prep_ymat(&m2, an->coremodel->twistangle); 
		rmMatrixMultiply(&m1, &m2, &m1);
		rmNodeGetRotateMatrix(an->node, &m2);
		/* keep the sword the same direction as the node */
		rmNodeSetRotateMatrix(attach->node, &m2);
		rmMatrixMultiply(&m1, &m2, &m1);
		v2.x = m1.m[3][0] * an->coremodel->scale;
		v2.y = m1.m[3][1] * an->coremodel->scale;
		v2.z = m1.m[3][2] * an->coremodel->scale;
		rmNodeGetTranslateVector(an->node, &v);
		v2.x += v.x;
		v2.y += v.y;
		v2.z += v.z;
		rmNodeSetTranslateVector(attach->node, &v2);
		/* now rotate the attached object */
		rmNodeGetRotateMatrix(attach->node, &m1);
		quat = (quat_t *)CalBone_GetRotationAbsolute(CalSkeleton_GetBone(CalModel_GetSkeleton(an->calmodel), attach->boneid));
		rmMatrixIdentity(&m2);
		quat_to_matrix(quat, &m2);
		rmMatrixMultiply(&m2, &m1, &m2);
		if (attach->xrot != 0.0) {
		    prep_xmat(&m1, attach->xrot);
		    rmMatrixMultiply(&m2, &m1, &m2);
		}
		if (attach->yrot != 0.0) {
		    prep_ymat(&m1, attach->yrot);
		    rmMatrixMultiply(&m2, &m1, &m2);
		}
		if (attach->zrot != 0.0) {
		    prep_zmat(&m1, attach->zrot);
		    rmMatrixMultiply(&m2, &m1, &m2);
		}
		rmNodeSetRotateMatrix(attach->node, &m2);
	    }
	}
    }
}

void renderModel(anim_t *an)
{
    struct CalRenderer *pCalRenderer;
    int meshCount;
    int meshId;

    an->vertexcount = 0;
    an->facecount = 0;

    /* get the renderer of the model */
    pCalRenderer = CalModel_GetRenderer(an->calmodel);

    glRotatef(an->coremodel->tiltangle, 1.0f, 0.0f, 0.0f);
    glRotatef(an->coremodel->twistangle, 0.0f, 0.0f, 1.0f);

    /*  begin the rendering loop */
    if(CalRenderer_BeginRendering(pCalRenderer)) {
	/* set global OpenGL states */
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	/* we will use vertex arrays, so enable them */
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	/* get the number of meshes */
	meshCount = CalRenderer_GetMeshCount(pCalRenderer);

	/* render all meshes of the model */
	for (meshId = 0; meshId < meshCount; meshId++) {
	    int submeshCount;
	    int submeshId;
	    /* get the number of submeshes */
	    submeshCount = CalRenderer_GetSubmeshCount(pCalRenderer, meshId);

	    /* render all submeshes of the mesh */
	    for (submeshId = 0; submeshId < submeshCount; submeshId++) {
		/* select mesh and submesh for further data access */
		if (CalRenderer_SelectMeshSubmesh(pCalRenderer, meshId, submeshId)) {
		    unsigned char meshColor[4];
		    GLfloat materialColor[4];
		    float shininess;
		    float *meshVertices, *meshNormals, *meshTextureCoordinates;
		    CalIndex *meshFaces;
		    int vertexCount, textureCoordinateCount, faceCount, za;

		    /* set the material ambient color */
		    CalRenderer_GetAmbientColor(pCalRenderer, &meshColor[0]);
		    materialColor[0] = meshColor[0] / 255.0;
		    materialColor[1] = meshColor[1] / 255.0;
		    materialColor[2] = meshColor[2] / 255.0;
		    materialColor[3] = meshColor[3] / 255.0;
		    glMaterialfv(GL_FRONT, GL_AMBIENT, materialColor);

		    /* set the material diffuse color */
		    CalRenderer_GetDiffuseColor(pCalRenderer, &meshColor[0]);
		    materialColor[0] = meshColor[0] / 255.0;
		    materialColor[1] = meshColor[1] / 255.0;
		    materialColor[2] = meshColor[2] / 255.0;
		    materialColor[3] = meshColor[3] / 255.0;
		    glMaterialfv(GL_FRONT, GL_DIFFUSE, materialColor);

		    /* set the material specular color */
		    CalRenderer_GetSpecularColor(pCalRenderer, &meshColor[0]);
		    materialColor[0] = meshColor[0] / 255.0;
		    materialColor[1] = meshColor[1] / 255.0;
		    materialColor[2] = meshColor[2] / 255.0;
		    materialColor[3] = meshColor[3] / 255.0;
		    glMaterialfv(GL_FRONT, GL_SPECULAR, materialColor);

		    /* set the material shininess factor
		       float shininess;
		       shininess = 50.0f; pCalRenderer->getShininess();
		    */
		    shininess = CalRenderer_GetShininess(pCalRenderer);
		    glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);

		    /* get the transformed vertices of the submesh */
		    vertexCount = CalRenderer_GetVertexCount(pCalRenderer);
		    meshVertices = malloc(sizeof(float) * (vertexCount+1) * 3);
		    (void)CalRenderer_GetVertices(pCalRenderer, meshVertices);

		    /* get the transformed normals of the submesh */
		    meshNormals = malloc(sizeof(float) * (vertexCount+1) * 3);
		    (void)CalRenderer_GetNormals(pCalRenderer, meshNormals);

		    /* get the texture coordinates of the submesh */
		    meshTextureCoordinates = malloc(sizeof(float) * (vertexCount+1) * 2);
		    textureCoordinateCount = CalRenderer_GetTextureCoordinates(pCalRenderer, 0, meshTextureCoordinates);

		    /* get the faces of the submesh */

		    faceCount = CalRenderer_GetFaceCount(pCalRenderer);
		    meshFaces = malloc(sizeof(CalIndex) * (faceCount+1) * 3);
		    (void)CalRenderer_GetFaces(pCalRenderer, meshFaces);

		    for (za=0; za < vertexCount*3; za++)
			meshVertices[za] *= an->coremodel->scale;

		    /* set the vertex and normal buffers */
		    glVertexPointer(3, GL_FLOAT, 0, meshVertices);
		    glNormalPointer(GL_FLOAT, 0, meshNormals);

		    /* set the texture coordinate buffer and state if necessary */
		    if((CalRenderer_GetMapCount(pCalRenderer) > 0) && (textureCoordinateCount > 0)) {
			glEnable(GL_TEXTURE_2D);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glEnable(GL_COLOR_MATERIAL);

			/* set the texture id we stored in the map user data */
			glBindTexture(GL_TEXTURE_2D, (GLuint)CalRenderer_GetMapUserData(pCalRenderer, 0));

			/* set the texture coordinate buffer */
			glTexCoordPointer(2, GL_FLOAT, 0, meshTextureCoordinates);
			glColor3f(1.0f, 1.0f, 1.0f);
		    }

		    /* draw the submesh */
                  
		    if (sizeof(CalIndex)==2)
			glDrawElements(GL_TRIANGLES, faceCount * 3, GL_UNSIGNED_SHORT, meshFaces);
		    else
			glDrawElements(GL_TRIANGLES, faceCount * 3, GL_UNSIGNED_INT, meshFaces);

		    /* disable the texture coordinate state if necessary */
		    if((CalRenderer_GetMapCount(pCalRenderer) > 0) && (textureCoordinateCount > 0)) {
			glDisable(GL_COLOR_MATERIAL);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
			glDisable(GL_TEXTURE_2D);
		    }

		    /* adjust the vertex and face counter */
		    an->vertexcount += vertexCount;
		    an->facecount += faceCount;
		    /* free malloc'd data */
		    free(meshVertices);
		    free(meshNormals);
		    free(meshTextureCoordinates);
		    free(meshFaces);
		}
	    }
	}

	// clear vertex array state
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	// clear light
	glDisable(GL_LIGHTING);
	glDisable(GL_LIGHT0);
	glDisable(GL_DEPTH_TEST);

	// end the rendering
	CalRenderer_EndRendering(pCalRenderer);
    }
}

/*(struct RMprimitive *p, RMnode *r, RMstate *s, RMpipe *renderPipe,
  RMstateCache *rsc) */
void rmCal3D OGLPRIMPARMLIST() {
    renderModel((anim_t *)p->clientData);
}

GLuint loadcalTexture(const char *filename, calcore_t *corem) {
    int depth, width, height;
    GLuint textureId;
    char file[MAX_PATH];
    unsigned char *pBuffer=NULL;

    if(strncmp(strrchr(filename, '.'), ".jpg", 4) == 0) {
	sprintf(file, "%s/%s/%s", settings.texdir, corem->tpath, filename);
	pBuffer = readJPEG(file, &width, &height);
	depth = 3;

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glGenTextures(1, &textureId);

	glBindTexture(GL_TEXTURE_2D, textureId);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, (depth == 3) ? GL_RGB : GL_RGBA, width, height, 0, (depth == 3) ? GL_RGB : GL_RGBA, GL_UNSIGNED_BYTE, pBuffer);
    } else if(strncmp(strrchr(filename, '.'), ".png", 4) == 0) {
	sprintf(file, "%s/%s/%s", settings.texdir, corem->tpath, filename);
	pBuffer = parse_pngRGBA(file, &width, &height);
	depth = 4;

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glGenTextures(1, &textureId);

	glBindTexture(GL_TEXTURE_2D, textureId);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, (depth == 3) ? GL_RGB : GL_RGBA, width, height, 0, (depth == 3) ? GL_RGB : GL_RGBA, GL_UNSIGNED_BYTE, pBuffer);
    } else {
	fprintf(stderr, "Unsupported texture format %s\n", filename);
	bailout();
    }
    if (pBuffer != NULL)
	free(pBuffer);
    return(textureId);
}

calcore_t *loadcalcoremodel(const char *filename)
{
    int materialId;
    FILE *file;
    calcore_t *corem;
    char *tmp, filenm[MAX_PATH];
    xmlDocPtr doc;
    xmlNodePtr cur, ncur, xcur;
    float pts[6];
    int i, anims=0;
    attach_t *attach;

    /* check if this core model is allready loaded */
    TAILQ_FOREACH(corem, &calcoretable, next) {
	if (strcmp(corem->filename, filename) == 0)
	    return(corem);
    }

    forced_redraw();

    corem = smalloc(calcore_t);
    file = fopen(filename, "r");
    corem->coremodel = CalCoreModel_New();
    corem->filename = strdup(filename);
    corem->bmin.x = corem->bmin.y = corem->bmin.z = 999999.0;
    corem->bmax.x = corem->bmax.y = corem->bmax.z = -999999.0;
    CalCoreModel_Create(corem->coremodel, (char *)filename);

    /* XXX for now, just 16 animations, fix later */
    corem->animations = safer_malloc(sizeof(int) * 16);

    doc = xmlParseFile(filename);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "calmodel")) {
	fprintf(stderr, "document of the wrong type, root node != calmodel");
	xmlFreeDoc(doc);
	bailout();
    }
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "scale"))
	    parsefloat(doc, cur, &corem->scale);
	if (!strcmp(cur->name, "tiltangle"))
	    parsefloat(doc, cur, &corem->tiltangle);
	if (!strcmp(cur->name, "twistangle"))
	    parsefloat(doc, cur, &corem->twistangle);
	if (!strcmp(cur->name, "mpath"))
	    parsestring(doc, cur, &corem->path);
	if (!strcmp(cur->name, "tpath"))
	    parsestring(doc, cur, &corem->tpath);

	if (!strcmp(cur->name, "skeleton")) {
	    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    sprintf(filenm, "%s/%s/%s", settings.archdir, corem->path, tmp);
	    CalCoreModel_LoadCoreSkeleton(corem->coremodel, filenm);
	    free(tmp);
	}
	if (!strcmp(cur->name, "animation")) {
	    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    sprintf(filenm, "%s/%s/%s", settings.archdir, corem->path, tmp);
	    CalCoreModel_LoadCoreAnimation(corem->coremodel, filenm);
	    free(tmp);
	    tmp = xmlGetProp(cur, "id");
	    corem->animations[anims] = atoi(tmp);
	    free(tmp);
	    anims++;
	}
	if (!strcmp(cur->name, "mesh")) {
	    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    sprintf(filenm, "%s/%s/%s", settings.archdir, corem->path, tmp);
	    CalCoreModel_LoadCoreMesh(corem->coremodel, filenm);
	    free(tmp);
	}
	if (!strcmp(cur->name, "material")) {
	    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    sprintf(filenm, "%s/%s/%s", settings.texdir, corem->tpath, tmp);
	    CalCoreModel_LoadCoreMaterial(corem->coremodel, filenm);
	    free(tmp);
	}
	if (!strcmp(cur->name, "attachments")) {
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "attachment")) {
		    xcur = ncur->xmlChildrenNode;
		    if (TAILQ_EMPTY(&corem->attachments))
			TAILQ_INIT(&corem->attachments);
		    attach = smalloc(attach_t);
		    attach->node = NULL;
		    while (xcur != NULL) {
			if (!strcmp(xcur->name, "apoint"))
			    parseint(doc, xcur, &attach->attachpoint);
			if (!strcmp(xcur->name, "boneid"))
			    parseint(doc, xcur, &attach->boneid);
			if (!strcmp(xcur->name, "xrot"))
			    parsefloat(doc, xcur, &attach->xrot);
			if (!strcmp(xcur->name, "yrot"))
			    parsefloat(doc, xcur, &attach->yrot);
			if (!strcmp(xcur->name, "zrot"))
			    parsefloat(doc, xcur, &attach->zrot);
			xcur = xcur->next;
		    }
		    TAILQ_INSERT_TAIL(&corem->attachments, attach, next);
		}
		ncur = ncur->next;
	    }
	}
	if (!strcmp(cur->name, "bboxes")) {
	    ncur = cur->xmlChildrenNode;
            while (ncur != NULL) {
                if (!strcmp(ncur->name, "min")) {
                    parsetrifloat(doc, ncur, &pts[0], &pts[1], &pts[2]);
                }
                if (!strcmp(ncur->name, "max")) {
                    parsetrifloat(doc, ncur, &pts[3], &pts[4], &pts[5]);
		    for (i=0; i<6; i++)
                        pts[i] = pts[i] * corem->scale;
                    if (pts[0] > corem->bmax.x)
                        corem->bmax.x = pts[0];
                    if (pts[3] > corem->bmax.x)
                        corem->bmax.x = pts[3];
                    if (pts[0] < corem->bmin.x)
                        corem->bmin.x = pts[0];
                    if (pts[3] < corem->bmin.x)
                        corem->bmin.x = pts[3];
                    if (pts[1] > corem->bmax.y)
                        corem->bmax.y = pts[1];
                    if (pts[4] > corem->bmax.y)
                        corem->bmax.y = pts[4];
                    if (pts[1] < corem->bmin.y)
                        corem->bmin.y = pts[1];
                    if (pts[4] < corem->bmin.y)
                        corem->bmin.y = pts[4];
                    if (pts[2] > corem->bmax.z)
                        corem->bmax.z = pts[2];
                    if (pts[5] > corem->bmax.z)
                        corem->bmax.z = pts[5];
                    if (pts[2] < corem->bmin.z)
                        corem->bmin.z = pts[2];
                    if (pts[5] < corem->bmin.z)
                        corem->bmin.z = pts[5];
                }
                ncur = ncur->next;
            }
	}
	cur = cur->next;
    }
    corem->animations[anims] = -1; /* tag the last one */
    for (materialId = 0; materialId < CalCoreModel_GetCoreMaterialCount(corem->coremodel); materialId++) {
	/* create a material thread */
	CalCoreModel_CreateCoreMaterialThread(corem->coremodel, materialId);
	/* initialize the material thread */
	CalCoreModel_SetCoreMaterialId(corem->coremodel, materialId, 0, materialId);
    }

    /* load all textures and store the opengl texture id in the
       corresponding map in the material */
    for (materialId = 0; materialId < CalCoreModel_GetCoreMaterialCount(corem->coremodel); materialId++) {
	/* get the core material */
	struct CalCoreMaterial *pCoreMaterial;
	int mapId;

	pCoreMaterial = CalCoreModel_GetCoreMaterial(corem->coremodel, materialId);
	/* loop through all maps of the core material */
	for (mapId = 0; mapId < CalCoreMaterial_GetMapCount(pCoreMaterial); mapId++) {
	    GLuint textureId;

	    /* get the filename of the texture */
	    tmp = CalCoreMaterial_GetMapFilename(pCoreMaterial, mapId);

	    /* load the texture from the file */
	    textureId = loadcalTexture(tmp, corem);

	    /* store the opengl texture id in the user data of the map */
	    CalCoreMaterial_SetMapUserData(pCoreMaterial, mapId, (CalUserData)textureId);
	}
    }
    TAILQ_INSERT_TAIL(&calcoretable, corem, next);
    xmlFreeDoc(doc);
    return(corem);
}

anim_t *create_calmodel(char *name) {
    int meshId, i;
    calcore_t *corem;
    anim_t *an;

    corem = loadcalcoremodel(name);

    an = smalloc(anim_t);
    an->coremodel = corem;
    an->lasttick = 0;
    an->paused = 0;
    an->curanimid = 0;
    an->animtimeleft = 0.0;
    an->blendtime = 0.0;
    an->vertexcount = 0;
    an->facecount = 0;

    an->calmodel = CalModel_New();
    CalModel_Create(an->calmodel, corem->coremodel);
    /* attach all meshes to the model */
    for(meshId = 0; meshId < CalCoreModel_GetCoreMeshCount(corem->coremodel); meshId++) {
	if (!CalModel_AttachMesh(an->calmodel, meshId))
	    CalError_PrintLastError();
    }

    /* set the material set of the whole model */
    CalModel_SetMaterialSet(an->calmodel, 0);
    /* set initial animation state */
    if (CalCoreModel_GetCoreAnimationCount(corem->coremodel) > 0) {
	i = find_anim_for_state(corem, ANIM_IDLE);
	if (i != -1) {
	    an->curanimid = i;
	    an->prevanimid = i;
	} else {
	    an->curanimid = 0;
	    an->prevanimid = 0;
	}
	an->animtimeleft = CalCoreAnimation_GetDuration(CalCoreModel_GetCoreAnimation(corem->coremodel, an->curanimid)) - an->blendtime;
	CalMixer_BlendCycle(CalModel_GetMixer(an->calmodel), an->curanimid, 1.0, 0.0);
    } else {
	an->curanimid = -1;
	an->prevanimid = -1;
	an->animtimeleft = -1.0;
    }
    TAILQ_INSERT_TAIL(&animtable, an, next);
    return(an);
}

void delete_calmodel(anim_t *an)
{
    TAILQ_REMOVE(&animtable, an, next);
    CalModel_Destroy(an->calmodel);
    CalModel_Delete(an->calmodel);
    free(an);
}
