/* $Id: ui.c,v 1.33 2004/04/01 16:27:54 xb49 Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <inttypes.h>
#include <sys/time.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkgl.h>
/*#include <gdk/x11/gdkglx.h>
  #include <gdk/x11/gdkglglxext.h>*/
#undef False
#undef True
#define bool enum Boolean
#include <cal3d/cal3d_wrapper.h>

#include "defines.h"
#include "timers.h"
#include "mappings.h"
#include "utils.h"
#include "camera.h"
#include "cmotion.h"
#include "cobject.h"
#include "shstr.h"
#include "cnet.h"
#include "skelanim.h"
#include "modelloader.h"
#include "settings.h"
#include "jpegloader.h"

extern RMnode *mainchar;
extern GtkWidget *commands;
extern GtkWidget *window, *glarea;
extern csettings_t settings;

#define MAXFPS 30
#define ROTATEBY 30
#define MOVEBY 1.0
#define CTRL_MOD_FACTOR 3

int MWidth=670, MHeight=512;
int TWidth=170, THeight=170;
long max_time = 1000000/MAXFPS;
int needs_redraw = 0;
int target_needs_redraw = 0;
int rotate_char = 0;
int ctrl_down = 0;
int motion_dir = 0;
int connected = 0;
GtkWidget *usernm, *passwd, *hostnm;
GdkGLConfig *glconfig, *target_glconfig;

RMpipe *mainpipe = NULL;
RMpipe *targetpipe = NULL;

RMnode *pickednode = NULL;
RMnode *targetnode = NULL;
RMnode *rootnode, *transnode;
RMnode *objnode, *tobjnode, *floornode, *bgnode, *groundnode;
RMnode *gamelogo;

RMtime curtime;
double ms_per_frame;

client_t *pickedobj = NULL;
/*int pickedobjnum = -1;*/

gint       glarea_button_release(GtkWidget*, GdkEventButton*);
gint       glarea_button_press   (GtkWidget*, GdkEventButton*);
gint       glarea_motion_notify  (GtkWidget*, GdkEventMotion*);
gint       glarea_draw           (GtkWidget*, GdkEventExpose*);
gint       glarea_reshape        (GtkWidget*, GdkEventConfigure*);
gint       glarea_init           (GtkWidget*);
gint       glarea_destroy        (GtkWidget*);
gint       glarea_target_draw           (GtkWidget*, GdkEventExpose*);
gint       glarea_target_reshape        (GtkWidget*, GdkEventConfigure*);
gint       glarea_target_init           (GtkWidget*);
gint       glarea_target_destroy        (GtkWidget*);
int        main                  (int, char**);


void render_screen(RMpipe *p, RMnode *n)
{
    RMtime now;
    /* XXX there is a horrific memory leak here 
       This fixes the wierd display glitch when expose events occur.  However
       it leaks about 500k-1M of ram per execution.  Completely unacceptable.
       Until someone figures out how to make the target screen a separate
       thread, we will just live with it.
    */
    /*glXMakeCurrent(rmxPipeGetDisplay(p), rmPipeGetWindow(p),
      rmPipeGetContext(p));*/
    if (p == mainpipe)
	prerender_calmodels();
    rmFrame(p, n);
    rmTimeCurrent(&now);
    ms_per_frame = rmTimeDifferenceMS(&curtime, &now);
    curtime = now;
    /*printf("FPS: %0.2f %f\n", 1000.0/ms_per_frame, ms_per_frame);*/
}

gint update_target_screen(GtkWidget *glarea)
{
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(glarea);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(glarea);

    /* quick checks to see if we even should bother */
    if (!target_needs_redraw)
	return(TRUE);
    if (gdk_gl_drawable_gl_begin(gldrawable, glcontext)) {
	gdk_gl_drawable_wait_gl(gldrawable);
	/*printf("redrawing targetscreen\n");*/
	render_screen(targetpipe, targetnode);
	gdk_gl_drawable_swap_buffers(gldrawable);
	gdk_gl_drawable_gl_end(gldrawable);
	target_needs_redraw = 0;
    }
    return(TRUE);
}

gint update_screen(GtkWidget *glarea)
{
    int ctrl_mod_factor = 1;
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(glarea);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(glarea);

    if (ctrl_down)
	ctrl_mod_factor *= CTRL_MOD_FACTOR;

    if (rotate_char > 0)
	rotate_self(mainpipe, mainchar, 30 / ctrl_mod_factor);
    else if (rotate_char < 0)
	rotate_self(mainpipe, mainchar, -30 / ctrl_mod_factor);

    /* quick checks to see if we even should bother */
    if (!needs_redraw)
      return(TRUE);
    if (!enough_time_passed())
      return(TRUE);
    /* for now */
    if (gdk_gl_drawable_gl_begin(gldrawable, glcontext)) {
	/*printf("redrawing screen\n");*/
	render_screen(mainpipe, rootnode);
	gdk_gl_drawable_swap_buffers(gldrawable);
	gdk_gl_drawable_gl_end(gldrawable);
	if (needs_redraw == 2)
	    needs_redraw = 0;
    }
    return(TRUE);
}

gint key_press_event(GtkWidget *widget, GdkEventKey *event)
{
    if (!connected)
	return(TRUE);
    switch (event->keyval) {
    case GDK_Left:
	rotate_char = 1;
	break;
    case GDK_Right:
	rotate_char = -1;
	break;
    case GDK_Up:
	req_self_movement(1.0, STATE_WALK);
	break;
    case GDK_Down:
	req_self_movement(-1.0, STATE_WALK);
	break;
    case GDK_Control_L:
    case GDK_Control_R:
	ctrl_down = 1;
	change_player_state(STATE_RUN);
	break;
    }
    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
    return(TRUE);
}

gint key_release_event(GtkWidget *widget, GdkEventKey *event)
{
    if (!connected)
	return(TRUE);
    switch (event->keyval) {
    case GDK_Left:
    case GDK_Right:
	rotate_char = 0;
	break;
    case GDK_Up:
    case GDK_Down:
	change_player_state(STATE_IDLE);
	break;
    case GDK_Control_L:
    case GDK_Control_R:
	ctrl_down = 0; /* this is broke XXX */
	change_player_state(STATE_WALK);
	break;
    case GDK_apostrophe:
	gtk_widget_grab_focus(GTK_WIDGET(commands));
	g_signal_handlers_block_by_func(GTK_OBJECT(window),
	     GTK_SIGNAL_FUNC(key_press_event), NULL);
	g_signal_handlers_block_by_func(GTK_OBJECT(window),
	     GTK_SIGNAL_FUNC(key_release_event), NULL);
	break;
    }
    /*printf("Key Release\n");*/
    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_release_event");
    return TRUE;
}

GtkWidget* create_target_glarea(void)
{
    GtkWidget* glarea;
    RMenum targetPlatform = RM_PIPE_GLX;
    RMenum processingMode = RM_PIPE_MULTISTAGE;

    targetpipe = rmPipeNew(targetPlatform);
    rmPipeSetProcessingMode(targetpipe, processingMode);
    rmPipeCreateContext(targetpipe);
    rmPipeMakeCurrent(targetpipe); 
    rmPipeSetDisplayListEnable(targetpipe, RM_FALSE);

    target_glconfig = gdk_gl_config_new_by_mode(GDK_GL_MODE_RGBA |
					 GDK_GL_MODE_DEPTH |
					 GDK_GL_MODE_DOUBLE);

    glarea = gtk_drawing_area_new();
    gtk_widget_set_size_request(glarea, TWidth, THeight);
    gtk_widget_set_gl_capability(glarea, target_glconfig, NULL, TRUE,
				 GDK_GL_RGBA_TYPE);

    gtk_widget_set_events(GTK_WIDGET(glarea), GDK_EXPOSURE_MASK);

    gtk_signal_connect(GTK_OBJECT(glarea), "expose_event",
		       GTK_SIGNAL_FUNC(glarea_target_draw), NULL);
    gtk_signal_connect(GTK_OBJECT(glarea), "configure_event",
		       GTK_SIGNAL_FUNC(glarea_target_reshape), NULL);
    gtk_signal_connect(GTK_OBJECT(glarea), "realize",
		       GTK_SIGNAL_FUNC(glarea_target_init), NULL);
    gtk_signal_connect(GTK_OBJECT(glarea), "destroy",
		       GTK_SIGNAL_FUNC(glarea_target_destroy), NULL);
    gtk_widget_set_usize(GTK_WIDGET(glarea), TWidth, THeight);

    return(glarea);
}

GtkWidget* create_glarea(void)
{
    GtkWidget* glarea;
    RMenum targetPlatform = RM_PIPE_GLX;
    RMenum processingMode = RM_PIPE_MULTISTAGE;

    rmInit();
    mainpipe = rmPipeNew(targetPlatform);
    rmPipeSetProcessingMode(mainpipe, processingMode);
    rmPipeCreateContext(mainpipe);
    rmPipeMakeCurrent(mainpipe);
    rmPipeSetDisplayListEnable(mainpipe, RM_FALSE);

    glconfig = gdk_gl_config_new_by_mode(GDK_GL_MODE_RGBA |
					 GDK_GL_MODE_DEPTH |
					 GDK_GL_MODE_DOUBLE);

    glarea = gtk_drawing_area_new();
    gtk_widget_set_size_request(glarea, MWidth, MHeight);
    gtk_widget_set_gl_capability(glarea, glconfig, NULL, TRUE,
				 GDK_GL_RGBA_TYPE);


    gtk_widget_set_events(GTK_WIDGET(glarea),
			  GDK_EXPOSURE_MASK|
			  GDK_BUTTON_PRESS_MASK|
			  GDK_BUTTON_RELEASE_MASK|
			  GDK_KEY_PRESS_MASK|
			  GDK_KEY_RELEASE_MASK);
    g_signal_connect(GTK_OBJECT(glarea), "button_press_event",
		       GTK_SIGNAL_FUNC(glarea_button_press), NULL);

    g_signal_connect(GTK_OBJECT(glarea), "expose_event",
		       GTK_SIGNAL_FUNC(glarea_draw), NULL);
    g_signal_connect(GTK_OBJECT(glarea), "configure_event",
		       GTK_SIGNAL_FUNC(glarea_reshape), NULL);
    g_signal_connect(GTK_OBJECT(glarea), "realize",
		       GTK_SIGNAL_FUNC(glarea_init), NULL);
    g_signal_connect(GTK_OBJECT(glarea), "destroy",
		       GTK_SIGNAL_FUNC(glarea_destroy), NULL);
/*
  gtk_signal_connect(GTK_OBJECT(glarea), "key_press_event",
  GTK_SIGNAL_FUNC(key_press_event), NULL);
  gtk_signal_connect(GTK_OBJECT(glarea), "key_release_event",
  GTK_SIGNAL_FUNC(key_release_event), NULL);
*/
    gtk_widget_set_usize(GTK_WIDGET(glarea), MWidth, MHeight);

    return(glarea);
}


void set_target(client_t *obj)
{
    RMnode *nn=NULL;
    client_t *tobj;
    RMcamera3D *c;
    RMprimitive *prim;
    RMvertex3D v;

    if (targetnode != NULL) {
	/* we need to free nn if we were using it */
	nn = rmNodeGetIthChild(targetnode, 1);
	if (nn != NULL) {
	    rmNodeRemoveChild(targetnode, nn);
	    tobj = (client_t *)nn->clientData;
	    if (tobj != NULL) {
		if (tobj->anim != NULL)
		    rmNodeDelete(nn);
		if (tobj->gmod != NULL)
		    tobj->gmod->refcount--;
	    }
	}
    } else {
	targetnode = rmNodeNew("target", RM_RENDERPASS_3D,
			       RM_RENDERPASS_ALL);
    }
    if (obj->gmod != NULL) {
	rmNodeAddChild(targetnode, obj->gmod->node);
	obj->gmod->refcount++;
	target_needs_redraw++;
    } else if (obj->anim != NULL) {
	nn = rmNodeNew("anim", RM_RENDERPASS_3D, RM_RENDERPASS_OPAQUE);
	prim = rmPrimitiveNew(RM_USERDEFINED_PRIM);
	rmPrimitiveSetRenderFunc(prim, rmCal3D);
	/*obj->anim = create_calmodel(obj->model);*/
	prim->clientData = (void *)obj->anim;
	rmPrimitiveSetBoundingBox(prim, &obj->anim->coremodel->bmin,
				  &obj->anim->coremodel->bmax);
	rmNodeSetBoundingBox(nn, &obj->anim->coremodel->bmin,
			     &obj->anim->coremodel->bmax);
	v.x = v.y = v.z = 0.0;
	rmPrimitiveSetVertex3D(prim, 1, &v, RM_COPY_DATA, NULL);
	rmNodeAddPrimitive(nn, prim);
	rmNodeUnionAllBoxes(nn);
	rmNodeComputeCenterFromBoundingBox(nn);
	rmNodeAddChild(targetnode, nn);
	target_needs_redraw++;
    }
    v.x = v.y = v.z = 0.0;
    rmNodeSetBoundingBox(targetnode, &v, &v);
    rmNodeUnionAllBoxes(targetnode);
    c = rmCamera3DNew();
    rmCamera3DComputeViewFromGeometry(c, targetnode, TWidth, THeight);
    c->eye.z = c->eye.z/2.0;
    if (obj->anim != NULL) {
	rmNodeGetCenter(nn, &v);
	rmCamera3DSetAt(c, &v);
	c->at.y += 2.0;
    }
    rmNodeSetSceneCamera3D(targetnode, c);
    rmCamera3DDelete(c);
    pickedobj = obj;
    /*pickedobjnum = obj->num;*/
}

gint glarea_button_press(GtkWidget* widget, GdkEventButton* event)
{
   int x = event->x;
   int y = event->y;
   RMpick *pickResults;
   RMpick *pr;
   client_t *obj;
   char *pname;
   int numret, i;

   if (event->button == 1) {
       /* Mouse button 1 was engaged */
       g_print("Button 1 press   (%d, %d)\n", x, y);
       /*pickResults = rmFramePick(mainpipe, rmRootNode(), x, y);*/
       /*pickednode = rmPickedNode(pickResults);*/

       /* clear the old pickedobj and num */
       pickedobj = NULL;
       /*pickedobjnum = -1;*/

       numret = rmFramePickList(mainpipe, rootnode, x, y, &pr);
       for (i=0; i<numret; i++) {
	   pickResults = pr+i;
	   pickednode = rmPickedNode(pickResults);
	   pname = rmNodeGetName(pickednode);
	   if (strncmp(pname, "ground", 6) == 0 ||
	       strncmp(pname, "mesh", 4) == 0||
	       strcmp(pname, "rmlib_root") == 0) {
	       free(pname);
	       continue;
	   }
	   if (pickednode->clientData == NULL) {
	       free(pname);
	       continue;
	   }
	   /*printf("Node Picked: %s Prim #%d zval=%f\n", pname, 
		  rmPickedPrimitive(pickResults),
		  rmPickedPrimitiveZval(pickResults));*/
	   free(pname);
	   /* we got one, now target that mutha */
	   obj = (client_t *)pickednode->clientData;
	   if (obj != NULL) {
	       set_target(obj);
	       break;
	   }
       }
       /*rmPickDelete(pickResults);*/
       rmPickDelete(pr);

       return(TRUE);
   }

    if (event->button == 2) {

	/* Mouse button 2 was engaged */
	g_print("Button 2 press   (%d, %d)\n", x, y);

	return(TRUE);

    }

    return(FALSE);
}

gint glarea_target_draw(GtkWidget* widget, GdkEventExpose* event)
{
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);

    /* Draw only on the last expose event. */
    if (event->count > 0) {
	return(TRUE);
    }
    /*g_print("Target Expose Event\n");*/

    if (gdk_gl_drawable_gl_begin (gldrawable, glcontext)) {
	/* Clear the drawing color buffer and depth buffers */
	/* before drawing.                                  */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
	render_screen(targetpipe, targetnode);
	target_needs_redraw++;
	gdk_gl_drawable_swap_buffers(gldrawable);
	gdk_gl_drawable_gl_end(gldrawable);
    }
    return(TRUE);
}

void forced_redraw(void)
{
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(glarea);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(glarea);

    if (gdk_gl_drawable_gl_begin (gldrawable, glcontext)) {
	/*printf("Forced redraw\n");*/
    
	render_screen(mainpipe, rootnode);
	if (!needs_redraw)
	    needs_redraw = 2;
	gdk_gl_drawable_swap_buffers(gldrawable);
	gdk_gl_drawable_gl_end(gldrawable);
    }
    /*printf("done\n");*/
}


gint glarea_draw(GtkWidget* widget, GdkEventExpose* event)
{
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);

    /* Draw only on the last expose event. */
    if (event->count > 0) {
	return(TRUE);
    }
    /*g_print("Expose Event\n");*/

    if (gdk_gl_drawable_gl_begin (gldrawable, glcontext)) {
	/* Clear the drawing color buffer and depth buffers */
	/* before drawing.                                  */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
	render_screen(mainpipe, rootnode);
	if (!needs_redraw)
	    needs_redraw = 2;
	gdk_gl_drawable_swap_buffers(gldrawable);
	gdk_gl_drawable_gl_end(gldrawable);
    }
    return(TRUE);
}

gint glarea_target_reshape(GtkWidget* widget, GdkEventConfigure* event)
{
    RMcamera3D *c3=NULL;
    int w = widget->allocation.width;
    int h = widget->allocation.height;
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);

    /*g_print("Target Reshape Event\n");*/

    if (targetpipe == NULL || targetnode == NULL)
	return(TRUE);

    if (gdk_gl_drawable_gl_begin (gldrawable, glcontext)) {
	rmPipeSetWindowSize(targetpipe, w, h);
	if ((rmNodeGetSceneCamera3D(targetnode, &c3)) != RM_WHACKED) {
	    rmCamera3DSetAspectRatio(c3, ((float)(w) / (float)(h)));
	    rmNodeSetSceneCamera3D(targetnode, c3);
	} else {
	    g_print("Borked camera in resize\n");
	}
	if (c3 != NULL)
	    rmCamera3DDelete(c3);
	render_screen(targetpipe, targetnode);
	gdk_gl_drawable_swap_buffers(gldrawable);
	gdk_gl_drawable_gl_end(gldrawable);
    }
    return(TRUE);
}

gint glarea_reshape(GtkWidget* widget, GdkEventConfigure* event)
{
    RMcamera3D *c3=NULL;
    int w = widget->allocation.width;
    int h = widget->allocation.height;
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);

    /*g_print("Reshape Event\n");*/

    if (mainpipe == NULL || rootnode == NULL)
	return(TRUE);

    if (gdk_gl_drawable_gl_begin (gldrawable, glcontext)) {
	rmPipeSetWindowSize(mainpipe, w, h);
	if ((rmNodeGetSceneCamera3D(rootnode, &c3)) != RM_WHACKED) {
	    rmCamera3DSetAspectRatio(c3, ((float)(w) / (float)(h)));
	    rmNodeSetSceneCamera3D(rootnode, c3);
	} else {
	    g_print("Borked camera in resize\n");
	}
	if (c3 != NULL)
	    rmCamera3DDelete(c3);
	render_screen(mainpipe, rootnode);
	gdk_gl_drawable_swap_buffers(gldrawable);
	gdk_gl_drawable_gl_end(gldrawable);
    }
    return(TRUE);
}

/* XXX TAKE THIS OUT */

void addLighting(RMnode *toModify)
{
    RMlight *l0;

    /*
     * light source is a directional light pointing from the +Z direction
     */
    RMcolor4D diffuse = { .7,.7,.7,1.0};
    RMcolor4D specular = { .5,0.5,0.5,1.};
/*    RMcolor4D diffuse2 = { .3, 0.3, 0.5, 1.0 };*/
    RMvertex3D direction = { 0.2, 1.0, 0.2}; 

    l0 = rmLightNew();

    rmLightSetType(l0, RM_LIGHT_DIRECTIONAL); 
    rmLightSetColor(l0, &diffuse, &diffuse, &specular); 
    rmLightSetXYZ(l0, &direction);

    rmNodeSetSceneLight(toModify, RM_LIGHT0, l0);

    rmLightDelete(l0);

    /* then set the light model parms. */
    {
        RMcolor4D defAmbient =  { .5,.5,.5,1.0};
        RMlightModel *lm;

        lm = rmLightModelNew();
        rmLightModelSetAmbient(lm, &defAmbient);

        rmLightModelSetTwoSided(lm, RM_FALSE);
        rmLightModelSetLocalViewer(lm, RM_TRUE);
        
        rmNodeSetSceneLightModel(toModify, lm); 

        rmLightModelDelete(lm);
    }
}

RMnode *makeBase(void)
{
    RMvertex3D v;
    char *strings[1];
    RMprimitive *p;
    RMnode *n;

    strings[0] = "No Target";

    n = rmNodeNew(strings[0], RM_RENDERPASS_2D, RM_RENDERPASS_OPAQUE);
    p = rmPrimitiveNew(RM_TEXT);
    v.x = v.y = v.z = 0.0;
    rmPrimitiveSetVertex3D(p, 1, &v, RM_COPY_DATA, NULL);
    rmPrimitiveSetText(p, 1, strings);
    rmNodeSetTranslateVector(n, &v);
    rmNodeAddPrimitive(n, p);
    return(n);
}

/* XXX end of stuff to remove */

gint glarea_target_init(GtkWidget* widget)
{
    Window myWindow;
    RMcolor4D bg={0.1,0.1,0.1,1.0};
    RMcamera3D *c;
    RMnode *text, *bgnode;
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);

    /*g_print("Target Realize Event\n");*/

    if (gdk_gl_drawable_gl_begin(gldrawable, glcontext)) {
	myWindow = GDK_WINDOW_XWINDOW(widget->window);
	rmPipeSetWindow(targetpipe, myWindow, TWidth, THeight);
	rmPipeSetContext(targetpipe, (GLXContext)
			 gdk_x11_gl_context_get_glxcontext(glcontext));

	targetnode = rmNodeNew("targetnode", RM_RENDERPASS_ALL,
			     RM_RENDERPASS_ALL);
	bgnode = rmNodeNew("background", RM_RENDERPASS_3D,
			     RM_RENDERPASS_OPAQUE);
	rmNodeSetSceneBackgroundColor(bgnode, &bg);

	text = makeBase();
	rmNodeAddChild(targetnode, bgnode);
	rmNodeAddChild(targetnode, text);

	addLighting(targetnode); /* XXX */
	c = rmCamera3DNew();
	rmDefaultCamera3D(c);
	rmNodeSetSceneCamera3D(targetnode, c);
	rmCamera3DDelete(c);

	rmNodeSetTraversalMaskDims(targetnode, RM_RENDERPASS_ALL);
	rmNodeSetTraversalMaskOpacity(targetnode, RM_RENDERPASS_ALL);
	gdk_gl_drawable_swap_buffers(gldrawable);
	gdk_gl_drawable_gl_end(gldrawable);
     }

    return(TRUE);
}

gint glarea_init(GtkWidget* widget)
{
    Window myWindow;
    RMcamera3D *c;
    RMcolor4D bg={0.1,0.88,0.88,1.0};
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);

    /*g_print("Realize Event\n");*/

    /* gtk_gl_area_make_current MUST be called before rendering */
    /* into the GtkGLArea.                                      */

    if (gdk_gl_drawable_gl_begin(gldrawable, glcontext)) {
	myWindow = GDK_WINDOW_XWINDOW(widget->window);
	rmPipeSetWindow(mainpipe, myWindow, MWidth, MHeight);
	rmPipeSetContext(mainpipe, (GLXContext)
			 gdk_x11_gl_context_get_glxcontext(glcontext));

	rootnode = rmNodeNew("rootnode", RM_RENDERPASS_ALL,
			     RM_RENDERPASS_ALL);
	bgnode = rmNodeNew("bgnode", RM_RENDERPASS_3D,
			     RM_RENDERPASS_OPAQUE);
	rmNodeSetSceneBackgroundColor(bgnode, &bg);
	transnode = rmNodeNew("transnode", RM_RENDERPASS_ALL,
			      RM_RENDERPASS_ALL);
	objnode = rmNodeNew("objnode", RM_RENDERPASS_ALL,
			     RM_RENDERPASS_OPAQUE);
	groundnode = rmNodeNew("groundnode", RM_RENDERPASS_3D,
			     RM_RENDERPASS_OPAQUE);
	tobjnode = rmNodeNew("tobjnode", RM_RENDERPASS_ALL,
			      RM_RENDERPASS_ALL);

	rmNodeAddChild(rootnode, bgnode);
	rmNodeAddChild(rootnode, groundnode);
	rmNodeAddChild(rootnode, objnode);
	rmNodeAddChild(rootnode, transnode);
	rmNodeAddChild(transnode, tobjnode);

	c = rmCamera3DNew();
	rmDefaultCamera3D(c);
	rmNodeSetSceneCamera3D(rootnode, c);
	rmCamera3DDelete(c);

 	rmNodeUnionAllBoxes(rootnode);
	gdk_gl_drawable_gl_end(gldrawable);
	rmTimeCurrent(&curtime);
     }

    return(TRUE);

}

/* doing a pipeclose causes a BadWindow event.  I have no idea why, but
   I'll leave it alone for right now */

gint glarea_target_destroy(GtkWidget* widget)
{
#if 0
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);

    if (gdk_gl_drawable_gl_begin(gldrawable, glcontext)) {
	g_print("GTK GL Area Target Destroy Event\n");
	rmPipeClose(targetpipe);
	/* rmPipeDelete(mainpipe);*/
	gdk_gl_drawable_gl_end(gldrawable);
    }
#endif
    return(TRUE);

}

gint glarea_destroy(GtkWidget* widget)
{

    g_print("GTK GL Area Destroy Event\n");
    /*rmPipeClose(mainpipe);*/
    /* rmPipeDelete(mainpipe);*/
    rmFinish();

    return(TRUE);

}
