/* $Id: defines.h,v 1.22 2004/03/31 03:45:46 garbled Exp $ */

#ifndef _DEFINES_H_
#define _DEFINES_H_

/**
   \file defines.h
   \author Tim Rightnour
   \brief Global definitions
*/

/**
   \brief object stats (dex, wis, int)
*/
typedef struct _stat_t {
    uint8_t base;	/**< \brief Eg, 18 */
    uint8_t pct;	/**< \brief 100, as in 18/100 */
} stat_t;

/**
   \brief A vertice
*/
typedef struct _vertex_t {
    float x; /**< \brief yaw */
    float y; /**< \brief pitch */
    float z; /**< \brief roll */
} vertex_t;

/**
   \brief object stats, dex, wis, hp, etc.
*/
typedef struct _stats_t {
    int sex;	/**< \brief 0 it 1 M 2 F */
    int type;	/**< \brief type of object */
    int subtype;	/**< \brief subtype of obj */
    int weight;	/**< \brief weight grams */
    int level;	/**< \brief level (1-??) */
    int exp;	/**< \brief exp */
    int material;	/**< \brief material makeup (will make a list) */
    int value;	/**< \brief value of obj (in lowest currency) */
    stat_t Str, Dex, Con, Wis, Int, Per; /**< \brief duh */
    int fame;	/**< \brief fame (freerange) */
    int appr;	/**< \brief appearance (freerange) */ 
    int hp,maxhp;	/**< \brief hp */
    int sp,maxsp;	/**< \brief sp */
    int stm,maxstm;	/**< \brief stamina */
    int food,water;	/**< \brief food/water of 1000 */
    int luck;	/**< \brief luck -100/100 */
    int ac;		/**< \brief Armor Class -100/100 (NOT AD&D) */
    int speed;	/**< \brief speed */
} stats_t;

/**
   \brief Equipment data
*/
typedef struct _equip_t {
    int flags[2];   	/**< \brief equip location flag */
    int loc;		/**< \brief welding location for client */
    int visible;	/**< \brief visible to others? */
} equip_t;

/* defines uses in parsing functions */

#define PO_ARCH	0 /**< \brief load object normally */
#define PO_MAP	1 /**< \brief overlay load of object over arch, used in maps */

#define DO_NOREC 	0 /**< \brief don't recurse in duplicate_object */
#define DO_RECURSE	1 /**< \brief recurse */

#define NO_LOCK	0	/**< \brief Don't lock */
#define DO_LOCK 1	/**< \brief Lock */

#define NO_DRAW 0	/**< \brief Don't draw */
#define DO_DRAW 1	/**< \brief Draw */

#define NO_EXTENDED 0	/**< \brief Don't send extended data */
#define DO_EXTENDED 1	/**< \brief Send extended data */

#define BUCKET_SPELL	0	/**< \brief Spell inventory bucket */
#define BUCKET_JOB	1	/**< \brief Job inventory bucket */
#define BUCKET_INV	2	/**< \brief inventory bucket */
#define BUCKET_SKILL	3	/**< \brief Skill inventory bucket */
#define BUCKET_KEY	4	/**< \brief Key item inventory bucket */
#define BUCKET_EQUIP	5	/**< \brief Equipment inventory bucket */

/* Type/SubType tables */

#define T_PLAYER	1
	#define ST_HUMAN	1
	#define ST_ELF		2
#define T_MONSTER	2
	#define ST_BLOB		1
	#define ST_BIPED	2
	#define ST_QUADRU	3
	#define ST_DRAGON	4
	#define ST_SNAKE	5
	#define ST_INSECT	6
#define T_WEAPON	3
	#define ST_SWORD	1
	#define ST_CLUB		2
	#define ST_DAGGER	3
	#define ST_AXE		4
	#define ST_SPEAR	5
	#define ST_ARROW	6
	#define ST_BOW		7
#define T_PLANT_EVERGREEN	4
#define T_PLANT_DECIDUOUS	5
    #define ST_TREE_CONIF	1
	#define ST_TREE_LEAFY	2
	#define ST_WEED		3
	#define ST_SHRUB	4
	#define ST_MUSHROOM	5
#define T_ARMOR		6
	#define ST_BODY		1
	#define ST_HEAD		2
	#define ST_SHOULDER 	3
	#define ST_ARM		4
	#define ST_GLOVES	5
	#define ST_BELT		6
	#define ST_ULEG		7
	#define ST_LLEG		8
	#define ST_SHOES	9
	#define ST_SHIELD	10
#define T_FURNITURE	7
	#define ST_FIREPLACE	1
	#define ST_CHAIR	2
	#define ST_TABLE	3
	#define ST_BOOKSHELF	4
	#define ST_COATRACK	5
	#define ST_GRAVESTONE	6
	#define ST_EXECUTION	7
	#define ST_SIGN		8
	#define ST_PAINTING	9
	#define ST_STATUE	10
#define T_FOOD		8
	#define ST_WHITE_MEAT	1
	#define ST_DARK_MEAT	2
	#define ST_RED_MEAT	3
	#define ST_VEGETABLE	4
	#define ST_GRAIN	5
	#define ST_FRUIT	6
	#define ST_NUT		7
	#define ST_CANDY	8
	#define ST_DAIRY	9
#define T_DRINK		9
	#define ST_WATER	1
	#define ST_MILK		2
	#define ST_FRUITJUICE	3
	#define ST_BEER		4
	#define ST_WINE		5
	#define ST_LIQUOR	6
	#define ST_BLOOD	7
	#define ST_SALTWATER	8
#define T_STRUCTURE	10
	#define ST_BRIDGE	1
	#define ST_HOUSE	2
	#define ST_TEMPLE	3
	#define ST_CRYPT	4
	#define ST_TOWER	5
	#define ST_FOUNTAIN	6
	#define ST_PARK		7
	#define ST_PIER		8
	#define ST_WALL		9
	#define ST_WELL		10

/* etc..  you get the idea */


/* Flags.. ala crossfire:
 * flags[0] is 0 to 31
 * flags[1] is 32 to 63
 * flags[2] is 64 to 95
 * flags[3] is 96 to 127
 You *must* use the macros below taken from Crossfire.
*/

/**
   \brief Set a flag
   \param xyz structure to set flag on
   \param p flag to set
*/
#define SET_FLAG(xyz, p) \
        ((xyz)->flags[p/32] |= (1U << (p % 32)))
/**
   \brief Clear a flag
   \param xyz structure to set flag on
   \param p flag to clear
*/
#define CLEAR_FLAG(xyz, p) \
        ((xyz)->flags[p/32] &= ~(1U << (p % 32)))
/**
   \brief Check a flag
   \param xyz structure to set flag on
   \param p flag to check
*/
#define QUERY_FLAG(xyz, p) \
        ((xyz)->flags[p/32] & (1U << (p % 32)))

#define FLAG_ALIVE	0 /**< \brief object is a living thing */
#define FLAG_NO_PICK	1 /**< \brief object cannot be picked up */
#define FLAG_AGGR	2 /**< \brief monster is aggressive */
#define FLAG_META_AGGR	3 /**< \brief monster is very aggressive */
#define FLAG_CURSED	4 /**< \brief object is cursed */
#define FLAG_BLESSED	5 /**< \brief object is blessed */
#define FLAG_BLIND	6 /**< \brief monster is blind */
#define FLAG_IMMOBILE	7 /**< \brief object cannot move on it's own */
#define FLAG_NOCOLLIDE	8 /**< \brief object is not eligible for collision */
#define FLAG_NOTIP	9 /**< \brief object will not tip over */
#define FLAG_ROLL	10 /**< \brief object can roll in any direction */
#define FLAG_ROLLX	11 /**< \brief object can roll in X plane */
#define FLAG_ROLLZ	12 /**< \brief object can roll in Z plane */
#define FLAG_ANCHOR	13 /**< \brief object will not slide or rise to surf */
#define FLAG_ANTIGRAV	14 /**< \brief object is immune to gravity */
#define FLAG_STABLE_ROLL 15 /**< \brief object has ceased rolling */
#define FLAG_STABLE_TIP	16 /**< \brief object has ceased tipping */
#define FLAG_STABLE_SLIDE 17 /**< \brief object has ceased sliding */
#define FLAG_DESTROYED	18	/**< \brief object is pending deletion */

/* Values can go up to 127 before the size of the flags array in the
 * object structure needs to be enlarged.
 */

/* states */

#define STATE_IDLE	0	/**< \brief obj is idle */
#define STATE_SLEEP	1	/**< \brief obj is asleep */
#define STATE_REST	2	/**< \brief obj is resting */
#define STATE_DEAD	3	/**< \brief obj is dead */
#define STATE_WALK	4	/**< \brief obj is walking */
#define STATE_RUN	5	/**< \brief obj is running */
#define STATE_SPRINT	6	/**< \brief obj is sprinting */
#define STATE_FALLING	7	/**< \brief obj is falling */

/**
   \brief Set an equipment flag
   \param xyz obj->cli.equip
   \param p flag to set
*/
#define SET_EFLAG(xyz, p) \
        ((xyz).flags[p/32] |= (1U << (p % 32)))
/**
   \brief Clear an equipment flag
   \param xyz obj->cli.equip
   \param p flag to clear
*/
#define CLEAR_EFLAG(xyz, p) \
        ((xyz).flags[p/32] &= ~(1U << (p % 32)))
/**
   \brief Check an equipment flag
   \param xyz obj->cli.equip
   \param p flag to check
*/
#define QUERY_EFLAG(xyz, p) \
        ((xyz).flags[p/32] & (1U << (p % 32)))


/* equipment flags */
#define EQ_UNEQ		0  /**< \brief Not equipped */
#define EQ_HEAD		1  /**< \brief Head */
#define EQ_NECK		2  /**< \brief Neck */
#define EQ_BODY		3  /**< \brief Body */
#define EQ_RSHOULDER	4  /**< \brief Right Shoulder */
#define EQ_RARM		5  /**< \brief Right Arm */
#define EQ_RFARM       	6  /**< \brief Right Forearm */
#define EQ_RHAND       	7  /**< \brief Right Hand */
#define EQ_WRHAND      	8  /**< \brief Right Hand Weapon */
#define EQ_RRING       	9  /**< \brief Right Ring Finger */
#define EQ_LSHOULDER	10 /**< \brief Left Shoulder */
#define EQ_LARM		11 /**< \brief Left Arm */
#define EQ_LFARM       	12 /**< \brief Left Forearm */
#define EQ_LHAND       	13 /**< \brief Left Hand */
#define EQ_WLHAND      	14 /**< \brief Left Hand Weapon */
#define EQ_LRING       	15 /**< \brief Left Ring Finger */
#define EQ_WAIST       	16 /**< \brief Waist */
#define EQ_ULEGS       	17 /**< \brief Upper Legs */
#define EQ_LLEGS       	18 /**< \brief Lower Legs */
#define EQ_FEET		19 /**< \brief Feet */


/* animation tags */

#define ANIM_IDLE	0  /**< \brief Idle */
#define ANIM_WALK	1  /**< \brief Walking */
#define ANIM_RUN	2  /**< \brief Running */
#define ANIM_SPRINT	3  /**< \brief Sprinting */
#define ANIM_SNEAK	4  /**< \brief Sneaking */

#define ANIM_WAVE	20 /**< \brief Waving hand */
#define ANIM_SHARROW	21 /**< \brief Shoot an arrow via bow */

/* map scaling factors */

#define SCALE_XZ 10.0	 	/**< \brief X and Z scale to map */
#define SCALE_Y 2.0	 	/**< \brief Y scale to mapheight */
#define POINTS_PER_METER 4	/**< \brief Scaled points per meter */


/* directory defines */

#define ARCH_SUFX	"/arch"	  /**< \brief Archetype directory */
#define ETC_SUFX       	"/etc"	  /**< \brief Etc directory */
#define MAP_SUFX       	"/map"	  /**< \brief Mapfile Directory */
#define TEX_SUFX	"/textures"	  /**< \brief Textures Directory */
#define VAR_SUFX	"/var"	  /**< \brief Var Directory */
#define LIB_SUFX	"/libdata" /**< \brief libdata directory */

#define MAX_PATH	2048	  /**< \brief Maximum path size */
/* pthread_self */
#ifdef LOCK_DEBUG
#define pthread_mutex_lock(mutex) \
 {\
    printf("Lock: %s:%d %p %x\n", __FILE__, __LINE__, mutex,(int)pthread_self()); \
    fflush(NULL); \
    pthread_mutex_lock(mutex); \
    printf("LockRun: %s:%d %p %x\n", __FILE__, __LINE__, mutex,(int)pthread_self()); \
    fflush(NULL); \
 }

#define pthread_mutex_unlock(mutex) \
  {\
    printf("Unlock: %s:%d %p %x\n", __FILE__, __LINE__, mutex,(int)pthread_self()); \
    fflush(NULL); \
    pthread_mutex_unlock(mutex); \
  }
#endif

#endif /* _DEFINES_H_ */
