/* $Id: netparser.h,v 1.11 2004/03/30 06:45:55 garbled Exp $ */

#ifndef _NETPARSER_H_
#define _NETPARSER_H_
#include "uid.h"
/**
   \file netparser.h
   \author Tim Rightnour
   \brief Network parsing definitions
*/

/**
   \brief Network command argument record
*/

typedef struct _pargs_t {
    int cword;		/**< \brief command word */
    int type;		/**< \brief data type, see PTDOUBLE etc */
    union {
        double d;	/**< \brief PTDOUBLE */
	float f;	/**< \brief PTFLOAT */
        char *c;	/**< \brief PTCHAR */
        int i;		/**< \brief PTINT */
	long l;		/**< \brief PTLONG */
	stat_t s;	/**< \brief PTSTAT */
	uniqid_t uid; /** \brief PTUID */
    } arg;		/**< \brief argument data */
} pargs_t;

/**
   \brief Argument record entry
*/
typedef struct _argtable_t {
    char *name;	/**< \brief Argument keyword */
    int num;	/**< \brief argument subcommand number */
    int type;	/**< \brief data type see PTDOUBLE etc */
} argtable_t;

#define PTDOUBLE	1
#define PTCHAR		2
#define PTINT		3
#define PTSTAT		4
#define PTLONG		5
#define PTUID		6
#define PTFLOAT		7

/* subcommands */

#define ID		1
#define NROF	       	2
#define DRAW		3
#define STATE		4
#define NAME		5
#define NAMEPL		6
#define MODEL		7
#define FACEX		8
#define FACEY		9
#define FACEZ		10
#define DIRX		11
#define DIRY		12
#define DIRZ		13
#define COORDX		14
#define COORDY		15
#define COORDZ		16
#define SCALEX		17
#define SCALEY		18
#define SCALEZ		19
#define SEX		20
#define TYPE		21
#define SUBTYPE		22
#define WEIGHT		23
#define LEVEL		24
#define EXP		25
#define MATERIAL	26
#define VALUE		27
#define STR		28
#define DEX		29
#define CON		30
#define WIS		31
#define INT		32
#define PER		33
#define FAME		34
#define APPEAR		35
#define HP		36
#define MAXHP		37
#define SP		38
#define MAXSP		39
#define STAMINA		40
#define MAXSTAMINA	41
#define FOOD		42
#define WATER		43
#define AC		44
#define SPEED		45
#define ACT		46
#define CONTAINER	47
#define BUCKET		48
#define EQUIP		49
#define SLOT		50
#define TARGET		51
#define VELOCITY	52
#define USERNAME	53
#define PASSWORD	54
#define PLAYER_EXISTS	55
#define LANGUAGE	56
#define DATA		57
#define VOLUME		58
#define SRVID		59
#define IPADDR		60
#define PORTNUM		61
#define DESC		62
#define MAPINFO		63
#define HOSTNAME	64
#define SRVVERSION     	65
#define WEBPAGE		66
#define OWNERNAME	67
#define OWNEREMAIL	68
#define LASTUPDATE	69
#define FIRSTUPDATE	70
#define NUMUSERS	71
#define USERHWM		72
#define MAXUSERS	73
#define PREFERIP	74
#define WIDTH		75
#define HEIGHT		76
#define BINARY		77
#define MAPID		78
#define SUBCOMMAX 	79 /* keep this last */

void init_argcomm(void);
pargs_t *parse_command(char **words, int *command, int count);
char **parse_netcommand(char *buf, int *arg_count);
int find_arg_by_id(pargs_t *args, int id);

#endif /* _NETPARSER_H_ */
