/* $Id: uid.h,v 1.5 2004/03/19 10:23:08 xb49 Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Christophe Antoine
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/
/*
* $Log: uid.h,v $
* Revision 1.5  2004/03/19 10:23:08  xb49
* uid_printf now return a char *
*
* Revision 1.4  2004/03/11 11:28:29  xb49
* added uid_isnull , uid_clear
*
* Revision 1.3  2004/03/11 09:03:26  xb49
* added uid_eq
*
* Revision 1.2  2004/03/10 20:24:02  garbled
* Change uid_t to uniqid_t so we don't collide with unix namespace.
* Minor cleanups in these files:  Spaces after commas, no ; after }, space
* between type and * in prototype, and break functions into multiple lines.
*
* Also.. Change copyright to Christophe Antoine, not me.  I didn't write
* these files.
*
* Revision 1.1  2004/03/10 15:19:39  xb49
* adding files for uid (uniq id)
*
*/
#ifndef __uid_h__
#define __uid_h__
#include <stdio.h>

typedef unsigned long uint32;

/** Uid_t is an uniq id */
typedef uint32 uniqid_t[4];

/** compare two uniqid (like strcmp) */
int uid_cmp(uniqid_t*id1, uniqid_t*id2);
/** compare two uniqid  
@return false if not equal
*/
int uid_eq(uniqid_t*id1, uniqid_t*id2);

/** copy an uid */
void uid_cpy(uniqid_t * dst, uniqid_t *src);

/** increment an uid */
void uid_inc(uniqid_t*id);

/** @brief test if an uid is null */
int uid_isnull(uniqid_t *id);
/** @brief set  an uid to null */
void uid_clear(uniqid_t *id) ;



/** print an uid */
void uid_print(uniqid_t*);
/** print an uid to a file*/
void uid_fprint(FILE*f, uniqid_t*);
/** print an uid to string (must be size >=36)*/
char * uid_sprint(char *s, uniqid_t*);


/** read an uid from a string 
@return 0 if fail*/
int uid_sread(char *s, uniqid_t*i);

/** read an uid from a file  
@return 0 if fail*/
int uid_fread(FILE *f, uniqid_t*i);


#endif
