/* $Id: utils.h,v 1.12 2004/03/30 06:45:55 garbled Exp $ */

#ifndef _UTILS_H_
#define _UTILS_H_

/**
   \file utils.h
   \author Tim Rightnour
   \brief Utilitiy definitions
*/

/**
   \brief A quick and simple safer malloc
*/
#define smalloc(foo) _safer_malloc(sizeof(foo), __FILE__, __LINE__)

/** \brief a macro interface to safer_malloc */
#define safer_malloc(foo) _safer_malloc(foo, __FILE__, __LINE__)

/** \brief a macro interface to bailout */
#define bailout() _bailout(__FILE__, __LINE__)

#define RANDOM random

#define LOG_DEBUG	1
#define LOG_WARNING	2
#define LOG_ERROR	3
#define LOG_FATAL	4

/**
   \brief Convert degrees to radians
   \param a degrees (should probably be a float or double)
*/
#define DEGREES_TO_RADIANS(a) ((a)*0.017453292)
/**
   \brief Convert radians to degrees
   \param a radians (should probably be a float or double)
*/
#define RADIANS_TO_DEGREES(a) ((a)*57.29577951)

/**
   \brief A triangle
*/
typedef struct _triangle_t {
    vertex_t side[3]; /**< \brief the vertices of the triangle */
} triangle_t;

/**
   \brief A bounding box
*/
typedef struct _bbox_t {
    vertex_t side[8];	/**< \brief The 8 corner vertices of the box */
    vertex_t cog;	/**< \brief The center of gravity */
    vertex_t center;	/**< \brief The perfect center */
} bbox_t;

/**
   \brief A 1-dimensional motion vector
*/
typedef struct _vector1_t {
    float a;	/**< \brief angle */
    float m;	/**< \brief momentum */
} vector1_t;

void openlog(void);
void closelog(void);
void LOG(int severity, const char * s, ...);
void _bailout(char *file, int line);
void *_safer_malloc(size_t size, char *file, int line);
float groundheight_at_coord(uint8_t **map, vertex_t coord, int h, int w);
float groundheight_at_xz(uint8_t **map, float x, float z, int h, int w);
int pnpoly(int npol, float *xp, float *yp, float x, float y);
float farthest_fall_rotation(int nrof, float *x, float *y, float r);
triangle_t *find_closest_hf_triangle(uint8_t **map, vertex_t coord);
void angle_between(vertex_t src, vertex_t dst, vertex_t *angle);
int find_lowest_point_tri(triangle_t *tri);
int find_highest_point_tri(triangle_t *tri);
void find_center_tri(triangle_t *tri, vertex_t *center);
bbox_t *create_bbox(vertex_t bmin, vertex_t bmax, vertex_t cog,
		    vertex_t coord);
int rndm(int min, int max);
char *print_stat(stat_t stat);
float add_euler(float base, float add);
float avg_euler(float a, float b);
float diff_euler(float a, float b);
float dist_between_squared(vertex_t ca, vertex_t cb);
float dist_between(vertex_t ca, vertex_t cb);
float dist_between_flat(vertex_t ca, vertex_t cb);
vector1_t *vector_1d_cross(float aa, float ma, float ab, float mb);
int same_within(float a, float b, float factor);

#endif /* _UTILS_H_ */
