/* $Id: timers.c,v 1.4 2004/03/30 01:02:24 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <unistd.h>
#include <sys/time.h>
#include <stdio.h>
#include "timers.h"

/**
   \file timers.c
   \author Tim Rightnour
   \brief Timer routines
*/

extern long max_time;
struct timeval last_time;

/**
   \brief Diff two timeval's and return the result
   \param tval1 first timeval
   \param tval2 second timeval
   \return difference
*/
struct timeval tvaldiff(struct timeval *tval1, struct timeval *tval2)
{
    struct timeval rtval, tmp;

    tmp = *tval1;

    if ((rtval.tv_usec = tmp.tv_usec - tval2->tv_usec) < 0)
    {
	rtval.tv_usec += 1000000;
	--tmp.tv_sec;
    }

    if ((rtval.tv_sec = tmp.tv_sec - tval2->tv_sec) < 0)
    {
	rtval.tv_usec = 0;
	rtval.tv_sec = 0;
    }

    return(rtval);
}

/* timer routines taken from Crossfire */

/**
   \brief has enough time passed since the last marked point?
   \return true/false
*/
int enough_time_passed(void)
{
    static struct timeval new_time;
    long elapsed_utime;

    (void)gettimeofday(&new_time, NULL);
    elapsed_utime = (new_time.tv_sec - last_time.tv_sec) * 1000000 +
	new_time.tv_usec - last_time.tv_usec;
    if (elapsed_utime > max_time) {
	last_time.tv_sec = new_time.tv_sec;
	last_time.tv_usec = new_time.tv_usec;
	return 1;
    }
    return 0;
}

/**
   \brief sleep the difference between the amount of time we spent prior to
   the sleep, and the maximum we were alotted for this timeslot.
*/
void sleep_delta(void)
{
    static struct timeval new_time;
    long sleep_sec, sleep_usec;

    (void)gettimeofday(&new_time, NULL);
    sleep_sec = last_time.tv_sec - new_time.tv_sec;
    sleep_usec = max_time - (new_time.tv_usec - last_time.tv_usec);

    /* This is very ugly, but probably the fastest for our use: */
    while (sleep_usec < 0) {
	sleep_usec += 1000000;
	sleep_sec -= 1;
    }
    while (sleep_usec > 1000000) {
	sleep_usec -= 1000000;
	sleep_sec +=1;
    }
    if (sleep_sec >= 0 && sleep_usec > 0) {
	sleep(sleep_sec);
	usleep(sleep_usec);
 	/* printf("SLEPT: %ld.%ld\n", sleep_sec, sleep_usec); */
    }
    (void)gettimeofday(&last_time, NULL);
}

/**
   \brief this function is used to time other functions and report on how
   long they took.  It is used for debugging purposes only.
   \param set 0=start timer, 1=end timer and printf result
*/

void time_action(int set)
{
    static struct timeval begin;
    struct timeval end, diff;

    if (!set) {
	(void)gettimeofday(&begin, NULL);
	return;
    }
    (void)gettimeofday(&end, NULL);
/*    diff = tvaldiff(&begin, &end);*/
    diff = tvaldiff(&end, &begin);
    printf("TOOK: %ld.%0.6ld\n", diff.tv_sec, diff.tv_usec);
}
