/* $Id: uid.c,v 1.5 2004/03/19 10:23:07 xb49 Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Christophe Antoine
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/
/*
* $Log: uid.c,v $
* Revision 1.5  2004/03/19 10:23:07  xb49
* uid_printf now return a char *
*
* Revision 1.4  2004/03/11 11:28:29  xb49
* added uid_isnull , uid_clear
*
* Revision 1.3  2004/03/11 09:03:25  xb49
* added uid_eq
*
* Revision 1.2  2004/03/10 20:24:02  garbled
* Change uid_t to uniqid_t so we don't collide with unix namespace.
* Minor cleanups in these files:  Spaces after commas, no ; after }, space
* between type and * in prototype, and break functions into multiple lines.
*
* Also.. Change copyright to Christophe Antoine, not me.  I didn't write
* these files.
*
* Revision 1.1  2004/03/10 15:19:39  xb49
* adding files for uid (uniq id)
*
*/
#include "uid.h"
#include <string.h>


#define SHORTOUT 

static uniqid_t null_uid = {0,0,0,0};

int uid_isnull(uniqid_t *id) {return uid_eq(id,&null_uid);}
void uid_clear(uniqid_t *id) {uid_cpy(id,&null_uid);}

int uid_cmp(uniqid_t *id1, uniqid_t *id2)
{
    return memcmp(id1, id2, sizeof(uniqid_t));
}
int uid_eq(uniqid_t *id1, uniqid_t *id2)
{
    return !memcmp(id1, id2, sizeof(uniqid_t));
}


void uid_inc(uniqid_t *id)  {
    int i;
    for(i = 3; i>=0; i--) {
	if(++(*id)[i]) break;
    }
}

void uid_cpy(uniqid_t *dst, uniqid_t *src)
{
    memcpy(dst, src, sizeof(uniqid_t));
}



#ifdef SHORTOUT
#define FSTR "%lx.%lx.%lx.%lx" 
#else
#define FSTR "%08lx.%08lx.%08lx.%08lx" 
#endif
/** print an uid */
void uid_print(uniqid_t *id) {
	printf(FSTR, (*id)[0], (*id)[1], (*id)[2], (*id)[3]);
};
/** print an uid to a file*/
void uid_fprint(FILE *f, uniqid_t *id) {
	fprintf(f, FSTR, (*id)[0], (*id)[1], (*id)[2], (*id)[3]);
};
/** print an uid to string (must be size >=36)*/
char * uid_sprint(char *s, uniqid_t *id){
	sprintf(s, FSTR, (*id)[0], (*id)[1], (*id)[2], (*id)[3]);
	return s;
};

#define FSSTR "%lx.%lx.%lx.%lx"

/** read an uid from a string */
int uid_sread(char *s, uniqid_t *id) {
	return (sscanf(s, FSSTR, &((*id)[0]), &((*id)[1]), &((*id)[2]),
		       &((*id)[3])) == 4);
}
int uid_fread(FILE *f, uniqid_t *id) {
	return (fscanf(f, FSSTR, &((*id)[0]), &((*id)[1]), &((*id)[2]),
		       &((*id)[3])) == 4);
}


