/* $Id: gmdl.c,v 1.6 2004/02/09 21:53:41 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <time.h>
#include <stdarg.h>
#include <math.h>
#include <sys/param.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <rm/rmaux.h>

#include "defines.h"
#include "modelloader.h"

int img_width=400,img_height=300;
RMnode *rootnode;
RMpipe *mainpipe;
char *filename;
char *progname;
char *archdir, *etcdir, *mapdir, *texdir;

void init_dirs()
{
    archdir = PACKAGE_DATA_DIR ARCH_SUFX;
    etcdir = PACKAGE_DATA_DIR ETC_SUFX;
    mapdir = PACKAGE_DATA_DIR MAP_SUFX;
    texdir = PACKAGE_DATA_DIR TEX_SUFX;
}

void my_set_scene(RMnode *camNode, RMenum stereo_format)
{
    RMcamera3D *c=rmCamera3DNew();


    rmDefaultCamera3D(c);       /* assign it some default parameters */
    
    /* adjust the "default" camera based upon rootnode's geometry */
    rmCamera3DComputeViewFromGeometry(c, rootnode, img_width, img_height);

    rmNodeSetSceneCamera3D(camNode, c);
    rmCamera3DDelete(c);
  
    /* use RM's default lighting model, apply at rmRootNode() */
    rmDefaultLighting(camNode);
}

void my_build_objs(void)
{
    RMnode *model, *stuff;
    RMnode *bgnode;
    RMcolor4D bg;
    RMvertex3D v;

    rootnode = rmNodeNew("rootnode", RM_RENDERPASS_3D, RM_RENDERPASS_ALL);
    rmNodeComputeBoundingBox(rootnode);
    rmNodeAddChild(rmRootNode(), rootnode); 
    /*
     * make the background black for fun
     */
    bgnode = rmNodeNew("bgnode", RM_RENDERPASS_3D, RM_RENDERPASS_OPAQUE);
    bg.r = bg.g = bg.b = bg.a = 0.0;
    rmNodeSetSceneBackgroundColor(bgnode, &bg);
    rmNodeAddChild(rootnode, bgnode);

    model = rmNodeNew("model", RM_RENDERPASS_3D, RM_RENDERPASS_ALL);
    stuff = rmNodeNew("stuff", RM_RENDERPASS_3D, RM_RENDERPASS_ALL);
    rmNodeAddChild(rootnode, stuff);
    parse_genmodel(model, filename);
    rmNodeAddChild(stuff, model);
    v.x = v.y = v.z = 0.0;
    rmNodeSetTranslateVector(model, &v);
    rmNodeUnionAllBoxes(model);
    rmNodeUnionAllBoxes(stuff);
    rmNodeUnionAllBoxes(rootnode);
    rmNodeUnionAllBoxes(rmRootNode()); 
    rmNodeComputeCenterFromBoundingBox(rmRootNode());
}

void myinitfunc(RMpipe *p, RMnode *n)
{
    my_build_objs();
    my_set_scene(rmRootNode(), rmPipeGetChannelFormat(p));

    rmauxSetGeomTransform(rootnode, p);
    rmauxSetCamera3DTransform(rmRootNode(), p);
    rmauxSetResizeFunc(p, rmRootNode(), rmauxDefaultResizeFunc);
    
    rmFrame(p, rmRootNode());
}

void myrenderfunc(RMpipe *p, RMnode *n)
{
    rmFrame(p, n);
}

int main(int argc, char **argv)
{
    RMenum targetPlatform = RM_PIPE_GLX;
    void *msg;
    RMenum processingMode = RM_PIPE_MULTISTAGE;
    Window w;
    int managed = RM_TRUE;

    if (argc < 2) {
	printf("usage: genmodelviewer genmodel.xgm\n");
	exit(1);
    }

    init_dirs();
    filename = strdup(argv[1]);

    rmInit();
    mainpipe = rmPipeNew(targetPlatform);
    rmPipeSetProcessingMode(mainpipe, processingMode);
    
    w = rmauxCreateXWindow(mainpipe, (Window)NULL,
	0, 0, img_width, img_height,
	"Genecys Model Viewer", "GenModelViewer", managed);

    rmPipeSetWindow(mainpipe, w, img_width, img_height);

    rmauxSetInitFunc(myinitfunc); 
    rmPipeMakeCurrent(mainpipe); 
    rmauxSetRenderFunc(myrenderfunc);
    rmauxSetKeyFunc(mainpipe, rmauxDefaultKeyFunc);
    rmNodeSetTraversalMaskDims(rmRootNode(), RM_RENDERPASS_3D);
    rmNodeSetTraversalMaskOpacity(rmRootNode(), RM_RENDERPASS_ALL);

    rmauxEventLoop(mainpipe, rmRootNode(), &msg);

    rmPipeDelete(mainpipe);
    rmFinish();
    return(1);
}
