/* $Id: modelloader.c,v 1.8 2004/03/16 19:18:47 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <sys/param.h>
#include <inttypes.h>
#include <sys/time.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <rm/rm.h>
#include <rm/rmaux.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "defines.h"
#include "timers.h"
#include "mappings.h"
#include "cmotion.h"
#include "shstr.h"
#include "jpegloader.h"
#include "pngloader.h"
#include "utils.h"
#include "xmlparse.h"
#include "modelloader.h"

TAILQ_HEAD(, _genmaterial_t) genmats = TAILQ_HEAD_INITIALIZER(genmats);

extern RMpipe *mainpipe;
extern char *texdir;

typedef struct _tmpmaterial_t {
    char *name;
    char *image;
    RMcolor4D ambient, diffuse, specular;
    RMtexture *tex;
    int alpha;
} tmpmaterial_t;

void parse_quadfloat(xmlDocPtr doc, xmlNodePtr cur, RMcolor4D *quad)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    sscanf(tmp, "%f %f %f %f", &quad->r, &quad->g, &quad->b, &quad->a);
    free(tmp);
}

tmpmaterial_t *parse_materials(xmlDocPtr doc, xmlNodePtr cur, int *num)
{
    tmpmaterial_t *mats;
    int nrofmats = 0;
    xmlNodePtr ncur;
    genmaterial_t *gmat;
    RMimage *tmpimg;
    RMenum format, type;
    int nh, nw;
    char *tmp, file[MAX_PATH];

    cur = cur->xmlChildrenNode;
    mats = malloc(sizeof(tmpmaterial_t));

    while (cur != NULL) {
        if (!strcmp(cur->name, "material")) {
	    nrofmats++;
	    mats = realloc(mats, sizeof(tmpmaterial_t)*nrofmats);
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "ambient"))
		    parse_quadfloat(doc, ncur, &mats[nrofmats-1].ambient);
		if (!strcmp(ncur->name, "diffuse"))
		    parse_quadfloat(doc, ncur, &mats[nrofmats-1].diffuse);
		if (!strcmp(ncur->name, "specular")) {
		    parse_quadfloat(doc, ncur, &mats[nrofmats-1].specular);
		}
		if (!strcmp(ncur->name, "name")) {
		    mats[nrofmats-1].name =
			xmlNodeListGetString(doc, ncur->xmlChildrenNode, 1);
		}
		if (!strcmp(ncur->name, "image")) {
		    int found=0;

		    mats[nrofmats-1].image =
			xmlNodeListGetString(doc, ncur->xmlChildrenNode, 1);
		    /* look for existing texture */
		    TAILQ_FOREACH(gmat, &genmats, next) {
			if (strcmp(mats[nrofmats-1].name, gmat->name) == 0) {
			    mats[nrofmats-1].tex = gmat->texture;
			    mats[nrofmats-1].alpha = gmat->alpha;
			    found++;
			    break;
			}
		    }
		    if (found)
			continue;
		    /* allocate a new texture */
		    gmat = malloc(sizeof(genmaterial_t));
		    gmat->name = strdup(mats[nrofmats-1].name);
		    gmat->imagefile = strdup(mats[nrofmats-1].image);
		    gmat->alpha = 0;

		    sprintf(file, "%s/%s", texdir, gmat->imagefile);

		    if (strncmp(strrchr(gmat->imagefile, '.'),
				".jpg", 4) == 0) {
			gmat->pixels = readJPEG(file, &gmat->width,
						&gmat->height);
			format = RM_IMAGE_RGB;
		    } else if (strncmp(strrchr(gmat->imagefile, '.'),
				       ".png", 4) == 0) {
			gmat->pixels = parse_pngRGBA(file, &gmat->width,
						     &gmat->height);
			format = RM_IMAGE_RGBA;
			gmat->alpha = 1;
		    }
		    type = RM_UNSIGNED_BYTE;
		    mats[nrofmats-1].alpha = gmat->alpha;
		    tmpimg = rmImageNew(2, gmat->width, gmat->height,
			1, format, type, RM_DONT_COPY_DATA);
		    rmImageSetPixelData(tmpimg, gmat->pixels,
		      RM_DONT_COPY_DATA, free);
		    nw = rmNearestPowerOfTwo(gmat->width);
		    nh = rmNearestPowerOfTwo(gmat->height);
		    if (nw != gmat->width || nh != gmat->height) {
			gmat->width = nw;
			gmat->height = nh;
			gmat->image = rmImageNew(2, gmat->width, gmat->height,
			    1, format, type, RM_COPY_DATA);
			rmImageResize(tmpimg, gmat->image, RM_SOFTWARE,
			    mainpipe);
			gmat->image->copy_flag = RM_DONT_COPY_DATA;
		    } else {
			gmat->image = tmpimg;
		    }
		    gmat->texture = rmTextureNew(2);
		    rmTextureSetImages(gmat->texture, &gmat->image, 1,
				       RM_FALSE);
		    rmTextureSetEnv(gmat->texture, GL_REPLACE, NULL);
		    rmTextureSetWrapMode(gmat->texture, GL_REPEAT);
		    TAILQ_INSERT_TAIL(&genmats, gmat, next);
		    mats[nrofmats-1].tex = gmat->texture;
		}
		ncur = ncur->next;
	    }
	}
	cur = cur->next;
    }
    num = &nrofmats;
    return(mats);
}

void parse_genmodel(RMnode *node, char *filename)
{
    xmlDocPtr doc;
    xmlNodePtr cur, ncur, xcur, ycur;
    RMprimitive *p;
    RMnode *n;
    RMvertex3D *v, *fv, *fn, *nor;
    RMvertex2D *t, *ft;
    tmpmaterial_t *mats;
    int nrofmats, nrofpoints, *gnor;
    char *tmp;

    doc = xmlParseFile(filename);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "genmodel")) {
        fprintf(stderr, 
                "document of the wrong type, root node != genmodel");
        xmlFreeDoc(doc);
        bailout();
    }

    cur = cur->xmlChildrenNode;

    while (cur != NULL) {
	if (!strcmp(cur->name, "name")) {
	    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	    rmNodeSetName(node, tmp);
	    free(tmp);
	}
        if (!strcmp(cur->name, "materials")) {
	    mats = parse_materials(doc, cur, &nrofmats);
	}
	if (!strcmp(cur->name, "surface")) {
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "vertexs")) {
		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL) {
			int nrofverts, cvert;

			if (!strcmp(xcur->name, "verts")) {
			    parseint(doc, xcur, &nrofverts);
			    v = malloc(sizeof(RMvertex3D) * nrofverts);
			    nor = malloc(sizeof(RMvertex3D) * nrofverts);
			    gnor = malloc(sizeof(int) * nrofverts);
			    memset(gnor, 0, sizeof(int) * nrofverts);
			    t = malloc(sizeof(RMvertex2D) * nrofverts);
			    cvert = -1;
			}
			if (!strcmp(xcur->name, "vertex")) {
			    cvert++;
			    ycur = xcur->xmlChildrenNode;
			    while (ycur != NULL) {
				if (!strcmp(ycur->name, "vertex_x"))
				    parsefloat(doc, ycur, &v[cvert].x);
				if (!strcmp(ycur->name, "vertex_y"))
				    parsefloat(doc, ycur, &v[cvert].y);
				if (!strcmp(ycur->name, "vertex_z"))
				    parsefloat(doc, ycur, &v[cvert].z);
				if (!strcmp(ycur->name, "nor_x")) {
				    gnor[cvert]++;
				    parsefloat(doc, ycur, &nor[cvert].x);
				}
				if (!strcmp(ycur->name, "nor_y"))
				    parsefloat(doc, ycur, &nor[cvert].y);
				if (!strcmp(ycur->name, "nor_z"))
				    parsefloat(doc, ycur, &nor[cvert].z);
				if (!strcmp(ycur->name, "tex_u"))
				    parsefloat(doc, ycur, &t[cvert].x);
				if (!strcmp(ycur->name, "tex_v"))
				    parsefloat(doc, ycur, &t[cvert].y);
				ycur = ycur->next;
			    }
			}
			xcur = xcur->next;
		    }
		}
		if (!strcmp(ncur->name, "triangle")) {
		    int type, cface=0, vnum, matnum, newnode=0;
		    char *mname;

		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL) {
			if (!strcmp(xcur->name, "name")) {
			    mname = xmlNodeListGetString(doc,
			        xcur->xmlChildrenNode, 1);
			}
			if (!strcmp(xcur->name, "matnum")) {
			    parseint(doc, xcur, &matnum);
			    n = NULL;
			    if (rmNodeGetNumChildren(node) > matnum)
				n = rmNodeGetIthChild(node, matnum);
			    if (n == NULL) {
				if (mats[matnum].alpha)
				    n = rmNodeNew(mname, RM_RENDERPASS_3D,
					      RM_RENDERPASS_TRANSPARENT);
				else
				    n = rmNodeNew(mname, RM_RENDERPASS_3D,
					      RM_RENDERPASS_OPAQUE);
				newnode++;
			    }
			    free(mname);
			    if (newnode) {
				/*rmNodeSetSceneTexture(n, mats[matnum].tex,
				  RM_DONT_COPY_DATA);*/
				rmNodeSetSceneTexture(n, mats[matnum].tex);
				rmNodeSetAmbientColor(n,&mats[matnum].ambient);
				rmNodeSetDiffuseColor(n,&mats[matnum].diffuse);
				rmNodeSetSpecularColor(n, &mats[matnum].specular);
			    }
			}
			if (!strcmp(xcur->name, "points")) {
			    parseint(doc, xcur, &nrofpoints);
			}
			if (!strcmp(xcur->name, "type")) {
			    tmp = xmlNodeListGetString(doc,
                                xcur->xmlChildrenNode, 1);
			    if (!strcmp(tmp, "strip")) {
				fv = malloc(sizeof(RMvertex3D) * nrofpoints);
				fn = malloc(sizeof(RMvertex3D) * nrofpoints);
				ft = malloc(sizeof(RMvertex2D) * nrofpoints);
				p = rmPrimitiveNew(RM_TRIANGLE_STRIP);
				type = 1;
			    }
			    else if (!strcmp(tmp, "block")) {
				fv = malloc(sizeof(RMvertex3D) * nrofpoints);
				fn = malloc(sizeof(RMvertex3D) * nrofpoints);
				ft = malloc(sizeof(RMvertex2D) * nrofpoints);
				p = rmPrimitiveNew(RM_TRIANGLES);
				type = 2;
			    }
			    free(tmp);
			}
			if (!strcmp(xcur->name, "face")) {
			    if (type == 1) {
				parseint(doc, xcur, &vnum);
				fv[cface] = v[vnum];
				fn[cface] = nor[vnum];
				if (gnor[vnum] == 0) {
				    fn[cface] = v[vnum];
				    (void)rmVertex3DNormalize(&fn[cface]);
				}
				ft[cface] = t[vnum];
				cface++;
			    } else if (type == 2) {
				int v1,v2,v3;

				parsetriint(doc, xcur, &v1, &v2, &v3);
				fv[cface] = v[v1];
				fn[cface] = nor[v1];
				if (gnor[v1] == 0) {
				    fn[cface] = v[v1];
				    (void)rmVertex3DNormalize(&fn[cface]);
				}
				ft[cface] = t[v1];
				cface++;
				fv[cface] = v[v2];
				fn[cface] = nor[v2];
				if (gnor[v2] == 0) {
				    fn[cface] = v[v2];
				    (void)rmVertex3DNormalize(&fn[cface]);
				}
				ft[cface] = t[v2];
				cface++;
				fv[cface] = v[v3];
				fn[cface] = nor[v3];
				if (gnor[v3] == 0) {
				    fn[cface] = v[v3];
				    (void)rmVertex3DNormalize(&fn[cface]);
				}
				ft[cface] = t[v3];
				cface++;
			    }
			}
			xcur = xcur->next;
		    } /* while triangle */
		    rmPrimitiveSetVertex3D(p, cface, fv, RM_COPY_DATA,
					   NULL);
		    rmPrimitiveSetNormal3D(p, cface, fn, RM_COPY_DATA,
					   NULL);
		    rmPrimitiveSetTexcoord2D(p, cface, ft, RM_COPY_DATA,
					     NULL);
		    /*rmPrimitiveComputeBoundingBox(p);*/
		    rmNodeAddPrimitive(n, p);
		    rmNodeComputeBoundingBox(n);
		    if (newnode)
			rmNodeAddChild(node, n);
		} /* triangle */
		ncur = ncur->next;
	    } /* while */
	    free(fn);
	    free(fv);
	    free(ft);
	    free(v);
	    free(nor);
	    free(gnor);
	    free(t);
	} /* surface */
	cur = cur->next;
    }
    free(mats);
}
