/* $Id: lodder.c,v 1.13 2004/03/23 17:45:52 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

/*
   This code is based directly on Lodder.cpp from Cal3D.
   Copyright (C) 2001, 2002 Bruno 'Beosil' Heidelberger
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#if defined(__linux__)
#include <values.h>
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <pthread.h>

#include "defines.h"
#include "shstr.h"
#include "utils.h"
#include "quadric.h"
#include "lodder.h"
#include "xmlparse.h"

extern char *optarg;

char *progname;

void write_material(xmlNodePtr node, Lmat_t *mat)
{
    xmlNodePtr tree;
    char buf[256];

    tree = xmlNewChild(node, NULL, "material", NULL);
    if (mat->name != NULL)
        write_string(tree, "name", mat->name);
    if (mat->image != NULL)
        write_string(tree, "image", mat->image);
    sprintf(buf, "%f %f %f %f", mat->ambient[0], mat->ambient[1],
        mat->ambient[2], mat->ambient[3]);
    write_string(tree, "ambient", buf);
    sprintf(buf, "%f %f %f %f", mat->diffuse[0], mat->diffuse[1],
        mat->diffuse[2], mat->diffuse[3]);
    write_string(tree, "diffuse", buf);
    sprintf(buf, "%f %f %f %f", mat->specular[0], mat->specular[1],
        mat->specular[2], mat->specular[3]);
    write_string(tree, "specular", buf);
    write_float(tree, "shine", mat->shine);
}

void write_vertex(xmlNodePtr node, Lvertex_t *vert, int vnum, int ver)
{
    xmlNodePtr tree;
    char buf[256];

    if (ver == 1) {
        tree = xmlNewChild(node, NULL, "vertex", NULL);
        sprintf(buf, "%d", vnum);
        xmlSetProp(tree, "id", buf);
        write_float(tree, "vertex_x", vert->x);
        write_float(tree, "vertex_y", vert->y);
        write_float(tree, "vertex_z", vert->z);
        if (!(vert->nx == 0.0 && vert->ny == 0.0 && vert->nz == 0.0)) {
            write_float(tree, "nor_x", vert->nx);
            write_float(tree, "nor_y", vert->ny);
            write_float(tree, "nor_z", vert->nz);
        }
        if (!(vert->tu == 0.0 && vert->tv == 0.0)) {
            write_float(tree, "tex_u", vert->tu);
            write_float(tree, "tex_v", vert->tv);
        }
        if (vert->faceCollapseCount > 0) {
            write_int(tree, "collid", vert->collapseid);
            write_int(tree, "collc", vert->faceCollapseCount);
        }
     }
}

void write_genmodel(char *file, Lmodel_t *model)
{
    xmlDocPtr doc;
    xmlNodePtr tree, subtree;
    int mno, vno, fno, tno, ano;
    Lface_t *face;

    doc = xmlNewDoc("1.0");
    doc->children = xmlNewDocNode(doc, NULL, "genmodel", NULL);
    write_string(doc->children, "name", model->name);
    if (model->modeledby != NULL)
	write_string(doc->children, "modeledby", model->modeledby);
    if (model->paintedby != NULL)
	write_string(doc->children, "paintedby", model->paintedby);
    write_trifloat(doc->children, "bmax", (float)model->bmax.x, (float)model->bmax.y, (float)model->bmax.z);
    write_trifloat(doc->children, "bmin", (float)model->bmin.x, (float)model->bmin.y, (float)model->bmin.z);
    write_trifloat(doc->children, "cog", (float)model->cog.x, (float)model->cog.y, (float)model->cog.z);
    write_trifloat(doc->children, "center", (float)model->center.x, (float)model->center.y, (float)model->center.z);
    write_float(doc->children, "volume", model->volume);
    if (!strcmp("$Id", model->id))
	write_string(doc->children, "id", "$I" "d$");
    else
	write_string(doc->children, "id", model->id);
    tree = xmlNewChild(doc->children, NULL, "materials", NULL);
    for (ano=0; ano < model->nummats; ano++)
        write_material(tree, &model->mats[ano]);

    for (mno=0; mno < model->nummeshes; mno++) {
        tree = xmlNewChild(doc->children, NULL, "surface", NULL);
	if (model->meshes[mno].nobox)
	    (void)xmlNewChild(tree, NULL, "nobox", NULL);
	subtree = xmlNewChild(tree, NULL, "vertexs", NULL);
        write_int(subtree, "verts", model->meshes[mno].numverts);
        for (vno=0; vno < model->meshes[mno].numverts; vno++) {
            write_vertex(subtree, &model->meshes[mno].verts[vno], vno, 1);
        }
        for (tno=0; tno < model->meshes[mno].numtriangles; tno++) {
            subtree = xmlNewChild(tree, NULL, "triangle", NULL);
            write_string(subtree, "name", model->meshes[mno].triangles[tno].name);
	    if (model->meshes[mno].triangles[tno].optmesh)
		xmlNewChild(tree, NULL, "optmesh", NULL);
            write_int(subtree, "matnum", model->meshes[mno].triangles[tno].matnum);
            write_int(subtree, "points", model->meshes[mno].triangles[tno].numfaces*3);
            write_string(subtree, "type", "block");
            write_int(subtree, "lodcount", model->meshes[mno].triangles[tno].lodcount);
            for (fno=0; fno < model->meshes[mno].triangles[tno].numfaces; fno++) {
                face = &model->meshes[mno].triangles[tno].trufaces[fno];
                write_triint(subtree, "face", face->vertexid[0],
                    face->vertexid[1], face->vertexid[2]);
            }
        }
    }
    xmlSaveFormatFile(file, doc, 1);
}

/* parse a quadfloat */
void parse_quadfloat(xmlDocPtr doc, xmlNodePtr cur, float *f)
{
    char *tmp;

    tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
    sscanf(tmp, "%f %f %f %f", &f[0], &f[1], &f[2], &f[3]);
    free(tmp);
}

/* XXX The realloc gives us unzeroed memory, screwing up the whole way
   this routine works. No idea how to fix this. */


void add_face(Lvertex_t *vert, int numt, int faceno, int n1, int n2)
{
    int i;

    if (faceno != -1) {
	if (vert->faceid == NULL)
	    vert->faceid = safer_malloc(sizeof(facen_t) * numt);
	else
	    vert->faceid = realloc(vert->faceid, sizeof(facen_t) * numt);
	if (vert->faceid[numt-1].nei == NULL) {
	    vert->faceid[numt-1].nei = smalloc(int);
	    i = 0;
	} else {
	    i = vert->faceid[numt-1].num;
	    vert->faceid[numt-1].nei = realloc(vert->faceid[numt-1].nei, (i+1) * sizeof(int));
	}
	vert->faceid[numt-1].nei[i] = faceno;
	vert->faceid[numt-1].num = i+1;
    }

    if (n1 == -1 && n2 == -1) {
	i=0; /* nop */
    } else {
	if (vert->neighbor == NULL)
	    vert->neighbor = safer_malloc(sizeof(facen_t) * numt);
	else
	    vert->neighbor = realloc(vert->neighbor, sizeof(facen_t) * numt);
	if (vert->neighbor[numt-1].nei == NULL) {
	    vert->neighbor[numt-1].nei = smalloc(int);
	    i = 0;
	} else {
	    i = vert->neighbor[numt-1].num;
	    vert->neighbor[numt-1].nei = realloc(vert->neighbor[numt-1].nei, (i+2) * sizeof(int));
	}
	vert->neighbor[numt-1].nei[i] = n1;
	if (n2 == -1) {
	    vert->neighbor[numt-1].num = i+1;
	} else {
	    vert->neighbor[numt-1].nei[i+1] = n2;
	    vert->neighbor[numt-1].num = i+2;
	}
    }
}

/* Parse a genmodel file into a lodder struct */

void parse_genmodel(char *filename, Lmodel_t *model)
{
    xmlDocPtr doc;
    xmlNodePtr cur, ncur, xcur, ycur;
    int nrofpoints, stripcur=0, x, y, z;
    char *tmp;

    doc = xmlParseFile(filename);
    if (doc == NULL)
        bailout();

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);
        bailout();
    }
    if (strcmp(cur->name, "genmodel")) {
        fprintf(stderr, 
                "document of the wrong type, root node != genmodel");
        xmlFreeDoc(doc);
        bailout();
    }
    model->meshes = smalloc(Lmesh_t);
    cur = cur->xmlChildrenNode;

    while (cur != NULL) {
	if (!strcmp(cur->name, "name"))
	    parsestring(doc, cur, &model->name);
	if (!strcmp(cur->name, "modeledby"))
	    parsestring(doc, cur, &model->modeledby);
	if (!strcmp(cur->name, "paintedby"))
	    parsestring(doc, cur, &model->paintedby);
	if (!strcmp(cur->name, "id"))
	    parsestring(doc, cur, &model->id);
	if (!strcmp(cur->name, "materials")) {
	    model->mats = smalloc(Lmat_t);
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "material")) {
		    xcur = ncur->xmlChildrenNode;
		    model->nummats++;
		    model->mats = realloc(model->mats, sizeof(Lmat_t)*model->nummats);
		    while (xcur != NULL) {
			if (!strcmp(xcur->name, "name"))
			    parsestring(doc, xcur, &model->mats[model->nummats-1].name);
			if (!strcmp(xcur->name, "image"))
			    parsestring(doc, xcur, &model->mats[model->nummats-1].image);
			if (!strcmp(xcur->name, "ambient"))
			    parse_quadfloat(doc, xcur, model->mats[model->nummats-1].ambient);
			if (!strcmp(xcur->name, "diffuse"))
			    parse_quadfloat(doc, xcur, model->mats[model->nummats-1].diffuse);
			if (!strcmp(xcur->name, "specular"))
			    parse_quadfloat(doc, xcur, model->mats[model->nummats-1].specular);
			if (!strcmp(xcur->name, "shine"))
			    parsefloat(doc, xcur, &model->mats[model->nummats-1].shine);
			xcur = xcur->next;
		    }
		}
		ncur = ncur->next;
	    }
	}
 	if (!strcmp(cur->name, "surface")) {
	    model->nummeshes++;
	    model->meshes = realloc(model->meshes, sizeof(Lmesh_t)*model->nummeshes);
	    model->meshes[model->nummeshes-1].triangles = smalloc(Ltriangle_t);
	    model->meshes[model->nummeshes-1].numtriangles = 0;
	    model->meshes[model->nummeshes-1].nobox = 0;
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "nobox"))
		    model->meshes[model->nummeshes-1].nobox = 1;
		if (!strcmp(ncur->name, "vertexs")) {
		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL) {
			int nrofverts;
			static int cvert;

			if (!strcmp(xcur->name, "verts")) {
			    parseint(doc, xcur, &nrofverts);
			    model->meshes[model->nummeshes-1].numverts = nrofverts;
			    model->meshes[model->nummeshes-1].verts =
			        safer_malloc(sizeof(Lvertex_t) * nrofverts);
			    cvert = 0;
			}
			if (!strcmp(xcur->name, "vertex")) {
			    ycur = xcur->xmlChildrenNode;
			    model->meshes[model->nummeshes-1].verts[cvert].lodid = -1;
			    model->meshes[model->nummeshes-1].verts[cvert].collapseid = -1;
			    model->meshes[model->nummeshes-1].verts[cvert].error = FLT_MAX;
			    model->meshes[model->nummeshes-1].verts[cvert].faceid = smalloc(facen_t);
			    model->meshes[model->nummeshes-1].verts[cvert].faceid[0].nei = NULL;
			    model->meshes[model->nummeshes-1].verts[cvert].faceid[0].num = 0;
			    model->meshes[model->nummeshes-1].verts[cvert].neighbor = smalloc(facen_t);
			    model->meshes[model->nummeshes-1].verts[cvert].neighbor[0].nei = NULL;
			    model->meshes[model->nummeshes-1].verts[cvert].neighbor[0].num = 0;

			    while (ycur != NULL) {
				if (!strcmp(ycur->name, "vertex_x"))
				    parsefloat(doc, ycur, &model->meshes[model->nummeshes-1].verts[cvert].x);
				if (!strcmp(ycur->name, "vertex_y"))
				    parsefloat(doc, ycur, &model->meshes[model->nummeshes-1].verts[cvert].y);
				if (!strcmp(ycur->name, "vertex_z"))
				    parsefloat(doc, ycur, &model->meshes[model->nummeshes-1].verts[cvert].z);
				if (!strcmp(ycur->name, "nor_x"))
				    parsefloat(doc, ycur, &model->meshes[model->nummeshes-1].verts[cvert].nx);
				if (!strcmp(ycur->name, "nor_y"))
				    parsefloat(doc, ycur, &model->meshes[model->nummeshes-1].verts[cvert].ny);
				if (!strcmp(ycur->name, "nor_z"))
				    parsefloat(doc, ycur, &model->meshes[model->nummeshes-1].verts[cvert].nz);
				if (!strcmp(ycur->name, "tex_u"))
				    parsefloat(doc, ycur, &model->meshes[model->nummeshes-1].verts[cvert].tu);
				if (!strcmp(ycur->name, "tex_v"))
				    parsefloat(doc, ycur, &model->meshes[model->nummeshes-1].verts[cvert].tv);
				ycur = ycur->next;
			    }
			    cvert++;
			}
			xcur = xcur->next;
		    }
		}
		if (!strcmp(ncur->name, "triangle")) {
		    int type=M_TRI_BLOCK, vnum, numt;
		    static int faceno;

		    model->meshes[model->nummeshes-1].numtriangles++;
		    numt = model->meshes[model->nummeshes-1].numtriangles;
		    model->meshes[model->nummeshes-1].triangles =
		        realloc(model->meshes[model->nummeshes-1].triangles,
			    sizeof(Ltriangle_t) * numt);
		    if (model->meshes[model->nummeshes-1].triangles == NULL)
			bailout();

		    xcur = ncur->xmlChildrenNode;
		    while (xcur != NULL) {
			if (!strcmp(xcur->name, "name")) {
			    parsestring(doc, xcur, &model->meshes[model->nummeshes-1].triangles[numt-1].name);
			}
			if (!strcmp(xcur->name, "optmesh"))
			    model->meshes[model->nummeshes-1].triangles[numt-1].optmesh = 1;
			if (!strcmp(xcur->name, "matnum")) {
			    parseint(doc, xcur, &model->meshes[model->nummeshes-1].triangles[numt-1].matnum);
			}
			if (!strcmp(xcur->name, "points")) {
			    parseint(doc, xcur, &nrofpoints);
			}
			if (!strcmp(xcur->name, "type")) {
			    tmp = xmlNodeListGetString(doc,
                                xcur->xmlChildrenNode, 1);
			    if (!strcmp(tmp, "strip")) {
				type = M_TRI_STRIP;
				model->meshes[model->nummeshes-1].triangles[numt-1].numfaces = nrofpoints - 2;
				stripcur = 0;
			    }
			    else if (!strcmp(tmp, "block")) {
				type = M_TRI_BLOCK;
				model->meshes[model->nummeshes-1].triangles[numt-1].numfaces = nrofpoints/3;
			    }
			    free(tmp);
			    model->meshes[model->nummeshes-1].triangles[numt-1].faces =
				safer_malloc(sizeof(Lface_t)*model->meshes[model->nummeshes-1].triangles[numt-1].numfaces);
			    model->meshes[model->nummeshes-1].triangles[numt-1].trufaces =
				safer_malloc(sizeof(Lface_t)*model->meshes[model->nummeshes-1].triangles[numt-1].numfaces);
			    faceno = 0;
			}
			if (!strcmp(xcur->name, "face")) {
			    model->meshes[model->nummeshes-1].triangles[numt-1].faces[faceno].lodid = 0;
			    model->meshes[model->nummeshes-1].triangles[numt-1].trufaces[faceno].lodid = 0;
			    if (type == M_TRI_STRIP) {
				parseint(doc, xcur, &vnum);
				model->meshes[model->nummeshes-1].triangles[numt-1].faces[faceno].vertexid[stripcur] = vnum;
				model->meshes[model->nummeshes-1].triangles[numt-1].trufaces[faceno].vertexid[stripcur] = vnum;

				add_face(&model->meshes[model->nummeshes-1].verts[vnum], numt, faceno, 0, 0);
				if (stripcur == 3) {
				    stripcur = 0;
				    faceno++;
				} else
				    stripcur++;
			    } else if (type == M_TRI_BLOCK) {
				parsetriint(doc, xcur, &x, &y, &z);
				model->meshes[model->nummeshes-1].triangles[numt-1].faces[faceno].vertexid[0] = x;
				model->meshes[model->nummeshes-1].triangles[numt-1].faces[faceno].vertexid[1] = y;
				model->meshes[model->nummeshes-1].triangles[numt-1].faces[faceno].vertexid[2] = z;
				model->meshes[model->nummeshes-1].triangles[numt-1].trufaces[faceno].vertexid[0] = x;
				model->meshes[model->nummeshes-1].triangles[numt-1].trufaces[faceno].vertexid[1] = y;
				model->meshes[model->nummeshes-1].triangles[numt-1].trufaces[faceno].vertexid[2] = z;

				add_face(&model->meshes[model->nummeshes-1].verts[x], numt, faceno, y, z);
				add_face(&model->meshes[model->nummeshes-1].verts[y], numt, faceno, x, z);
				add_face(&model->meshes[model->nummeshes-1].verts[z], numt, faceno, y, x);
				faceno++;
			    }
			}
			xcur = xcur->next;
		    } /* while triangle */
		} /* triangle */
		ncur = ncur->next;
	    } /* while */
	} /* surface */
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

void calc_all_quadrics(Lmesh_t *mesh)
{
   Ltriangle_t *tri;
   Lface_t *face;
   int t,f;
   float x1, y1, z1, x2, y2, z2, nx, ny, nz, length, area;

   for (t=0; t < mesh->numtriangles; t++) {
      tri = &mesh->triangles[t];
      for (f=0; f < tri->numfaces; f++) {
	face = &tri->faces[f];
	x1 = mesh->verts[face->vertexid[1]].x - mesh->verts[face->vertexid[0]].x;
	y1 = mesh->verts[face->vertexid[1]].y - mesh->verts[face->vertexid[0]].y;
	z1 = mesh->verts[face->vertexid[1]].z - mesh->verts[face->vertexid[0]].z;
	x2 = mesh->verts[face->vertexid[2]].x - mesh->verts[face->vertexid[0]].x;
	y2 = mesh->verts[face->vertexid[2]].y - mesh->verts[face->vertexid[0]].y;
	z2 = mesh->verts[face->vertexid[2]].z - mesh->verts[face->vertexid[0]].z;
	 
	nx = y1 * z2 - y2 * z1;
	ny = z1 * x2 - z2 * x1;
	nz = x1 * y2 - x2 * y1;

	length = sqrt(nx * nx + ny * ny + nz * nz);
	area = 0.5f * length;

	nx /= length;
	ny /= length;
	nz /= length;

	if (mesh->verts[face->vertexid[0]].nx != 0.0 &&
	    mesh->verts[face->vertexid[0]].ny != 0.0 &&
	    mesh->verts[face->vertexid[0]].nz != 0.0) {
	    if (nx * mesh->verts[face->vertexid[0]].nx +
	    	ny * mesh->verts[face->vertexid[0]].ny +
	    	nz * mesh->verts[face->vertexid[0]].nz < 0.0) {
		    nx = -nx;
		    ny = -ny;
		    nz = -nz;
	    }
	} else {
	    mesh->verts[face->vertexid[0]].nx = nx;
	    mesh->verts[face->vertexid[0]].ny = ny;
	    mesh->verts[face->vertexid[0]].nz = nz;
	}
	set_quadric_plane(nx, ny, nz, -(mesh->verts[face->vertexid[0]].x * nx +
		mesh->verts[face->vertexid[0]].y * ny +
		mesh->verts[face->vertexid[0]].z * nz), area, &face->quadrad);
	scale_quadric(area, &face->quadrad);
	add_quadric(&mesh->verts[face->vertexid[0]].quadrad, &face->quadrad);
	add_quadric(&mesh->verts[face->vertexid[1]].quadrad, &face->quadrad);
	add_quadric(&mesh->verts[face->vertexid[2]].quadrad, &face->quadrad);
      }
   }
}

int freeze_borders(Lmesh_t *mesh)
{
    int vertid, neid, trid, faceid, freezecount=0;
    Lvertex_t *vert, *nvert;
    Ltriangle_t *tri;
    Lface_t *face;
    quadric_t q;
    float nx, ny, nz, length;

    for (trid=0; trid < mesh->numtriangles; trid++) {
	tri = &mesh->triangles[trid];
	for (vertid=0; vertid < mesh->numverts; vertid++) {
	    vert = &mesh->verts[vertid];
	    for (neid=0; neid < vert->neighbor[trid].num; neid++) {
		int facecount = 0, onefaceid = -1;

		nvert = &mesh->verts[vert->neighbor[trid].nei[neid]];
		for (faceid = 0; faceid < vert->faceid[trid].num; faceid++) {
		    face = &tri->faces[vert->faceid[trid].nei[faceid]];
		    if ((face->vertexid[0] == vert->neighbor[trid].nei[neid]) ||
			(face->vertexid[1] == vert->neighbor[trid].nei[neid]) ||
			(face->vertexid[2] == vert->neighbor[trid].nei[neid])) {
			onefaceid = faceid;
			facecount++;
		    }
		}
		if (facecount < 2) {
		    face = &tri->faces[onefaceid];
		    nx = mesh->verts[face->vertexid[0]].nx +
			mesh->verts[face->vertexid[1]].nx +
			mesh->verts[face->vertexid[2]].nx;
		    ny = mesh->verts[face->vertexid[0]].ny +
			mesh->verts[face->vertexid[1]].ny +
			mesh->verts[face->vertexid[2]].ny;
		    nz = mesh->verts[face->vertexid[0]].nz +
			mesh->verts[face->vertexid[1]].nz +
			mesh->verts[face->vertexid[2]].nz;
		    length = sqrt(nx * nx + ny * ny + nz * nz);
		    nx /= length;
		    ny /= length;
		    nz /= length;

		    set_quadric_edge(vert->x, vert->y, vert->z, nvert->x,
			nvert->y, nvert->z, nx, ny, nz, &q);
		    add_quadric(&vert->quadrad, &q);
		    add_quadric(&nvert->quadrad, &q);
		    vert->frozen = 1;
		    freezecount++;
		    break;
		}
	    }
	}
    }
    return(freezecount);
}

float calc_collapse_error(Lmesh_t *mesh, Lvertex_t *vert, Lvertex_t *nvert,
	int vid, int nvid, int trid)
{
    Lface_t *face;
    int faceid;

    for (faceid=0; faceid < vert->faceid[trid].num; faceid++) {
	face = &mesh->triangles[trid].faces[vert->faceid[trid].nei[faceid]];

	if ((face->vertexid[0] == nvid || face->vertexid[1] == nvid ||
	     face->vertexid[2] == nvid) &&
	    face->vertexid[0] != vid && face->vertexid[1] != vid &&
	    face->vertexid[2] != vid) {
	    int v1, v2, v3;
	    float x1, y1, z1, x2, y2, z2, nx, ny, nz, length, nx2, ny2, nz2;

	    v1 = face->vertexid[0];
	    v2 = face->vertexid[1];
	    v3 = face->vertexid[2];

	    x1 = mesh->verts[v2].x - mesh->verts[v1].x;
	    y1 = mesh->verts[v2].y - mesh->verts[v1].y;
	    z1 = mesh->verts[v2].z - mesh->verts[v1].z;

	    x2 = mesh->verts[v3].x - mesh->verts[v1].x;
	    y2 = mesh->verts[v3].y - mesh->verts[v1].y;
	    z2 = mesh->verts[v3].z - mesh->verts[v1].z;

	    nx = y1 * z2 - y2 * z1;
	    ny = z1 * x2 - z2 * x1;
	    nz = x1 * y2 - x2 * y1;

	    length = sqrt(nx * nx + ny * ny + nz * nz);
	    nx /= length;
	    ny /= length;
	    nz /= length;

	    if (v1 == vid)
		v1 = nvid;
	    if (v2 == vid)
		v2 = nvid;
	    if (v3 == vid)
		v3 = nvid;

	    x1 = mesh->verts[v2].x - mesh->verts[v1].x;
	    y1 = mesh->verts[v2].y - mesh->verts[v1].y;
	    z1 = mesh->verts[v2].z - mesh->verts[v1].z;

	    x2 = mesh->verts[v3].x - mesh->verts[v1].x;
	    y2 = mesh->verts[v3].y - mesh->verts[v1].y;
	    z2 = mesh->verts[v3].z - mesh->verts[v1].z;

	    nx2 = y1 * z2 - y2 * z1;
	    ny2 = z1 * x2 - z2 * x1;
	    nz2 = x1 * y2 - x2 * y1;

	    length = sqrt(nx2 * nx2 + ny2 * ny2 + nz2 * nz2);
	    nx2 /= length;
	    ny2 /= length;
	    nz2 /= length;

	    if (nx * nx2 + ny * ny2 + nz * nz2 < 0.0)
		return(FLT_MAX/2.0);
	}
    }

    return(eval_quadric(vert->x, vert->y, vert->z, &vert->quadrad) / vert->quadrad.m_area);
}

void calc_vert_collapse(Lmesh_t *mesh, Lvertex_t *vert, int vid, int trid)
{
    int neid, nvid;
    Lvertex_t *nvert;

    vert->collapseid = -1;
    vert->error = FLT_MAX;

    if (vert->neighbor[trid].num == 0) {
	vert->error = -1.0;
	return;
    }
    for (neid=0; neid < vert->neighbor[trid].num; neid++) {
	float error;

	nvert = &mesh->verts[vert->neighbor[trid].nei[neid]];
	nvid = vert->neighbor[trid].nei[neid];
	error = calc_collapse_error(mesh, vert, nvert, vid, nvid, trid);
	if (error < vert->error && vid != nvid) {
	    vert->error = error;
	    vert->collapseid = nvid;
	}
    }
}

int get_next_collapsing_vertex(Lmesh_t *mesh)
{
    float error=FLT_MAX;
    int colv=-1, vertid;

    for (vertid=0; vertid < mesh->numverts; vertid++) {
	if (!mesh->verts[vertid].collapsed && !mesh->verts[vertid].frozen &&
	    mesh->verts[vertid].error < error) {
	    colv = vertid;
	    error = mesh->verts[vertid].error;
	}
    }
    return(colv);
}

/* pop out one of the values and collapse inward */

void erase_neighbor(int *nei, int nrof, int num)
{
   int i;

   for (i=num; i < nrof; i++) {
	nei[i] = nei[i+1];
   }
}

void pnei(facen_t *foo)
{
    int i;

    printf("%d::", foo->num);
    for (i=0; i < foo->num; i++)
	printf("%d ", foo->nei[i]);
    printf("\n");
}

void collapse_vertex(Lmesh_t *mesh, int nvid, int trid)
{
    Lvertex_t *vert;
    Lface_t *face;
    int faceid, nrofmods=0, i, j, k, found;
    int modverts[100000];

    vert = &mesh->verts[nvid];

    vert->lodid = mesh->nextvertexlodid--;
    vert->collapsed = 1;

    if (vert->collapseid == -1)
	printf("vert %d borked\n", nvid);

    add_quadric(&vert->quadrad, &mesh->verts[vert->collapseid].quadrad);
    for (faceid=0; faceid < vert->faceid[trid].num; ) {
	k = vert->faceid[trid].nei[faceid];
	face = &mesh->triangles[trid].faces[k];
	if (face->vertexid[0] == vert->collapseid ||
	    face->vertexid[1] == vert->collapseid ||
	    face->vertexid[2] == vert->collapseid) {
	    for (j=0; j < 3; j++) {
	    for (i=0, found=0; i < mesh->verts[face->vertexid[j]].faceid[trid].num; i++) {
		if (mesh->verts[face->vertexid[j]].faceid[trid].nei[i] == k) {
		   /* printf("erase vert %d from face slot %d\n",k, i);
		    pnei(&mesh->verts[face->vertexid[j]].faceid[trid]); */
	    	    erase_neighbor(mesh->verts[face->vertexid[j]].faceid[trid].nei, mesh->verts[face->vertexid[j]].faceid[trid].num, i);
		    mesh->verts[face->vertexid[j]].faceid[trid].num--;
		    modverts[nrofmods] = face->vertexid[j];
		    if (!found) {
			nrofmods++;
			found++;
		    }
		}
	    }
	    }
	    face->lodid = mesh->triangles[trid].nextfacelodid--;
	    vert->faceCollapseCount++;
	    faceid = 0;
	} else {
	    faceid++;
	}
    }
    if (vert->collapseid != -1) {
    for (faceid=0; faceid < vert->faceid[trid].num; faceid++) {
	k = vert->faceid[trid].nei[faceid];
	face = &mesh->triangles[trid].faces[k];
	if (face->vertexid[0] == nvid)
	    face->vertexid[0] = vert->collapseid;
	if (face->vertexid[1] == nvid)
	    face->vertexid[1] = vert->collapseid;
	if (face->vertexid[2] == nvid)
	    face->vertexid[2] = vert->collapseid;

	add_face(&mesh->verts[vert->collapseid], trid+1, k, -1, -1);
    }

    for (i=0; i < vert->neighbor[trid].num; ) {
	j = vert->neighbor[trid].nei[i];
	for (k=0,found=0; k < mesh->verts[j].neighbor[trid].num; ) {
	    if (mesh->verts[j].neighbor[trid].nei[k] == nvid) {
		/* printf("erase vert %d from nei %d slot %d\n", nvid, j, k); */
		erase_neighbor(mesh->verts[j].neighbor[trid].nei, mesh->verts[j].neighbor[trid].num, k);
		mesh->verts[j].neighbor[trid].num--;
		found++;
		k = 0;
	    } else {
		k++;
	    }
	}
	if (found) {
	    /* mesh->verts[j].neighbor[trid].nei[mesh->verts[j].neighbor[trid].num] = vert->collapseid;
	    mesh->verts[j].neighbor[trid].num++; */
	    add_face(&mesh->verts[j], trid+1, -1, vert->collapseid, -1);
	    add_face(&mesh->verts[vert->collapseid], trid+1, -1, j, -1);
	    modverts[nrofmods] = j;
            nrofmods++;
	    i = 0;
        } else {
	    i++;
	}
    }
    }
    for (i=0; i < nrofmods; i++) {
	if (modverts[i] == nvid) {
    	    erase_neighbor(modverts, nrofmods, i);
	    nrofmods--;
	    break;
	}
    }
    if (vert->collapseid != -1) {
	modverts[nrofmods] = vert->collapseid;
	nrofmods++;
    }
    for (i=0; i < nrofmods; i++) {
	if (modverts[i] != nvid)
	    calc_vert_collapse(mesh, &mesh->verts[modverts[i]], modverts[i], trid);
    }
}

void calc_levels(Lmesh_t *mesh)
{
    int faceid, vertid, trid;
    Lvertex_t *vert;
    Ltriangle_t *tri;

    for (trid=0; trid < mesh->numtriangles; trid++) {
	tri = &mesh->triangles[trid];
	for (vertid=0; vertid < mesh->numverts; vertid++) {
	    vert = &mesh->verts[vertid];
	    calc_vert_collapse(mesh, vert, vertid, trid);
	}
	mesh->nextvertexlodid = mesh->numverts - 1;
	tri->nextfacelodid = tri->numfaces - 1;

	while (1) {
	    int nvid;

	    nvid = get_next_collapsing_vertex(mesh);
	    if (nvid == -1)
		break;
	    collapse_vertex(mesh, nvid, trid);
	}
	mesh->nextvertexlodid = 0;
	tri->nextfacelodid = 0;

	for (vertid = 0; vertid < mesh->numverts; vertid++) {
	    if (mesh->verts[vertid].lodid == -1) {
		mesh->verts[vertid].lodid = mesh->nextvertexlodid++;
		mesh->verts[vertid].collapseid = -1;
	    }
	}

	for (vertid = 0; vertid < mesh->numverts; vertid++) {
	    if (mesh->verts[vertid].collapseid != -1) {
		mesh->verts[vertid].collapseid =
 		    mesh->verts[mesh->verts[vertid].collapseid].lodid;
	    }
	}

	for (faceid = 0; faceid < tri->numfaces; faceid++) {
	    if (tri->faces[faceid].lodid == -1)
		tri->faces[faceid].lodid = tri->nextfacelodid++;
	}
	tri->lodcount = mesh->numverts - mesh->nextvertexlodid;
    }
}


#if 0
void com_face(Lface_t *face, Lmesh_t *mesh, float *vol, vertex_t *vert)
{
   vertex_t p1, p2, p3;
   float x1, x2, y1, y2, z1, z2, nx, ny, nz;

   p1.x = mesh->verts[face->vertexid[0]].x;
   p1.y = mesh->verts[face->vertexid[0]].y;
   p1.z = mesh->verts[face->vertexid[0]].z;
   p2.x = mesh->verts[face->vertexid[1]].x;
   p2.y = mesh->verts[face->vertexid[1]].y;
   p2.z = mesh->verts[face->vertexid[1]].z;
   p3.x = mesh->verts[face->vertexid[2]].x;
   p3.y = mesh->verts[face->vertexid[2]].y;
   p3.z = mesh->verts[face->vertexid[2]].z;

   x1 = p2.x - p1.x;
   y1 = p2.y - p1.y;
   z1 = p2.z - p1.z;
   
   x2 = p3.x - p1.x;
   y2 = p3.y - p1.y;
   z2 = p3.z - p1.z;

   nx = y1*z2 - z1*y2;
   ny = z1*x2 - x1*z2;
   nz = x1*y2 - y1*x2;

   vert->x += nx*(p1.x*p1.x + p2.x*p2.x + p3.x*p3.x +
		  p1.x*p2.x + p1.x*p3.x + p2.x*p3.x);
   vert->y += ny*(p1.y*p1.y + p2.y*p2.y + p3.y*p3.y +
		  p1.y*p2.y + p1.y*p3.y + p2.y*p3.y);
   vert->z += nz*(p1.z*p1.z + p2.z*p2.z + p3.z*p3.z +
		  p1.z*p2.z + p1.z*p3.z + p2.z*p3.z);

   printf("p1x=%f p1y=%f p1x=%f\n", p1.x, p1.y, p1.z);
   printf("p2x=%f p2y=%f p2x=%f\n", p2.x, p2.y, p2.z);
   printf("p3x=%f p3y=%f p3x=%f\n", p3.x, p3.y, p3.z);
   printf("vx=%f vy=%f vz=%f\n", nx*(p1.x*p1.x + p2.x*p2.x + p3.x*p3.x +
				     p1.x*p2.x + p1.x*p3.x + p2.x*p3.x),
	  ny*(p1.y*p1.y + p2.y*p2.y + p3.y*p3.y +
	      p1.y*p2.y + p1.y*p3.y + p2.y*p3.y),
	  nz*(p1.z*p1.z + p2.z*p2.z + p3.z*p3.z +
	      p1.z*p2.z + p1.z*p3.z + p2.z*p3.z));

   printf("nx=%f ny=%f nz=%f\n", nx, ny, nz);

   /*printf("nx=%f p1.x=%f p2.x=%f p3.x=%f\n", nx, p1.x, p2.x, p3.x);*/

/*   *vol += fabs(nx)*(p1.x + p2.x + p3.x); */
   *vol += (p1.x * (p2.y * p3.z - p2.z *p3.y) +
	    p2.x * (p3.y * p1.z - p3.z * p1.y) +
	    p3.x * (p1.y * p2.z - p1.z * p2.y));
   printf("vol=%f\n", (p1.x * (p2.y * p3.z - p2.z *p3.y) +
		       p2.x * (p3.y * p1.z - p3.z * p1.y) +
		       p3.x * (p1.y * p2.z - p1.z * p2.y)));
}
#endif

void com_face(Lface_t *face, Lmesh_t *mesh, float *vol, vertex_t *vert)
{
   vertex_t p1, p2, p3;

   p1.x = mesh->verts[face->vertexid[0]].x;
   p1.y = mesh->verts[face->vertexid[0]].y;
   p1.z = mesh->verts[face->vertexid[0]].z;
   p2.x = mesh->verts[face->vertexid[1]].x;
   p2.y = mesh->verts[face->vertexid[1]].y;
   p2.z = mesh->verts[face->vertexid[1]].z;
   p3.x = mesh->verts[face->vertexid[2]].x;
   p3.y = mesh->verts[face->vertexid[2]].y;
   p3.z = mesh->verts[face->vertexid[2]].z;

   vert->x = (p1.x + p2.x + p3.x)/3.0;
   vert->y = (p1.y + p2.y + p3.y)/3.0;
   vert->z = (p1.z + p2.z + p3.z)/3.0;

/*   *vol += fabs(nx)*(p1.x + p2.x + p3.x); */
   *vol += (p1.x * (p2.y * p3.z - p2.z *p3.y) +
	    p2.x * (p3.y * p1.z - p3.z * p1.y) +
	    p3.x * (p1.y * p2.z - p1.z * p2.y));
}

void com_model(Lmodel_t *model)
{
    int mno, fno, tno, nfaces=0;

    model->cog.x = 0.0;
    model->cog.y = 0.0;
    model->cog.z = 0.0;
    model->volume = 0.0;

    for (mno=0; mno < model->nummeshes; mno++) {
	for (tno=0; tno < model->meshes[mno].numtriangles; tno++) {
	    for (fno=0; fno < model->meshes[mno].triangles[tno].numfaces; fno++) {
		com_face(&model->meshes[mno].triangles[tno].trufaces[fno],
		    &model->meshes[mno], &model->volume, &model->cog);
		nfaces++;
	    }
	}
    }
    model->cog.x /= nfaces;
    model->cog.y /= nfaces;
    model->cog.z /= nfaces;
/*    
      if (model->volume != 0.0) {
        model->cog.x /= 4.0 * model->volume;
	model->cog.y /= 4.0 * model->volume;
	model->cog.z /= 4.0 * model->volume;
      }
*/
    model->volume /= 6.0;
}

void calc_bbox(Lmodel_t *model)
{
    int mno, vno;

    model->bmax.x = -99999999.0;
    model->bmax.y = -99999999.0;
    model->bmax.z = -99999999.0;

    model->bmin.x = 99999999.0;
    model->bmin.y = 99999999.0;
    model->bmin.z = 99999999.0;

    for (mno=0; mno < model->nummeshes; mno++) {
	if (!model->meshes[mno].nobox) {
	    for (vno=0; vno < model->meshes[mno].numverts; vno++) {
		if (model->meshes[mno].verts[vno].x < model->bmin.x)
		    model->bmin.x = model->meshes[mno].verts[vno].x;
		if (model->meshes[mno].verts[vno].y < model->bmin.y)
		    model->bmin.y = model->meshes[mno].verts[vno].y;
		if (model->meshes[mno].verts[vno].z < model->bmin.z)
		    model->bmin.z = model->meshes[mno].verts[vno].z;
		if (model->meshes[mno].verts[vno].x > model->bmax.x)
		    model->bmax.x = model->meshes[mno].verts[vno].x;
		if (model->meshes[mno].verts[vno].y > model->bmax.y)
		    model->bmax.y = model->meshes[mno].verts[vno].y;
		if (model->meshes[mno].verts[vno].z > model->bmax.z)
		    model->bmax.z = model->meshes[mno].verts[vno].z;
	    }
	}
    }
    model->center.x = (model->bmax.x + model->bmin.x)/2.0;
    model->center.y = (model->bmax.y + model->bmin.y)/2.0;
    model->center.z = (model->bmax.z + model->bmin.z)/2.0;
}

void scale_model(Lmodel_t *model, float scale)
{
    int mno, vno;

    for (mno=0; mno < model->nummeshes; mno++) {
	for (vno=0; vno < model->meshes[mno].numverts; vno++) {
	    model->meshes[mno].verts[vno].x *= scale;
	    model->meshes[mno].verts[vno].y *= scale;
	    model->meshes[mno].verts[vno].z *= scale;
        }
    }
}

void usage(void)
{
    fprintf(stderr,
	    "Valid options:\n"
	    "-f <filename>  : input filename\n"
	    "-o <outfile>   : output filename\n"
	    "-s <scale>     : scale model\n"
	    "-n             : don't write out new model\n"
	    "-l             : regenerate LOD data\n");
    exit(1);
}

int main(int argc, char **argv)
{
    Lmodel_t model;
    int i, frozen, nowrite=0, smodel=0, regenlod=0;
    char ch;
    char *filename=NULL, *outfile=NULL;
    float scale=1.0;

    while ((ch = getopt(argc, argv, "s:f:o:nl")) != -1) {
	switch (ch) {
	case 's': scale = atof(optarg); smodel = 1; break;
	case 'f': filename = strdup(optarg); break;
	case 'o': outfile = strdup(optarg); break;
	case 'n': nowrite = 1; break;
	case 'l': regenlod = 1; break;
	default: usage(); break;
	}
    }

    if (filename == NULL)
	usage();

    model.nummeshes = 0;
    model.nummats = 0;
    model.id = "$I" "d$"; /* don't want CVS "fixing" this */
    model.modeledby = NULL;
    model.paintedby = NULL;
    parse_genmodel(filename, &model);
    if (smodel)
	scale_model(&model, scale);

    if (regenlod) {
	for (i = 0; i < model.nummeshes; i++) {
	    calc_all_quadrics(&model.meshes[i]);
	    frozen = freeze_borders(&model.meshes[i]);
	    printf("mesh %d verts %d tris %d frozen %d\n", i,
		   model.meshes[i].numverts, model.meshes[i].numtriangles,
		   frozen);
	    calc_levels(&model.meshes[i]);
	    printf("calc'd\n");
	}
    }
    com_model(&model);
    calc_bbox(&model);
    if (!nowrite) {
	if (outfile == NULL)
	    write_genmodel("file.xml", &model);
	else
	    write_genmodel(outfile, &model);
    }
    printf("cog %f %f %f vol %f\n", model.cog.x, model.cog.y, model.cog.z,
	   model.volume);
    printf("bmax %f %f %f\n", model.bmax.x, model.bmax.y, model.bmax.z);
    printf("bmin %f %f %f\n", model.bmin.x, model.bmin.y, model.bmin.z);
    printf("center %f %f %f\n", model.center.x, model.center.y,
	   model.center.z);
    return(0);
}
