/* $Id: metaserver.c,v 1.4 2004/02/24 05:36:41 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <time.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "defines.h"
#include "utils.h"
#include "shstr.h"
#include "timers.h"
#include "mnetwork.h"
#include "netparser.h"
#include "msettings.h"
#include "xmlparse.h"
#include "metaserv.h"

const char *progrname;
settings_t settings;
long max_time = 250000;
struct timeval last_time;

TAILQ_HEAD(, _mservdata) mservdata = TAILQ_HEAD_INITIALIZER(mservdata);

void init_settings()
{
    settings.etcdir = PACKAGE_DATA_DIR ETC_SUFX;
    settings.vardir = PACKAGE_DATA_DIR VAR_SUFX;
    settings.bindport = BIND_PORT;
}

void write_mservdata(void)
{
    xmlDocPtr doc;
    xmlNodePtr tree;
    mservdata_t *dat;
    char file[MAX_PATH], *sf;

    if (TAILQ_EMPTY(&mservdata))
	return;

    sf = "metaserverdata.xml";
    sprintf(file, "%s/%s", settings.vardir, sf);

    doc = xmlNewDoc("1.0");
    doc->children = xmlNewDocNode(doc, NULL, "Servers", NULL);
    TAILQ_FOREACH(dat, &mservdata, next) {
	tree = xmlNewChild(doc->children, NULL, "Server", NULL);
	xmlNewProp(tree, "id", dat->id);
	write_string(tree, "name", dat->name);
	write_string(tree, "ip", dat->ip);
	write_int(tree, "port", dat->port);
	write_string(tree, "desc", dat->desc);
	write_string(tree, "mapinfo", dat->mapinfo);
	write_string(tree, "hostname", dat->hostname);
	write_string(tree, "version", dat->version);
	write_string(tree, "webpage", dat->webpage);
	write_string(tree, "ownername", dat->ownername);
	write_string(tree, "owneremail", dat->owneremail);
	write_long(tree, "lastupdate", dat->lastupdate);
	write_long(tree, "firstupdate", dat->firstupdate);
	write_int(tree, "userhwm", dat->userhwm);
	write_int(tree, "maxusers", dat->maxusers);
	if (dat->preferip)
	    write_int(tree, "preferip", dat->preferip);
    }
    xmlSaveFormatFile(file, doc, 1);
    xmlFreeDoc(doc);
}

void parse_mservdata(void)
{
    char file[MAX_PATH], *sf;
    xmlDocPtr doc;
    xmlNodePtr cur, ncur;
    mservdata_t *new;

    sf = "metaserverdata.xml";

    sprintf(file, "%s/%s", settings.vardir, sf);
    doc = xmlParseFile(file);
    if (doc == NULL) {
	/* we don't know about any servers yet.  now what? */
	LOG(LOG_ERROR, "No MetaServerData file at %s.", sf);
	return;
    }

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
	LOG(LOG_ERROR, "MetaServerData file %s is empty.", sf);
	xmlFreeDoc(doc);
	return;
    }
    if (strcmp(cur->name, "Servers")) {
	LOG(LOG_ERROR, "%s: root node != Servers", sf);
	xmlFreeDoc(doc);
	return;
    }
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "Server")) {
	    new = smalloc(mservdata_t);
	    if (xmlHasProp(cur, "id"))
		new->id = xmlGetProp(cur, "id");
	    ncur = cur->xmlChildrenNode;
	    while (ncur != NULL) {
		if (!strcmp(ncur->name, "name"))
		    parsestring(doc, ncur, &new->name);
		if (!strcmp(ncur->name, "ip"))
		    parsestring(doc, ncur, &new->ip);
		if (!strcmp(ncur->name, "port"))
		    parseint(doc, ncur, &new->port);
		if (!strcmp(ncur->name, "desc"))
		    parsestring(doc, ncur, &new->desc);
		if (!strcmp(ncur->name, "mapinfo"))
		    parsestring(doc, ncur, &new->mapinfo);
		if (!strcmp(ncur->name, "hostname"))
		    parsestring(doc, ncur, &new->hostname);
		if (!strcmp(ncur->name, "version"))
		    parsestring(doc, ncur, &new->version);
		if (!strcmp(ncur->name, "webpage"))
		    parsestring(doc, ncur, &new->webpage);
		if (!strcmp(ncur->name, "ownername"))
		    parsestring(doc, ncur, &new->ownername);
		if (!strcmp(ncur->name, "owneremail"))
		    parsestring(doc, ncur, &new->owneremail);
		if (!strcmp(ncur->name, "lastupdate"))
		    parselong(doc, ncur, &new->lastupdate);
		if (!strcmp(ncur->name, "firstupdate"))
		    parselong(doc, ncur, &new->firstupdate);
		if (!strcmp(ncur->name, "userhwm"))
		    parseint(doc, ncur, &new->userhwm);
		if (!strcmp(ncur->name, "maxxusers"))
		    parseint(doc, ncur, &new->maxusers);
		if (!strcmp(ncur->name, "preferip"))
		    new->preferip = 1;
		ncur = ncur->next;
	    }
	    if (TAILQ_EMPTY(&mservdata))
		TAILQ_INIT(&mservdata);
	    TAILQ_INSERT_TAIL(&mservdata, new, next);
	}
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

void parse_settings(void)
{
    char file[MAX_PATH], *sf;
    xmlDocPtr doc;
    xmlNodePtr cur;

    sf = "metaserverconfig.xml";

    sprintf(file, "./%s", sf);
    doc = xmlParseFile(file);
    if (doc == NULL) {
	sprintf(file, "%s/%s", settings.etcdir, sf);
	doc = xmlParseFile(file);
    }
    if (doc == NULL) {
	sprintf(file, "/etc/%s", sf);
	doc = xmlParseFile(file);
    }
    if (doc == NULL) {
	LOG(LOG_WARNING, "Warning, cannot find settings file %s", sf);
	return;
    }
    /* if we got this file, we found one */

    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        LOG(LOG_WARNING, "Settings file %s is empty", file);
        xmlFreeDoc(doc);
	return;
    }
    if (strcmp(cur->name, "MetaServerSettings")) {
        LOG(LOG_WARNING,"%s: root node != MetaServerSettings", file);
        xmlFreeDoc(doc);
	return;
    }
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
	if (!strcmp(cur->name, "etcdir"))
	    parsestring(doc, cur, &settings.etcdir);
	if (!strcmp(cur->name, "vardir"))
	    parsestring(doc, cur, &settings.vardir);
	if (!strcmp(cur->name, "bindport"))
	    parseint(doc, cur, &settings.bindport);
	cur = cur->next;
    }
    xmlFreeDoc(doc);
}

int main(int argc, char **argv)
{
    xmlKeepBlanksDefault(0);
    init_hash_table();
#ifdef __NetBSD__
    progrname = getprogname();
    setproctitle(NULL);
#endif
    openlog();
    init_argcomm();
    init_settings();
    parse_settings();
    parse_mservdata();

    init_network();

    network_loop(NULL);
    closelog();

    return(0);
}
