/* $Id: mparser.c,v 1.8 2004/04/03 12:29:18 xb49 Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <string.h>
#include <sys/param.h>
#include <sys/time.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
#define bool char
#include <time.h>
#else
#include <sys/queue.h>
#endif

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "defines.h"
#include "timers.h"
#include "shstr.h"
#include "utils.h"
#include "xmlparse.h"
#include "msettings.h"
#include "netparser.h"
#include "mnetwork.h"
#include "mparser.h"
#include "metaserv.h"

extern TAILQ_HEAD(, _mservdata) mservdata;
/* XXX */
extern void write_mservdata(void);

argtable_t commtable[] = {
    {"up", COMMAND_UP, 0},
    {"get", COMMAND_GET, 0}
};

const size_t comm_size = sizeof(commtable) / sizeof(argtable_t);

void parse_command_null(descriptor_data_t *d, pargs_t *args)
{
    return;
}

void parse_command_get(descriptor_data_t *d, pargs_t *args)
{
    char *buf, *buf2;
    mservdata_t *tmp;
    size_t len=0;
    time_t now;

    now = time(NULL);
    TAILQ_FOREACH(tmp, &mservdata, next) {
	/* skip over servers that haven't updated in over 24hrs */
	if (difftime(tmp->lastupdate, now) > 86400.0)
	    continue;
	len += strlen(tmp->id);
	len += ((tmp->name) ? strlen(tmp->name) : 0);
	len += ((tmp->ip) ? strlen(tmp->ip) : 0);
	len += ((tmp->desc) ? strlen(tmp->desc) : 0);
	len += ((tmp->mapinfo) ? strlen(tmp->mapinfo) : 0);
	len += ((tmp->hostname) ? strlen(tmp->hostname) : 0);
	len += ((tmp->version) ? strlen(tmp->version) : 0);
	len += ((tmp->webpage) ? strlen(tmp->webpage) : 0);
	len += ((tmp->ownername) ? strlen(tmp->ownername) : 0);
	len += ((tmp->owneremail) ? strlen(tmp->owneremail) : 0);
	len += 256;
	buf = safer_malloc(len * sizeof(char));
	buf2 = safer_malloc(len * sizeof(char));
	sprintf(buf, "server servid:%s numusers:%d", tmp->id, tmp->numusers);
	if (tmp->name) {
	    sprintf(buf2, " nm:\"%s\"", tmp->name);
	    strcat(buf, buf2);
	}
	if (tmp->ip) {
	    sprintf(buf2, " ipaddr:\"%s\"", tmp->ip);
	    strcat(buf, buf2);
	}
	if (tmp->port) {
	    sprintf(buf2, " portnum:%d", tmp->port);
	    strcat(buf, buf2);
	}
	if (tmp->desc) {
	    sprintf(buf2, " desc:\"%s\"", tmp->desc);
	    strcat(buf, buf2);
	}
	if (tmp->mapinfo) {
	    sprintf(buf2, " mapinfo:\"%s\"", tmp->mapinfo);
	    strcat(buf, buf2);
	}
	if (tmp->hostname) {
	    sprintf(buf2, " hostname:\"%s\"", tmp->hostname);
	    strcat(buf, buf2);
	}
	if (tmp->version) {
	    sprintf(buf2, " srvversion:\"%s\"", tmp->version);
	    strcat(buf, buf2);
	}
	if (tmp->webpage) {
	    sprintf(buf2, " webpage:\"%s\"", tmp->webpage);
	    strcat(buf, buf2);
	}
	if (tmp->ownername) {
	    sprintf(buf2, " ownername:\"%s\"", tmp->ownername);
	    strcat(buf, buf2);
	}
	if (tmp->owneremail) {
	    sprintf(buf2, " owneremail:\"%s\"", tmp->owneremail);
	    strcat(buf, buf2);
	}
	if (tmp->lastupdate) {
	    sprintf(buf2, " lastup:%ld", tmp->lastupdate);
	    strcat(buf, buf2);
	}
	if (tmp->firstupdate) {
	    sprintf(buf2, " firstup:%ld", tmp->firstupdate);
	    strcat(buf, buf2);
	}
	if (tmp->userhwm) {
	    sprintf(buf2, " userhwm:%d", tmp->userhwm);
	    strcat(buf, buf2);
	}
	if (tmp->maxusers) {
	    sprintf(buf2, " maxusers:%d", tmp->maxusers);
	    strcat(buf, buf2);
	}
	if (tmp->preferip)
	    strcat(buf, " preferip:1\r");
	else
	    strcat(buf, "\r");
	do_write(d, buf);
	free(buf);
	free(buf2);
    }
    /* the client should disconnect after getting an msrvdone */
    buf = safer_malloc(sizeof(char) * 10);
    sprintf(buf, "msrvdone\r");
    do_write(d, buf);
    free(buf);
}

void parse_command_up(descriptor_data_t *d, pargs_t *args)
{
    int i;
    mservdata_t *dat, *tmp;
    char id[32];

    i = find_arg_by_id(args, PORTNUM);
    if (i == -1) 
	/* assume 5501 */
	sprintf(id, "%s:5501", d->ipaddr);
    else
	sprintf(id, "%s:%d", d->ipaddr, args[i].arg.i);

    dat = NULL;
    if (TAILQ_EMPTY(&mservdata))
	TAILQ_INIT(&mservdata);
    else
	TAILQ_FOREACH(tmp, &mservdata, next)
	    if (!strcmp(id, tmp->id)) {
		dat = tmp;
		break;
	    }

    if (dat == NULL) {
	dat = smalloc(mservdata_t);
	dat->id = add_string(id);
	dat->ip = add_string(d->ipaddr);
	if (i != -1)
	    dat->port = args[i].arg.i;
	else
	    dat->port = 5501;
	dat->firstupdate = time(NULL);
	TAILQ_INSERT_TAIL(&mservdata, dat, next);
    }
    /* now see if the server has updated or changed anything it
       wants us to record.  new servers will update everything. */
    for (i=0; args[i].cword != -1; i++) {
        switch (args[i].cword) {
        case NAME: dat->name = add_string(args[i].arg.c); break;
        case DESC: dat->desc = add_string(args[i].arg.c); break;
        case MAPINFO: dat->mapinfo = add_string(args[i].arg.c); break;
        case HOSTNAME: dat->hostname = add_string(args[i].arg.c); break;
        case SRVVERSION: dat->version = add_string(args[i].arg.c); break;
        case WEBPAGE: dat->webpage = add_string(args[i].arg.c); break;
        case OWNERNAME: dat->ownername = add_string(args[i].arg.c); break;
        case OWNEREMAIL: dat->owneremail = add_string(args[i].arg.c); break;
        case NUMUSERS: dat->numusers = args[i].arg.i; break;
        case USERHWM: dat->userhwm = args[i].arg.i; break;
        case MAXUSERS: dat->maxusers = args[i].arg.i; break;
        case PREFERIP: dat->preferip = args[i].arg.i; break;
	}
    }
    dat->lastupdate = time(NULL);
    /* we got the update, now close out the connection */
    close_desc(d->descriptor);
    free_desc(d);
    d = 0;
    /* someone updated, lets save the data */
    if (dat->hostname == NULL)
	dat->preferip = 1;
    write_mservdata();
}
