/* $Id: cmd_emote.c,v 1.12 2004/03/24 14:30:00 xb49 Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif
#include <unistd.h>
#include <math.h>
#include <sys/param.h>

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "collision.h"
#include "tod.h"
#include "xmlparse.h"
#include "matrix.h"
#include "event.h"
#include "player.h"
#include "network.h"
#include "netparser.h"
#include "ninterface.h"
#include "commands.h"

/**
   \file cmd_emote.c
   \brief Commands having to do with emotions
   \author Tim Rightnour

   This file contains all emotions, both physically animated and those
   expressed via text, and all chat and communication related commands.
*/

/**
   \brief Schedules an event to perform an animation action
   \return void
   \param obj the obejct to perform the action
   \param action the action to be performed see ANIM_* in commands.h
   \note expects obj to be locked
*/

void do_emote(object_t *obj, int action)
{
    event_t *ev;

    ev = new_obj_event(Q_PLAYER, ETYPE_NONE, "obj_eact", obj);
    ev->action = action;
    queue_event(ev, obj->mapgroup);
}

/**
   \brief Calls do_emote to perform a wave animation
   \return int success
   \param obj the object to perform the wave
   \param args arguments to the command, for this command there are none
   \note expects obj to be locked
*/

int command_wave(object_t *obj, pargs_t *args)
{
    do_emote(obj, ANIM_WAVE);
    return(1);
}

/**
   \brief Schedules a "who is online" event
   \return int success
   \param obj the object asking for a list of players online
   \param args arguments to the command, for this command there are none
   \note expects obj to be locked
*/

int command_who(object_t *obj, pargs_t *args)
{
    event_t *ev;

    ev = new_obj_event(Q_PLAYER, ETYPE_NONE, "event_who", obj);
    queue_event(ev, obj->mapgroup);
    return(1);
}

/**
   \brief Schedules a say event with text from the client
   \return int success
   \param obj the object doing the speaking
   \param args arguments to the command
   \note expects obj to be locked
*/

int command_say(object_t *obj, pargs_t *args)
{
    int i, lang=0, volume=2;
    char *data=NULL;
    event_t *ev;

    for (i=0; args[i].cword != -1; i++) {
	switch (args[i].cword) {
	case DATA:
	    data = add_string(args[i].arg.c);
	    break;
	case LANGUAGE:
	    lang = args[i].arg.i;
	    break;
	case VOLUME:
	    volume = args[i].arg.i;
	    break;
	}
    }
    if (data == NULL)
	return(0);

    ev = new_obj_event(Q_PLAYER, ETYPE_NONE, "event_say", obj);
    
    ev->action = lang;
    evarg_setastr(ev,0,add_string(data));
    evarg_setint(ev,1,volume);

    queue_event(ev, obj->mapgroup);
    free_string(data);
    return(1);
}
