/* $Id: commands.c,v 1.12 2004/03/11 04:06:45 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif
#include <unistd.h>
#include <ctype.h>
#include <sys/param.h>

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "collision.h"
#include "tod.h"
#include "xmlparse.h"
#include "matrix.h"
#include "event.h"
#include "player.h"
#include "netparser.h"
#include "network.h"
#include "commands.h"

/**
   \file commands.c
   \brief Command interpreters
   \author Tim Rightnour
*/

/** The command table */
commands_t commands[] = {
    {"drop", command_drop, Q_PLAYER},
    {"get", command_get, Q_PLAYER},
    {"unwear", command_unwear, Q_PLAYER},
    {"wield", command_wield, Q_PLAYER},
    {"wave", command_wave, Q_PLAYER},
    {"say", command_say, Q_PLAYER},
    {"who", command_who, Q_PLAYER}
};

/** The size of the command table */
const size_t commands_size = sizeof(commands) / sizeof(commands_t);

/**
   \brief compare two commands lexographically
   \param a first command
   \param b second command
   \return see strcmp()
*/

static int compare_command(const void *a, const void *b)
{
    return strcmp(((commands_t *)a)->name, ((commands_t *)b)->name);
}

/**
   \brief Initialize the commands table
*/

void init_commands(void)
{
    qsort((char *)commands, commands_size, sizeof(commands_t),
	  compare_command);
}

/**
   \brief Handle a command from the network
   \param d descriptor command came from
   \param command command to execute
   \param args arguments to command
*/

int parsed_command(descriptor_data_t *d, char *command, pargs_t *args)
{
    commands_t *asp, dummy;
    object_t *obj;
    char *cp;
    int ret;

    if (d->pl == NULL)
	return(-1);
    if (d->pl->obj == NULL)
	return(-1);
    obj = d->pl->obj;

    for (cp=command; *cp; cp++) {
	*cp = tolower(*cp);
    }
    dummy.name = command;
    asp = (commands_t *)bsearch((void *)&dummy, (void *)commands,
				commands_size, sizeof(commands_t),
				compare_command);
    if (asp) {
	ret = asp->func(obj, args);
	return(ret);
    } else {
	return(-1); /* command not found */
    }
}
