/* $Id: event.h,v 1.23 2004/03/24 14:30:01 xb49 Exp $ */

#ifndef _EVENT_H_
#define _EVENT_H_

/**
   \file event.h
   \brief Event queue definitions
   \author Tim Rightnour
*/

/* Queue priorities */
#define Q_URGENT	1	/**< \brief Urgent queue */
#define Q_PLAYER	2	/**< \brief Player queue */
#define Q_NPC		3	/**< \brief NPC queue */
#define Q_TIMER		4	/**< \brief Timered events */
#define Q_MAINT		5	/**< \brief Server maintenance */
#define Q_NPCLIFE	6	/**< \brief NPC life functions */
#define Q_WEATHER	7	/**< \brief Weather */
#define	Q_MAX		8	/**< \brief The number of queues */


#define ETYPE_NONE	0	/**< \brief Default event type */
#define ETYPE_TIMER	1	/**< \brief Timer event */
#define ETYPE_DNOB	2	/**< \brief Do Not Open Before */
#define ETYPE_NPC	3	/**< \brief NPC event */


#define EACT_WANDER	0	/**< \brief Action wander */
#define EACT_MOVEOBJS	1	/**< \brief Action move objects */
#define EACT_NPCLIFE	2	/**< \brief Action do npc_life() */

#define EVFLAG_SIZE 1 /**<\brief number of int needed for event flag */

#define EVFLAG_QUEUED	0	/**< \brief Event is queued */
#define EVFLAG_NOSAVE	1	/**< \brief Event should not be saved */
#define EVFLAG_OBJ	2	/**< \brief This is an object event */
#define EVFLAG_FOLLOW_OBJ  3 /** <\brief If the innitiator isn't in the eventqueue mapgroup requeue it to the good one */
#define EVFLAG_SERVER_IGNORE  4 /** <\brief ignore the event if it is in the server event queue */
#define EVFLAG_DEL_NOT_IN_MAPG  5 /** <\brief delete the event if the initiator isn't in the event queue mapg */
#define EVFLAG_INITIATOR 6 /** <\brief initiator must exist */

#include "tod.h"
#include "defines.h"
/*#include "object.h"*/
/*#include "outnet.h" */

struct object_t;
struct _mapgroup_t;
struct _event_t ;
struct _servout_t ;
struct _map_head_t;
/** \brief an event queue struct 
this event_queue is more or less thread safe.
\note some of the function will be executed later if it's the currant item in the loop
\see event_queue_init event_queue_first event_queue_next
*/

typedef struct _event_queue_t {
	TAILQ_HEAD(,_event_t) queue; /**< real queue */
	struct _event_t * _cur; /**< current item */
	int _cur_action; /**< do we need to delete or move curant item */
	struct _event_queue_t *_cur_move;  /**< do we need to move currant */
	struct _mapgroup_t *mapg;
}event_queue_t;


#define EVARG_NBR 8 /**< \brief number of arg in envent */
enum {
    EVARG_NONE=0,  /**< \brief type none for event arg */
    EVARG_INT,  /** < \brief type int for event arg */
    EVARG_FLOAT,  /** < \brief type float for event arg */
    EVARG_DSTRING, /** < \brief dynamically allocated string (alloc)*/
    EVARG_ASTRING, /** < \brief string allocated with add_string*/
    EVARG_SSTRING, /** < \brief static  string (no free) */
    EVARG_OBJ, /** < \brief type object */
};

/** \brief an event queue entry */
typedef struct _event_t {
    int queue;		/**< \brief which queue does this belong in? */
    int type;		/**< \brief type of action requested */
    int action;		/**< \brief action requested */
    int dnob;		/**< \brief do not open before */
    /* it's a flag now int objectevent;	*/ /**< \brief is this an object event? */
    char *func;		/**< \brief function name to call */
    struct _object_t *initiator;/**< \brief object that this event revolves around */
   /* struct _object_t *owner; REMOVED */	/**< \brief object event's owner object */
    struct _servout_t *out;	/**< \brief output socket */
    char arg_type[EVARG_NBR]; /** < \brief type of args */
    union {
	float f;
	int i;
	char *c;
	struct _object_t * obj;
    } arg_val[EVARG_NBR];		/**< \brief 8 possible data fields */
    int flags[EVFLAG_SIZE];	/**< \brief event flags */
    TAILQ_ENTRY(_event_t) next; /**< \brief next event in the queue */
    TAILQ_ENTRY(_event_t) stack;/**< \brief for obj inventory of events */
    struct _maphead_t *maph;	/**< \brief which map in the mapgroup is this on? */
    event_queue_t * event_queue;	/**< cur event_queue */ 
} event_t;

/** \brief an event function record */
typedef struct _funcrec_t {
    char *name;			/**< \brief Name of event */
    int (*function)(event_t *);	/**< \brief function to call */
    int flags[EVFLAG_SIZE]; /**<\brief default flags */
} funcrec_t;

int empty_func_handler(event_t *event);
int find_efunc_by_event(event_t *event);
void init_efuncrec(void);
void init_time(void);
void init_server_queues(void);
void init_queues(struct _mapgroup_t *mapg);
event_t *new_event(int queue, int type, char *func);
event_t *new_obj_event(int queue, int type, char *func, struct _object_t *obj);
void get_tod(timeofday_t *tod);
void queue_event(event_t *event, struct _mapgroup_t *mapg);
void queue_obj_event(event_t *event);
void unqueue_event(event_t *event);
void init_server_events(void);
void init_events(struct _mapgroup_t *mapg);
int save_persistent_data(event_t *event);
int metaserver_update(event_t *event);
int update_all_maps_object_motion(event_t *event);
void *event_loop(void *mapgroup);
void *server_event_loop(void *nada);

/* function on event_queue */
void event_queue_init(event_queue_t *queue,struct _mapgroup_t *mapg);
event_t * event_queue_first(event_queue_t *queue);
event_t * event_queue_next(event_queue_t *queue);

void delete_event(event_t * event);

event_queue_t * get_event_queue( struct _mapgroup_t *mapg,int n);

struct _mapgroup_t * event_mapg(event_t *e);

void add_event_obj(event_t*event);

void delete_all_obj_event(struct _object_t *obj);
void unqueue_all_obj_event(struct _object_t *obj);
void queue_all_obj_event(struct _object_t *obj,struct _mapgroup_t *mapg);

/* evarg func */
void evarg_setint(event_t * event,int num,int v);
void evarg_setfloat(event_t * event,int num,float v);
void evarg_setdstr(event_t * event,int num,char *s);
void evarg_setsstr(event_t * event,int num,char *s);
void evarg_setastr(event_t * event,int num,char *s);
void evarg_setobj(event_t * event,int num,struct _object_t *obj);

int evarg_getint(event_t * event,int num);
float evarg_getfloat(event_t * event,int num);
char * evarg_getstr(event_t * event,int num);
struct _object_t * evarg_getobj(event_t * event,int num);

#define evarg_isset(event,num) ((event)->arg_type[num] != EVARG_NONE)



#endif /* _EVENT_H_ */
