/* $Id: map.h,v 1.19 2004/03/30 06:45:55 garbled Exp $ */

#ifndef _MAP_H_
#define _MAP_H_

/**
   \file map.h
   \author Tim Rightnour
   \brief Map definitions
*/

#include "defines.h"

/**
   \brief The map structure
   map_t is a separate structure because it is the meat of the maps.  This
   allows us to have all the mapheads loaded by the server, and selectively
   load the map_t's
*/

typedef struct _map_t {
    char *elevfile;	      /**< \brief png file containing elevation data */
    uint8_t **elevation;	/**< \brief array of spaces.  alititude */
    char *terrfile;		/**< \brief png file containing terrain data */
    uint8_t **terrain;		/**< \brief array of spaces.  texture */
    /* linked list of players on map (shortcut) */
    TAILQ_HEAD(, _object_t) objs; /**< \brief linked list of objects on map */
} map_t;

/**
   \brief Map header structure
*/

typedef struct _maphead_t {
    char *name;			/**< \brief map name */
    char *path;			/**< \brief map path */
    int num;			/**< \brief unique map number */
    int mapnumber;		/**< \brief map number within the mapgroup */
    uint32_t reset_time;	/**< \brief time this map should reset */
    uint32_t reset_timeout;	/**< \brief how many seconds before reset */
    int	width;			/**< \brief map width */
    int height;			/**< \brief map height */
    int difficulty;		/**< \brief map difficulty */
    float x, y, z;	   /**< \brief x and z 0 coordinates for map. y=elev */
    struct _maphead_t *tile[9]; /**< \brief map tiling path data */
    /* what about skymaps, fog, skytone, stafields and lighting?
       With the exception of lighting, the server doesn't need to know about
       the above, but it needs to tell the client */
    int flags[2];		/**< \brief map flags, if any */
    map_t *map;			/**< \brief the real map data */
    struct _mapgroup_t *mapg;	/**< \brief pointer to the mapgroup */
    TAILQ_ENTRY(_maphead_t) next; /**< \brief next maphead in list */
} maphead_t;

#include "event.h"

/**
   \brief Map group structure
*/

typedef struct _mapgroup_t {
    char *name;	       		/**< \brief name of this map grouping */
    int width;			/**< \brief width of mapset */
    int nrofmaps;		/**< \brief nrof maps in mapset */
    pthread_mutex_t mutex;	/**< \brief per group mutex */
    TAILQ_ENTRY(_mapgroup_t) next; /**< \brief next group in list */
    TAILQ_HEAD(, _maphead_t) maps; /**< \brief mapheaders I contain */
    struct _event_queue_t queue[Q_MAX]; /**< \brief My event queue */
    pthread_t thread;		/**< \brief the event thread for this group */
} mapgroup_t;

#define PNGMAPTYPE_ELEV	1	/**< \brief elevation map */
#define PNGMAPTYPE_TERR	2	/**< \brief terrain map */

#define POOM_NOSPACE	0	/**< \brief Don't put object in collision */
#define POOM_SPACE	1	/**< \brief Put object in collision */

#define MAPFLAG_LOADED		0   /**< \brief this map is loaded */
#define MAPFLAG_STARTING	1   /**< \brief a designated starting map */

#include "object.h"

maphead_t *find_map_byname(char *name, mapgroup_t *mapg);
maphead_t *find_map_bynum(int num, mapgroup_t *mapg);
void set_map_loaded(maphead_t *maph);
void set_map_unloaded(maphead_t *maph);
void remove_obj_from_map(object_t *obj, maphead_t *maph, int space, int group);
map_t *new_map(void);
void parsemap(xmlDocPtr doc, xmlNodePtr cur, map_t *map);
void parsemaphead(xmlDocPtr doc, xmlNodePtr cur, maphead_t *maph);
void put_obj_on_map(object_t *obj, maphead_t *maph, int space);
void put_obj_on_map_closeto(object_t *obj, maphead_t *maph, int space);
int in_range_from_border(object_t *obj, int border, float range);
int obj_close_to_border(object_t *obj);
int map_shift_direction(maphead_t *old, maphead_t *new);
maphead_t *obj_on_map(object_t *obj);
void parse_mapfile(char *filename, char *path, int mapnum, mapgroup_t *mapg,
		   float basex, float basey, float basez, int mapsize);
void parse_mapgroup(char *filename);
void parse_mapgroups(char *filename, char *path);
#endif /* _MAP_H_ */
