/* $Id: matrix.h,v 1.7 2004/03/30 06:45:55 garbled Exp $ */

#ifndef _MATRIX_H_
#define _MATRIX_H_

/**
   \file matrix.h
   \brief Matrix definitions
   \author Tim Rightnour
*/

#include "defines.h"

/**
   \brief A 4x4 matrix
*/

typedef struct _matrix_t {
    float m[4][4];	/** \brief matrix members */
} matrix_t; 

int ASSERT(const void *a, const char *s);
matrix_t * MatrixNew(void);
int MatrixCopy(matrix_t *d, const matrix_t *s);
int MatrixDelete(matrix_t *m);
int MatrixIdentity(matrix_t *m);
int MatrixTranspose(const matrix_t *s, matrix_t *d);
int MatrixMultiply(const matrix_t *a, const matrix_t *b, matrix_t *c);
int MatrixInverse(const matrix_t *in, matrix_t *out);
int PrintMatrix(const matrix_t *m);
void new_coord(vertex_t direction, float motion, vertex_t coord,
	       vertex_t *newcoord);
void euler_to_quat4(float ax, float ay, float az, float quat4[4]);
void euler_to_matrix(vertex_t euler, matrix_t *mat);
void prep_xmat(matrix_t *m, float dir);
void prep_ymat(matrix_t *m, float dir);
void prep_zmat(matrix_t *m, float dir);
void rotate_bbox(bbox_t *box, vertex_t dir, vertex_t coord);
void rotate_point_on_center(vertex_t coord, vertex_t center, vertex_t dir,
			    vertex_t *new);
int find_lowest_bbox_point(bbox_t *box);
int find_lowest_bbox_side(bbox_t *box, maphead_t *maph, int side[4]);
float bbox_groundheight_diff(bbox_t *box, maphead_t *maph, vertex_t coord);

#endif /* _MATRIX_H_ */
