/* $Id: network.h,v 1.6 2004/03/11 04:06:46 garbled Exp $ */

#ifndef _NETWORK_H_
#define _NETWORK_H_

/**
   \file network.h
   \brief Network defines
   \author Tim Rightnour
   \author Dave Watson
*/

#include <time.h>
#include "player.h"

/**
   \brief the network descriptor
*/

typedef struct _descriptor_data_t {
    int	descriptor;	/**< \brief The descriptor */
    char *in_buf;	/**< \brief Incoming buffer */
    char *out_buf;	/**< \brief outgoing buffer */
    int outbuflen;	/**< \brief length of outgoing buffer */
    char host[50];	/**< \brief hostname of other side */
    time_t idle;	/**< \brief how long have we idled? */
    player_t *pl;	/**< \brief backlink to the player object */
    pthread_mutex_t mutex;	/**< \brief Mutex to protect the desc. */
    TAILQ_ENTRY(_descriptor_data_t) next; /**< \brief Next in list */
} descriptor_data_t;

#define BIND_PORT	5501	/** \brief TCP port to bind to
				    \note To bind to
				    ports that are < 1024, you usually need
				    superuser permission on that system */
#define USEC_TIMEOUT	250000	/** \brief  Timeout between polls in ms */
#define IDLE_TIMEOUT	300	/** Timeout before disconnecting an idle
				 * connection in seconds. (0=No Timeout) */
#define NO_NAGLE		/** Socket option TCP_NODELAY. Read comments
				 * in init_socket() for details. */
#define NONBLOCK		/** Turns on non-blocking on descriptors. Read
				 * comments in init_desc() for details. */
#define USE_INET_NTOA		/** This should only be undefined if you have
				 * problems with inet_ntoa() */

void *network_loop(void *nada);
void init_network(void);
void do_write(descriptor_data_t *d, char *buf);
void lock_do_write(descriptor_data_t *d, char *buf);

#endif /* _NETWORK_H_ */
