/* $Id: player.h,v 1.26 2004/03/30 06:45:55 garbled Exp $ */

#ifndef _PLAYER_H_
#define _PLAYER_H_

/**
   \file player.h
   \brief Player definitions
   \author Tim Rightnour
*/

#include "defines.h"
#include "object.h"
#include "map.h"
/**
   \brief Things this player knows about
*/

typedef struct _pl_known_t {
    uniqid_t uid;	/**< \brief obj id of known item */
    int lu;	/**< \brief update counter */
    /* things we know about this obj */
    int nrof;	/**< \brief How many? */
    int state;	/**< \brief It's state? */
    char *name;	/**< \brief It's name */
    char *name_pl;	/**< \brief plural name */
    char *model;	/**< \brief Model filename */
    vertex_t facing;	/**< \brief direction facing */
    vertex_t direction;	/**< \brief direction moving */
    vertex_t coord;	/**< \brief location */
    vertex_t scale;	/**< \brief scale */
    float velocity;	/**< \brief speed */
    unsigned int bits;	/**< \brief bitfield of last changes */
    TAILQ_ENTRY(_pl_known_t) next; /**< \brief next in list */
} pl_known_t;

/**
   \brief The player structure
*/

typedef struct _player_t {
    char *username;   /**< \brief the name of the user who owns this char */
    char *charname;		     /**< \brief the character's name */
    struct _descriptor_data_t *d;    /**< \brief descriptor */
    object_t *obj;		  /**< \brief backlink to the primary object */
    int known_maps;		     /**< \brief maps the player knows about */
    TAILQ_HEAD(, _pl_known_t) known; /**< \brief player knowledge structure */
    TAILQ_ENTRY(_player_t) next;     /**< \brief link to next player */
} player_t;

#define PLKN_NUM	(1<<1)	/**< \brief Player knows num */
#define PLKN_NROF	(1<<2)	/**< \brief Player knows nrof */
#define PLKN_STATE	(1<<3)	/**< \brief Player knows state */
#define PLKN_NAME	(1<<4)	/**< \brief Player knows name */
#define PLKN_NAMEPL	(1<<5)	/**< \brief Player knows plural name */
#define PLKN_MODEL	(1<<6)	/**< \brief Player knows model filename */
#define PLKN_FACING	(1<<7)	/**< \brief Player knows facing */
#define PLKN_DIR	(1<<8)	/**< \brief Player knows direction */
#define PLKN_COORD	(1<<9)	/**< \brief Player knows location */
#define PLKN_SCALE	(1<<10)	/**< \brief Player knows scale */
#define PLKN_VELOC	(1<<11)	/**< \brief Player knows velocity */

#include "event.h"
#include "network.h"

player_t *new_player(void);
pl_known_t *find_pl_knowledge(player_t *pl, struct _object_t*obj);
int obj_removed_from_map(struct _event_t *event);
int obj_eact(struct _event_t *event);
int obj_wore_obj(struct _event_t *event);
int obj_unwore_obj(struct _event_t *event);
int update_pl_knowledge(player_t *pl, object_t *obj);
int player_perc_surrounds(struct _event_t *event);
int object_teach_players_self(struct _event_t *event);
int tell_player_surr_changes(struct _event_t *event);
int remove_player_game(struct _event_t *event);
void shift_known_maps(player_t *pl, int dir);
void inform_border_maps(player_t *pl, int mapnum);
int player_find_surrounds(struct _event_t *event);
int add_player_game(struct _descriptor_data_t *d, char *username,
		    char *charname, char *password);
int player_check_border_distance(struct _event_t *event);
int player_modpl(struct _event_t *event);
int player_rotate(struct _event_t *event);
int player_update_all_motion(struct _event_t *event);
int player_enter_mapg(struct _event_t *event);
void send_player_mapg(object_t *obj, struct _mapgroup_t *mapg);

#endif /* _PLAYER_H_ */
