/* $Id: sutils.h,v 1.9 2004/04/03 06:48:13 xb49 Exp $ */

#ifndef _SUTILS_H_
#define _SUTILS_H_

/**
   \file sutils.h
   \brief Server utility definitions
   \author Tim Rightnour
*/

#include "defines.h"
#include "object.h"
#include "map.h"

#define PREFER_HIGH     1 /**< \brief we prefer a high random number */
#define PREFER_LOW      0 /**< \brief we prefer a low random number */

/**
   \brief is x,z outside of maph?
   \param maph maphead to check
   \param xx x coordinate
   \param zz z coordinate
*/
#define xz_out_of_map(maph, xx, zz) \
    ((xx > (maph->x + SCALE_XZ * maph->width)) || \
     (zz > (maph->z + SCALE_XZ * maph->height)) || \
     (xx < maph->x) || (zz < maph->z))

/**
   \brief are the coordinates outside of maph?
   \param maph maphead to check
   \param coord vertex_t coordniates
*/
#define coord_out_of_map(maph, coord) \
    (xz_out_of_map(maph, (coord).x, (coord).z))

/**
   \brief what is the groundheight at our current coordinates?
   \param maph maphead to check
   \param coord vertex_t coordinates
*/
#define groundheight_at_mapcoord(maph, coord) \
    groundheight_at_mapxz(maph, coord.x, coord.z)

/**
   \brief is x,z outside the mapgroup?
   \param maph maphead to check
   \param xx x coordinate
   \param zz z coordinate
*/
#define xz_out_of_mapgroup(maph, xx, zz) \
    (xx < 0 || xx > maph->mapg->width * maph->width * SCALE_XZ || zz < 0 || \
     zz > (maph->mapg->nrofmaps/maph->mapg->width) * maph->height * SCALE_XZ)

/**
   \brief is the coordinate out of the mapgroup?
   \param maph maphead to check
   \param coord vertex_t coordinates
*/
#define coord_out_of_mapgroup(maph, coord) \
    (xz_out_of_mapgroup(maph, (coord).x, (coord).z))

int random_roll(int min, int max, object_t *obj, int goodbad);
int die_roll(int num, int size, object_t *obj, int goodbad);
int mapnum_at_xz(maphead_t *maph, float x, float z);
float groundheight_at_mapxz(maphead_t *maph, float x, float z);
int init_mutex(pthread_mutex_t *mutex);
#endif /* _SUTILS_H_ */
