/* $Id: ninterface.c,v 1.40 2004/04/01 02:53:50 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "tod.h"
#include "xmlparse.h"
#include "player.h"
#include "network.h"
#include "netparser.h"
#include "ninterface.h"
#include "settings.h"
#include "text.h"
/**
   \file ninterface.c
   \brief Interface between event and network engines
   \author Tim Rightnour
   \todo parse_command_adpl needs to be fixed
*/

extern settings_t settings;

argtable_t commtable[] = {
    {"adpl", COMMAND_ADPL, 0},		/* client */
    {"modpl", COMMAND_MODPL, 0},	/* client */
    {"rot", COMMAND_ROT, 0},		/* client */
    {"ready", COMMAND_READY, 0}		/* metaserver, other server */
};

const size_t comm_size = sizeof(commtable) / sizeof(argtable_t);

/**
   \brief generic null command parser
   \param d descriptor that triggered this parser
   \param args command arguments
*/

void parse_command_null(descriptor_data_t *d, pargs_t *args)
{
    return;
}

/**
   \brief Add a player to the game
   \param d descriptor that triggered this parser
   \param args command arguments
   \todo THIS FUNCTION IS BROKEN
*/

void parse_command_adpl(descriptor_data_t *d, pargs_t *args)
{
    int i;
    char *charname=NULL, *username=NULL, *password=NULL;

    for (i=0; args[i].cword != -1; i++) {
        switch (args[i].cword) {
	case NAME:
	    charname = add_string(args[i].arg.c);
	    break;
	case USERNAME:
	    username = add_string(args[i].arg.c);
	    break;
	case PASSWORD:
	    password = add_string(args[i].arg.c);
	    break;
	}
    }
    /* XXX FOR NOW */
    if (username == NULL)
	username = charname;
    if (username == NULL || charname == NULL || password == NULL)
	return;
    /* XXX we should probably send back a failure notice */

    /* XXX break this into two events:
       1) create the player and add it to the game, server_event
       2) put the player on a map, mapg event.
    */
    add_player_game(d, username, charname, password);
}

/**
   \brief client requests a modification to the player
   \param d descriptor that triggered this parser
   \param args command arguments
   we only allow certain stats to be modified by the player
*/

void parse_command_modpl(descriptor_data_t *d, pargs_t *args)
{
    int i;
    event_t *event;

    if (d->pl == NULL)
	return;
    if (d->pl->obj == NULL)
	return;

    for (i=0; args[i].cword != -1; i++) {
        switch (args[i].cword) {
	case STATE:
	    event = new_obj_event(Q_PLAYER, ETYPE_NONE, "player_modpl",
				  d->pl->obj);
	    event->action = args[i].arg.i;
	    queue_obj_event(event);
	    break;
	}
    }
}

/**
   \brief parse a player rotation request
   \param d descriptor that triggered this parser
   \param args command arguments
*/

void parse_command_rot(descriptor_data_t *d, pargs_t *args)
{
    int i;
    event_t *event;

    if (d->pl == NULL)
	return;
    if (d->pl->obj == NULL)
	return;

    event = new_obj_event(Q_PLAYER, ETYPE_NONE, "player_rotate", d->pl->obj);
    /* unset argument stay of type EVARG_NONE */
    for (i=0; args[i].cword != -1; i++) {
        switch (args[i].cword) {
	case FACEX:
	    evarg_setfloat(event, 0, args[i].arg.f);
	    break;
	case FACEY:
	    evarg_setfloat(event, 1, args[i].arg.f);
	    break;
	case FACEZ:
	    evarg_setfloat(event, 2, args[i].arg.f);
	    break;
	case DIRX:
	    evarg_setfloat(event, 3, args[i].arg.f);
	    break;
	case DIRY:
	    evarg_setfloat(event, 4, args[i].arg.f);
	    break;
	case DIRZ:
	    evarg_setfloat(event, 5, args[i].arg.f);
	    break;
	}
    }
    queue_obj_event(event);
}

/* tellings */

/* tell player about an action that occurred */

char *real_tell_eact(object_t *obj, int action)
{
    char *buf;
    char b2[40];

    buf = malloc(sizeof(char) * 1024);
    sprintf(buf, "eact id:%s act:%d\r", uid_sprint(b2,&obj->cli.uid), action);
    return(buf);
}

void tell_eact(player_t *pl, object_t *obj, int action)
{
    char *buf;

    buf = real_tell_eact(obj, action);
    lock_do_write(pl->d, buf);
    free(buf);
}

/* tell player that an object he allready knows about, belongs in the
   inventory of another object.  Inventory is done on demand for objects that
   are not the player, except in the case of externally visible equipment, and
   the player's own inventory
*/

char *real_tell_inv_obj(object_t *obj, object_t *cont, int bucket)
{
    char *buf;
    char b2[40],b3[40];

    buf = malloc(sizeof(char) * 1024);
    sprintf(buf, "inv id:%s cont:%s buk:%d\r", uid_sprint(b2,&obj->cli.uid), uid_sprint(b3,&cont->cli.uid),
	    bucket);
    return(buf);
}
   
void tell_inv_obj(player_t *pl, object_t *obj, object_t *cont, int bucket)
{
    char *buf;

    buf = real_tell_inv_obj(obj, cont, bucket);
    lock_do_write(pl->d, buf);
    free(buf);
}

char *real_tell_rminv_obj(object_t *obj, object_t *cont, int bucket)
{
    char *buf;
    char b2[40],b3[40];
    buf = malloc(sizeof(char) * 1024);
    sprintf(buf, "rminv id:%s cont:%s buk:%d\r", uid_sprint(b2,&obj->cli.uid), uid_sprint(b3,&cont->cli.uid),
	    bucket);
    return(buf);
}

void tell_rminv_obj(player_t *pl, object_t *obj, object_t *cont, int bucket)
{
    char *buf;

    buf = real_tell_rminv_obj(obj, cont, bucket);
    lock_do_write(pl->d, buf);
    free(buf);
}

char *real_tell_equip_obj(object_t *obj, object_t *eq, int slot, int draw)
{
    char *buf;
    char b2[40], b3[40];

    buf = malloc(sizeof(char) * 1024);
    sprintf(buf, "eq id:%s eq:%s sl:%d draw:%d\r",
	    uid_sprint(b2, &obj->cli.uid), uid_sprint(b3, &eq->cli.uid),
	    slot, draw);
    return(buf);
}

void tell_equip_obj(player_t *pl, object_t *obj, object_t *eq, int slot,
		    int draw)
{
    char *buf;

    /* XXX why? Because we do not properly handle obj->cli.equip.visible*/
    draw = 1;
    buf = real_tell_equip_obj(obj, eq, slot, draw);
    lock_do_write(pl->d, buf);
    free(buf);
}

char *real_tell_unequip_obj(object_t *obj, object_t *eq, int draw)
{
    char *buf,b2[40],b3[40];

    buf = malloc(sizeof(char) * 1024);
    sprintf(buf, "ueq id:%s eq:%s draw:%d\r", uid_sprint(b2,&obj->cli.uid), uid_sprint(b3,&eq->cli.uid), draw);
    return(buf);
}

void tell_unequip_obj(player_t *pl, object_t *obj, object_t *eq, int draw)
{
    char *buf;

    draw = 1; /* XXX why? */
    buf = real_tell_unequip_obj(obj, eq, draw);
    lock_do_write(pl->d, buf);
    free(buf);
}

/**
   tell a player about a new object (to them).  obj is the object to
   describe.  DO_EXTENDED will tell the player about
   the stats.  DO_DRAW/NO_DRAW to tell the client to draw or not draw the obj.
   \note expects obj to be locked
*/


char *real_tell_new_obj(object_t *obj, int extended, int draw)
{
    char *buf,b2[40];
    char *str, *con, *dex, *wis, *Int, *per;
#if 0
    float bb1x, bb1y, bb1z, bb2x, bb2y, bb2z;
    char buf2[256];
#endif

    buf = malloc(sizeof(char) * 1024);
    if (extended) {
	str = print_stat(obj->cli.stats.Str);
	dex = print_stat(obj->cli.stats.Dex);
	con = print_stat(obj->cli.stats.Con);
	wis = print_stat(obj->cli.stats.Wis);
	per = print_stat(obj->cli.stats.Per);
	Int = print_stat(obj->cli.stats.Int);
	sprintf(buf, "nuob id:%s nrof:%d st:%d nm:\"%s\" nmp:\"%s\" mdl:%s"
		" fx:%0.1f fy:%0.1f fz:%0.1f dx:%0.1f dy:%0.1f dz:%0.1f"
		" cx:%0.1f cy:%0.1f cz:%0.1f"
		" sx:%f sy:%f sz:%f v:%0.2f"
		" sex:%d type:%d sub:%d wght:%d lvl:%d xp:%d mat:%d"
		" val:%d str:%s dex:%s con:%s wis:%s int:%s per:%s fame:%d"
		" appr:%d hp:%d mhp:%d sp:%d msp:%d stm:%d mstm:%d"
		" food:%d wat:%d ac:%d spd:%d draw:%d\r", 
		uid_sprint(b2,&obj->cli.uid), obj->cli.nrof, obj->cli.state,
		obj->cli.name, obj->cli.name_pl, obj->cli.model,
		obj->cli.facing.x, obj->cli.facing.y, obj->cli.facing.z,
		obj->cli.direction.x, obj->cli.direction.y,
		obj->cli.direction.z,
		obj->cli.coord.x, obj->cli.coord.y, obj->cli.coord.z,
		obj->cli.scale.x, obj->cli.scale.y, obj->cli.scale.z,
		obj->cli.velocity, obj->cli.stats.sex, obj->cli.stats.type,
		obj->cli.stats.subtype, obj->cli.stats.weight,
		obj->cli.stats.level, obj->cli.stats.exp,
		obj->cli.stats.material, obj->cli.stats.value,
		str, dex, con, wis, Int, per,
		obj->cli.stats.fame, obj->cli.stats.appr, obj->cli.stats.hp,
		obj->cli.stats.maxhp, obj->cli.stats.sp, obj->cli.stats.maxsp, 
		obj->cli.stats.stm, obj->cli.stats.maxstm,
		obj->cli.stats.food, obj->cli.stats.water, obj->cli.stats.ac,
		obj->cli.stats.speed, draw);
	free(str);
	free(con);
	free(dex);
	free(wis);
	free(Int);
	free(per);
    } else {
	sprintf(buf, "nuob id:%s nrof:%d st:%d nm:\"%s\" nmp:\"%s\" mdl:%s"
		" fx:%0.1f fy:%0.1f fz:%0.1f dx:%0.1f dy:%0.1f dz:%0.1f"
		" cx:%0.1f cy:%0.1f cz:%0.1f sx:%f sy:%f sz:%f v:%0.2f"
		" draw:%d\r", uid_sprint(b2,&obj->cli.uid), 
		obj->cli.nrof,
	    obj->cli.state, obj->cli.name, obj->cli.name_pl, obj->cli.model,
	    obj->cli.facing.x, obj->cli.facing.y, obj->cli.facing.z,
	    obj->cli.direction.x, obj->cli.direction.y, obj->cli.direction.z,
	    obj->cli.coord.x, obj->cli.coord.y, obj->cli.coord.z,
	    obj->cli.scale.x, obj->cli.scale.y, obj->cli.scale.z,
	    obj->cli.velocity, draw);
    }
    return(buf);

#if 0
    printf("OBJ: %s :\r", obj->cli.name);
    dtBBObject(obj, &bb1x, &bb1y, &bb1z, &bb2x, &bb2y, &bb2z);
    bb1x -= obj->cli.coord.x;
    bb1y -= obj->cli.coord.y;
    bb1z -= obj->cli.coord.z;
    bb2x -= obj->cli.coord.x;
    bb2y -= obj->cli.coord.y;
    bb2z -= obj->cli.coord.z;
    printf("%f %f %f : %f %f %f\r", bb1x, bb1y, bb1z, bb2x, bb2y, bb2z);
    sprintf(buf2, "bbox id:%d lx:%f ly:%f lz:%f ux:%f uy:%f uz:%f\r",
	    obj->cli.num, bb1x, bb1y, bb1z, bb2x, bb2y, bb2z);
    do_write(pl->d, buf2);
#endif
}

/**
   \note expects obj to be locked
*/

void tell_new_obj(player_t *pl, object_t *obj, int extended, int draw)
{
    char *buf;

    buf = real_tell_new_obj(obj, extended, draw);
    lock_do_write(pl->d, buf);
    free(buf);
}

/**
   Tell the player a mapshift has occurred.  For us, this is easy, for the
   client, it requires a bit more work.
*/

void tell_mapshift(player_t *pl, maphead_t *newmap)
{
    int dir;
    char buf[256];

    dir = map_shift_direction(pl->obj->maphead, newmap);
    if (dir) {
	sprintf(buf, "maps act:%d\r", dir);
	lock_do_write(pl->d, buf);
	shift_known_maps(pl, dir);
    }
}

/**
   There are 9 potential maps we could tell the player about. Inform
   the player about whichever one was asked for.
*/

void tell_new_map(player_t *pl, int mapnum)
{
    char buf[256], *filedata, filename[MAX_PATH], *p;
    FILE *file;
    struct stat sb;
    maphead_t *maph;

    if (pl->obj->maphead->tile[mapnum] == NULL)
	return;
    maph = pl->obj->maphead->tile[mapnum];

    sprintf(filename, "%s/%s", maph->path, maph->map->elevfile); 

    if (stat(filename, &sb) == -1)
	bailout();
    filedata = safer_malloc(sizeof(char)*sb.st_size);
    file = fopen(filename, "r");
    if (file == NULL)
        bailout();
    if (fread(filedata, sb.st_size, 1, file) != 1)
	bailout();
    fclose(file);

    sprintf(buf, "mape mapid:%d width:%d height:%d cx:%0.1f cy:%0.1f cz:%0.1f bin:%d\r",
	    mapnum, maph->width, maph->height, maph->x, maph->y, maph->z,
	    (int)sb.st_size);
    LOG(LOG_WARNING, "SEND:%s %s\n", filename, buf);
    pthread_mutex_lock(&pl->d->mutex);
    do_write(pl->d, buf);
    /* we know this is a realloc, because do_write would have malloced it */
    pl->d->out_buf = realloc(pl->d->out_buf,
			     sizeof(char) * (pl->d->outbuflen+sb.st_size+2));
    p = pl->d->out_buf;
    p += pl->d->outbuflen;
    memcpy(p, filedata, sb.st_size);
    free(filedata);
    pl->d->outbuflen += sb.st_size;
    pthread_mutex_unlock(&pl->d->mutex);

    /* now we do the terrain file */

    sprintf(filename, "%s/%s", maph->path, maph->map->terrfile); 

    if (stat(filename, &sb) == -1)
	bailout();
    filedata = safer_malloc(sizeof(char)*sb.st_size);
    file = fopen(filename, "r");
    if (file == NULL)
        bailout();
    if (fread(filedata, sb.st_size, 1, file) != 1)
	bailout();
    fclose(file);

    sprintf(buf, "mapt mapid:%d width:%d height:%d cx:%0.1f cy:%0.1f cz:%0.1f bin:%d\r",
	    mapnum, maph->width, maph->height, maph->x, maph->y, maph->z,
	    (int)sb.st_size);
    
    pthread_mutex_lock(&pl->d->mutex);
    do_write(pl->d, buf);
    /* we know this is a realloc, because do_write would have malloced it */
    pl->d->out_buf = realloc(pl->d->out_buf,
			     sizeof(char) * (pl->d->outbuflen+sb.st_size+2));
    p = pl->d->out_buf;
    p += pl->d->outbuflen;
    memcpy(p, filedata, sb.st_size);
    free(filedata);
    pl->d->outbuflen += sb.st_size;
    pthread_mutex_unlock(&pl->d->mutex);

}

void tell_rmobj(player_t *pl,object_t *obj) {tell_rmobjbyid(pl, &obj->cli.uid);}

void tell_rmobjbyid(player_t *pl, uniqid_t*id)
{
    char buf[256],b2[40];

    sprintf(buf, "rmob id:%s\r",uid_sprint(b2,id));
    lock_do_write(pl->d, buf);
}

char *real_tell_text(char *data, int datatype, object_t *from)
{
    char *buf;

    buf = malloc(sizeof(char)*strlen(data) + 70);
    if(!from)
	sprintf(buf, "text type:%d data:\"%s\"\r", datatype, data);
    else
	sprintf(buf, "text nm:\"%s\" type:%d data:\"%s\"\r", from->cli.name,
		datatype, data);
    return(buf);
}

void tell_text(player_t *pl, char *data, int datatype, object_t *from)
{
    char *buf;

    buf = real_tell_text(data, datatype, from);
    lock_do_write(pl->d, buf);
    free(buf);
}

void say_text(player_t *pl, char *data, object_t *from, int volume )
{
    char *buf;

    buf = malloc(sizeof(char)*strlen(data) + 100);
    if(!from)
	sprintf(buf, "text type:%d volume:%d data:\"%s\"\r", (int)DATA_SAY,
		volume, data);
    else
	sprintf(buf, "text nm:\"%s\" type:%d volume:%d data:\"%s\"\r",
		from->cli.name, (int)DATA_SAY, volume, data);
    lock_do_write(pl->d, buf);
    free(buf);
}

void tell_servdefs(player_t *pl)
{
    char buf[256];

    sprintf(buf, "srvd hp:%d\r", HOURS_PER_DAY);
    lock_do_write(pl->d, buf);
}
