/* $Id: npc.c,v 1.15 2004/03/20 07:21:36 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "event.h"
#include "npc.h"

/**
   \file npc.c
   \brief NPC routines and AI
   \author Tim Rightnour
*/

extern TAILQ_HEAD(, _mapgroup_t) maps;
extern pthread_mutex_t maps_mutex;

/**
   \brief Make the npc perform some sort of AI routine
   \param event event that called us
   \return event status (allways 1)
   \todo this function is very very far from complete
*/

int npc_ehandler(event_t *event)
{
    switch (event->action) {
    case EACT_WANDER:
	npc_wander(event->initiator);
	break;
    }
    return(1);
}

/**
   \brief give the NPC something to do
   \param obj NPC to give something to do, to
*/

void npc_life(object_t *obj)
{
    event_t *event;

    /* for now, lets just make him wander */
    event = new_obj_event(Q_NPCLIFE, ETYPE_NPC, "npc_ehandler", obj);
    event->action = EACT_WANDER;
    queue_obj_event(event);
}

/**
   \brief make all the NPCs on the mapgroup do something.
   \param event event that called us
   \return event status (allways 0)
*/

int npc_life_handler(event_t *event)
{
    maphead_t *maph;
    object_t *obj;
    mapgroup_t * mapg = event_mapg(event);

    TAILQ_FOREACH(maph, &mapg->maps, next) {
	if (QUERY_FLAG(maph, MAPFLAG_LOADED)) {
	    TAILQ_FOREACH(obj, &maph->map->objs, next) {
		if (QUERY_FLAG(obj, FLAG_ALIVE) && obj->pl == NULL &&
		    !QUERY_FLAG(obj, FLAG_IMMOBILE)) {
		    npc_life(obj);
		}
	    }
	}
    }

    return(0);
}

/**
   \brief make the NPC wander randomly
   \param npc the NPC to make wander
*/

void npc_wander(object_t *npc)
{
    double ndir;

    printf("calling wander\n");
    if (rndm(1, 10) == 1) {
	ndir = (double)rndm(0, 359);
	npc->cli.direction.y = ndir;
    }
    /* make the NPC face the direction of motion */
    npc->cli.facing = npc->cli.direction;
    npc->cli.state = STATE_WALK;
}
