/* $Id: text.c,v 1.12 2004/03/24 14:30:01 xb49 Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif
#include <unistd.h>
#include <math.h>
#include <sys/param.h>

#include <SOLID/solid.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defines.h"
#include "object.h"
#include "map.h"
#include "shstr.h"
#include "utils.h"
#include "models.h"
#include "collision.h"
#include "tod.h"
#include "xmlparse.h"
#include "matrix.h"
#include "event.h"
#include "player.h"
#include "network.h"
#include "netparser.h"
#include "ninterface.h"
#include "commands.h"
#include "settings.h"
#include "text.h"
#include "stattables.h"

/**
   \file text.c
   \brief interface for sending raw text to the client
   \author Tim Rightnour
*/

extern settings_t settings;
extern TAILQ_HEAD(, _player_t) players;
extern pthread_mutex_t players_mutex;

/**
   \brief send the motd to the client
   \param event event that called us
   \return event status
*/

int send_motd(event_t *event)
{
    FILE *file;
    long int motdlen;
    char *buf, filename[MAX_PATH];
    object_t *obj;

    sprintf(filename, "%s/motd", settings.etcdir);
    file = fopen(filename, "r");
    if (file == NULL)
	return(1);

    obj = event->initiator;
    if (obj->pl == NULL)
	return(1);

    (void)fseek(file, 0, SEEK_END);
    motdlen = ftell(file);
    rewind(file);
    buf = malloc(sizeof(char) * motdlen+1);
    fread(buf, motdlen, 1, file);
    buf[motdlen] = '\0';
    fclose(file);
    tell_text(obj->pl, buf, DATA_MOTD, NULL);
    free(buf);
    return(1);
}

/**
   \brief tell players and NPCs around us what we said
   \param event event that called us
   \return event status
   \todo eventually, this function will need to pass the data through the
   translator, if required
   \todo this function needs to talk to NPC's as well
*/

int event_say(event_t *event)
{
    object_t *obj, *tmp;
    int volume;
    float rvol;
    char * data;

    obj = event->initiator;
    data = evarg_getstr(event,0);
    volume = evarg_getint(event,1);
    switch (volume) {
	case 1: rvol = .01 ; break ; /*whisper */
	case 2: rvol = .1 ; break;  /*say */
	case 3: rvol = 1. ; break ; /*shout */
	default : rvol = 5.; 
    };
    
    printf("SAY ============ %s =================\n",obj->cli.name);
    TAILQ_FOREACH(tmp, &obj->maphead->map->objs, next) {

	    if (tmp->pl != NULL){ /* only talk to palyers now*/
	    printf ("SAY %p obj:%s hd = %f  rvol=%f dist=%f\n",tmp,tmp->cli.name,(float)hearing_dist(tmp->cli.stats.Per) ,rvol,(float)	dist_between(tmp->cli.coord, obj->cli.coord));
	    if ((tmp == obj) || (hearing_dist(tmp->cli.stats.Per) * rvol >= dist_between(tmp->cli.coord, obj->cli.coord))) 
		say_text(tmp->pl, data,  obj,volume);
	}
    }
    return(1);
}

/**
   \brief gather a list of what players are online
   \param event event that called us
   \return event status
*/
   
int event_who(event_t *event)
{
    object_t *obj;
    player_t *pl;
    int nrofplayers=0;
    char *data, buf[256];

    obj = event->initiator;

    pthread_mutex_lock(&players_mutex);
    TAILQ_FOREACH(pl, &players, next)
	nrofplayers++;

    data = malloc(sizeof(char) * (nrofplayers+2) * 256);
    sprintf(data, "LVL  Name\n");
    sprintf(buf, "---+--------------------------------------------------\n");
    strcat(data, buf);

    TAILQ_FOREACH(pl, &players, next) {
	sprintf(buf, "% 3d  %s\n", pl->obj->cli.stats.level, pl->obj->cli.name);
	strcat(data, buf);
    }
    pthread_mutex_unlock(&players_mutex);
    sprintf(buf, "%d players are online\n", nrofplayers);
    tell_text(obj->pl, data, DATA_NEWS, NULL);
    return(1);
}
