/* $Id: plates.c,v 1.2 2004/03/18 07:59:11 kstenger Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Karla Stenger
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/


#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include "plates.h"
#include "var.h"
#include "defines.h"
#include "utils.h"

void plates_alloc(void){
    /* Allocates space for the MAXPLATE plates we will use. */
    struct plate *pl = safer_malloc(sizeof(struct plate));
    
    p.first = pl;
    p.last = pl;
    pl->id = 1;
    
    while (pl->id < MAXPLATE) { 
	pl->area = 0;
	pl->age = 0;
	pl->rx = 0;
	pl->ry = 0;
	pl->odx = 0;
	pl->ody = 0;
	pl->dx = 0;
	pl->dy = 0;
	pl->in_use = 0;
	
	pl->next = safer_malloc(sizeof(struct plate));
	pl->next->id = pl->id + 1;
	pl = pl->next;
	p.last = pl;
    }
    p.last->next = NULL;
    

}


void plates_free () {
    /* free()s the whole plate's structure. */
    struct plate *pl1 = p.first;
    struct plate *pl2;
    int i;
    
    for (i=1; i < MAXPLATE; i++) {
	pl2 = pl1->next;
	free(pl1);
	pl1 = pl2;
    }
}


int plate_add (void) {
    /* Finds an unused plate in the whole plate's structure returning it's id.
     * Marks it as beeing used for future reference. */
    struct plate *pl = p.first;
     
    while ((pl != p.last) && pl->in_use)
	pl = pl->next;

    if (pl->in_use)
	panic("No free plates!\n");

    pl->in_use = 1;

    return pl->id;
}


struct plate *get_plate(int n) {
    /* Finds plate number n and returns it.
     * returns NULL if it's not in_use or was not found. */
    struct plate *pl = p.first;
    int i;
    
    for (i=1; i != n; i++)
	if (pl->next == p.last)
	    return NULL;
	else
	    pl = pl->next;

    if (pl->in_use)
	return pl;
    else 
	return NULL;
}


void plate_remove(int n){
    /* Finds plate number n, clear it's values to zero.
     * Marks it as not beeing used anymore. */
    struct plate *pl = p.first;
    int i;
    
    for (i=1; i != n; i++)
	if (pl->next == p.last)
	    panic ("Tried to remove unexisting plate.\n");
	else
	    pl = pl->next;
    
    pl->area = 0;
    pl->age = 0;
    pl->rx = 0;
    pl->ry = 0;
    pl->odx = 0;
    pl->ody = 0;
    pl->dx = 0;
    pl->dy = 0;
    pl->in_use = 0;
    
}



