/* $Id: unix.c,v 1.15 2004/03/26 06:23:15 kstenger Exp $ */

/* This program is Copyright (c) 1990 David Allen.  It may be freely
   distributed as long as you leave my name and copyright notice on it.
   I'd really like your comments and feedback; send e-mail to
   davea@vlsi.ll.mit.edu, or send us-mail to David Allen, 10 O'Moore Ave,
   Maynard, MA 01754. 
   
   Copyright (C) 2003 Karla Stenger
   Copyright (C) 2003 Genecys Developer Team
*/

/* This file substitutes for ami.c.  The function draw() is called by 
   onestep() once every DRAWEVERY steps, and it just calls tecst() in 
   tec1.c.  Most of the functions in this file are dummies; see ami.c for
   more details on what they were. */


#include <sys/types.h>
#include <sys/timeb.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <inttypes.h>

#include "const.h"
#include "var.h"
#include "defines.h"
#include "utils.h"

extern int step;

extern char* optarg;
extern int   optind;

int XSIZE = 250, YSIZE = 250; /*mapsize*/
int MAXSTEP = 150;
int MAXBUMP = 250;
int BUMPTOL = 5,	DRAWMODE = 4;
int DRAWEVERY = 150,	BLOBLEVEL = 30;
int ZINIT = 122,	ZSUBSUME = 16,	ZCOAST = 16;
int ZSHELF = 1,		ZMOUNTAIN = 48, ZERODE = 6;
int RIFTPCT = 40,	DOERODE = 1,	ERODERND = 9/* 1 - 16 */;
int MAXCTRTRY = 500,	RIFTDIST = 20,	BENDEVERY = 2;
int BENDBY = 350,	SPEEDRNG = 300,	SPEEDBASE = 250;
int MAXLIFE = 10; /* Length of MR vector */
int XPERC = 90,		YPERC = 90;/* blob size over the whole map (percentage)*/
int MAXPLATE = 40;

double MR [] = { 1.0,1.0,1.0,0.7,0.4,0.1,-0.2,-0.5,-0.8,-1.0 };

/* These are all defined in tec1.c */
extern void init (void);
extern void onestep (void);
extern void tecst (int src, int drawmode, int index);


void panic (char *s);
void draw (int src);


int main (int argc, char **argv) { 
   time_t seed;

   char ch;

   /* Initialize random number generator */
   seed = time(NULL);

   while ((ch = getopt(argc, argv, \
		"x:y:X:Y:t:l:p:o:d:b:i:u:c:f:m:r:e:E:n:h:v:w:z:s:g:R:"))!=-1)
       switch (ch) {
	   case 'x': XSIZE = atoi(optarg); break;
	   case 'y': YSIZE = atoi(optarg); break;
	   case 'X': XPERC = atoi(optarg); break;
	   case 'Y': YPERC = atoi(optarg); break;
	   case 't': MAXSTEP = atoi(optarg); break;
	   case 'l': MAXLIFE = atoi(optarg); break;
	   case 'p': MAXBUMP = atoi(optarg); break;
	   case 'o': BUMPTOL = atoi(optarg); break;
	   case 'd': DRAWEVERY = atoi(optarg); break;
	   case 'b': BLOBLEVEL = atoi(optarg); break;
	   case 'i': ZINIT = atoi(optarg); break;
	   case 'u': ZSUBSUME = atoi(optarg); break;
	   case 'c': ZCOAST = atoi(optarg); break;
	   case 'f': ZSHELF = atoi(optarg); break;
	   case 'm': ZMOUNTAIN = atoi(optarg); break;
	   case 'r': RIFTPCT = atoi(optarg); break;
	   case 'e': DOERODE = atoi(optarg); break;
	   case 'E': ZERODE = atoi(optarg); break;
	   case 'n': ERODERND = atoi(optarg); break;
	   case 'h': MAXCTRTRY = atoi(optarg); break;
	   case 'v': RIFTDIST = atoi(optarg); break;
	   case 'w': BENDEVERY = atoi(optarg); break;
	   case 'z': BENDBY = atoi(optarg); break;
	   case 's': SPEEDBASE = atoi(optarg); break;
	   case 'g': SPEEDRNG = atoi(optarg); break;
	   case 'R': seed = atoi(optarg); break;
	   case '?': printf("Valid options are: x,y,X,Y,t,l,p,o,d,b,i,u,c,f,m,r,e,E,n,h,v,w,z,s,g,R,\nsee params.doc for details");return 0;
       }

   srandom(seed);
   printf("-- Generating map with random seed %d --\n", (int)seed);
   
   /* Run the program */
   init();
   printf("Done with initialization.\n");
   for (step=0; step<=MAXSTEP; step++) 
       onestep ();
   
   printf("-- Map generated with random seed %d --\n", (int)seed);
   exit (0);
}

void panic (char *s) {
    printf ("PANIC: %s\n", s); 
    exit (1); 
}

void draw (int src) {
    tecst (src, DRAWMODE, step);
}

