/* $Id: translate.c,v 1.1 2004/03/19 07:36:50 garbled Exp $ */
/*
    Genecys, A Multiplayer Online RPG

    Copyright (C) 2003 Tim Rightnour
    Copyright (C) 2003 Genecys Developer Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The authors can be reached via e-mail at genecys-devel@lists.sourceforge.net
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/stat.h>
#if defined(__linux__)
#include "queue.h"
#else
#include <sys/queue.h>
#endif

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "defines.h"
#include "xmlparse.h"
#include "utils.h"
#include "language.h"

/**
   \file translate.c
   \brief Language testing tool.
   \author Tim Rightnour
*/

TAILQ_HEAD(, _lang_t) langs;
extern char *optarg;


void usage(void)
{
    fprintf(stderr,
            "Valid options:\n"
            "-f <filename>  : input filename\n"
            "-l <language>  : language number to translate to\n"
	    "-k <knowledge> : 1-100, how well you know lang.\n"
	    "Must provide language number and filename.\n");
    exit(1);
}


int main(int argc, char **argv)
{
    FILE *file;
    char *filename=NULL, *inbuf, *outbuf, ch;
    int lang=0, knowledge=0;
    struct stat sb;

    while ((ch = getopt(argc, argv, "l:f:k:")) != -1) {
        switch (ch) {
	case 'l': lang=atoi(optarg); break;
	case 'k': knowledge=atoi(optarg); break;
	case 'f': filename = strdup(optarg); break;
	default: usage(); break;
	}
    }
    if (filename == NULL || lang == 0)
	usage();

    parse_languages("data/libdata", "data/etc/languages.xml");

    if (stat(filename, &sb) == -1)
        bailout();
    inbuf = safer_malloc(sizeof(char)*sb.st_size+1);
    file = fopen(filename, "r");
    if (file == NULL)
        bailout();
    if (fread(inbuf, sb.st_size, 1, file) != 1)
        bailout();
    fclose(file);
    inbuf[sb.st_size] = '\0';

    outbuf = translate_string(inbuf, lang, knowledge);
    printf("%s", outbuf);
    return(0);
}
